package de.heinerkuecker.asm_improved;

import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import de.heinerkuecker.asm_improved.annotation_use.AsmAnnotationUse;
import de.heinerkuecker.asm_improved.type.AsmClassType;
import de.heinerkuecker.asm_improved.type.AsmType;
import de.heinerkuecker.asm_improved.type.generics.param.AsmAbstractTypeParam;
import de.heinerkuecker.asm_improved.type.generics.param.AsmExtendsTypeParam;
import de.heinerkuecker.asm_improved.type.util.TypeArgsUtil;
import de.heinerkuecker.asm_improved.util.CollectionUtil;
import de.heinerkuecker.asm_improved.util.ExtendedArrayList;
import de.heinerkuecker.asm_improved.util.Fqn;
import de.heinerkuecker.asm_improved.util.FqnAndSourceFileNameTuple;
import de.heinerkuecker.asm_improved.util.ImmutableArray;
import de.heinerkuecker.asm_improved.util.ImmutableMap;
import de.heinerkuecker.asm_improved.util.SeparatedStrBuff;

/**
 * Informationen zu einer Klasse, Interface, Enum oder Annotation,
 * also aller Entitten, die durch eine *.class-Datei reprsentiert
 * werden.
 * 
 * TODO eigentlich sollte ich hier eine Lsung entsprechend algebraischer Datentypen machen, AbstractClassInfo, InterfaceInfo, NonAbstractClassInfo
 */
public final class AsmClassInfo
{
	/**
	 * *.class-File oder *.jar-File
	 * aus welchem die Informationen
	 * dieses Objekts gescannt wurden
	 */
	public final File file;

	/**
	 * Vollqualifizierter Name mit
	 * eindeutiger Abtrennung Package
	 * und eventuell Punkt-getrenntem
	 * Namen fr innere oder
	 * geschachtelte Klasse.
	 */
	public final Fqn fqn;

	public final boolean isPrivate;

	public final boolean isProtected;

	public final boolean isPublic;

	public final boolean isAbstract;

	public final boolean isStatic;

	public final boolean isFinal;

	public final boolean isInterface;

	public final boolean isEnum;

	public final boolean isAnnotation;

	//public final boolean isTopLevel;

	//public final boolean isInnerOrNested;

	/**
	 * Typ-Parameter.
	 * 
	 * TODO eventuell wre eine Map besser, Name ist wichtig, Reihenfolge ist egal
	 */
	public final ImmutableArray<? extends AsmAbstractTypeParam> typeParams;

	/**
	 */
	public final AsmClassType superType;

	/**
	 */
	public final ImmutableArray<? extends AsmClassType> interfaceTypes;

	/**
	 * Felder.
	 */
	public final ImmutableMap<String, ? extends AsmFieldInfo> fields;

	/**
	 * Methoden.
	 */
	public final ImmutableMap<String, ? extends ImmutableArray<? extends AsmMethodInfo>> methods;

	/**
	 * Quelldatei-Name aus Bytecode ausgelesen.
	 */
	public final String sourceFileName;

	/**
	 * Annotationen.
	 */
	public final ImmutableArray<? extends AsmAnnotationUse> annotations;

	/**
	 * Bei anonymer oder lokaler Klasse die {@link Fqn} der einschliessenden Klasse.
	 */
	public final Fqn enclosingClassFqn;

	/**
	 * Bei anonymer oder lokaler Klasse der Name der einschliessenden Methode.
	 */
	public final String enclosingMethodName;
	
	/**
	 * Konstruktor.
	 * 
	 */
	AsmClassInfo(
			final File file ,
			final Fqn fqn ,
			final boolean isPrivate ,
			final boolean isProtected ,
			final boolean isPublic ,
			final boolean isAbstract ,
			final boolean isStatic ,
			final boolean isFinal ,
			final boolean isInterface ,
			final boolean isEnum ,
			final boolean isAnnotation ,
			//final boolean isTopLevel ,
			//final boolean isInnerOrNested ,
			final ImmutableArray<? extends AsmAbstractTypeParam> typeParams ,
			final AsmClassType superType ,
			final ImmutableArray<? extends AsmClassType> interfaceTypes ,
			final ImmutableMap<String, ? extends AsmFieldInfo> fields ,
			final ImmutableMap<String, ? extends ImmutableArray<? extends AsmMethodInfo>> methods ,
			final String sourceFileName ,
			final ImmutableArray<? extends AsmAnnotationUse> annotations ,
			final Fqn enclosingClassFqn ,
			final String enclosingMethodName )
	{
		//if ( fqn.toFqnStr().equals( "de.heinerkuecker.parser.java8.test.class_.inner.ClassWithInnerClassWithGenerics_0.Inner" ) )
		//{
		//	System.out.println( fqn + " static " + isStatic );
		//}
		//if ( fqn.toFqnStr().startsWith( "de.heinerkuecker.parser.java8.test.field.from_expression.outer_class.ClassWithAnonymusClass_1$" ) )
		//{
		//	System.out.println( "debug break" );
		//}

		if ( isInterface )
		{
			if ( superType != null )
			{
				throw new IllegalArgumentException(
						"interface with super class " +
								/*superClassTuple*/superType.fqn );
			}
		}
		else if ( ( ! "java.lang.Object".equals( fqn.toFqnStr() ) ) &&
				( superType == null ) )
		{
			throw new IllegalArgumentException(
					"class without super class " +
							String.valueOf( fqn ) );
		}

		if ( file == null )
		{
			throw new IllegalArgumentException(
					"file is null" );
		}
		if ( fqn == null )
		{
			throw new IllegalArgumentException(
					"fqn is null" );
		}

		// TODO enclosingMethodName == null kommt bei statischen Methoden vor
		//if ( ( enclosingClassFqn == null ) != ( enclosingMethodName == null ) )
		//	// beide null oder beide nicht null
		//{
		//	throw new IllegalArgumentException( "enclosing" );
		//}

		this.file = file;
		this.fqn = fqn;
		this.isPrivate = isPrivate;
		this.isProtected = isProtected;
		this.isPublic = isPublic;
		this.isAbstract = isAbstract;
		this.isStatic = isStatic;
		this.isFinal = isFinal;
		this.isInterface = isInterface;
		this.isEnum = isEnum;
		this.isAnnotation = isAnnotation;
		//this.isTopLevel() = isTopLevel;
		//this.isInnerOrNested = isInnerOrNested;
		this.typeParams = typeParams;
		this.superType = superType;
		this.interfaceTypes = interfaceTypes;
		this.fields = fields;
		this.methods = methods;
		this.sourceFileName = sourceFileName;
		this.annotations = annotations;
		this.enclosingClassFqn = enclosingClassFqn;
		this.enclosingMethodName = enclosingMethodName;
	}

	/**
	 * @return default/package scope
	 */
	public boolean isDefault()
	{
		return ! ( this.isPublic || this.isProtected || this.isPrivate );
	}

	public boolean isInner()
	{
		return
				( ! this.isStatic ) &&
				(
						// anonyme oder lokale Klasse
						this.enclosingClassFqn != null 
						||
						// Name hat Punkt
						( ! this.fqn.dottedName.isTopLevel() ) 
						);
	}

	public boolean isNested()
	{
		return
				this.isStatic &&
				( ! this.fqn.dottedName.isTopLevel() );
	}

	public boolean isInnerOrNested()
	{
		return isInner() || isNested();
	}

	/**
	 * @return
	 */
	public boolean isTopLevel()
	{
		return ! isInnerOrNested();
	}

	/**
	 * @return {@link Fqn} der usseren Klasse dieser Klasse
	 * @throws IllegalStateException wenn es keine ussere Klasse gibt
	 */
	public Fqn getOuterClassFqn()
	{
		if ( /*this.isTopLevel*/( ! this.isInnerOrNested() ) ||
				isStatic || this.fields == null || this.fields.isEmpty() )
		{
			throw new IllegalStateException( "no outer class" );
			// eventuell: return null;
		}

		//AsmFieldInfo outerField = null;
		//
		//for ( final AsmFieldInfo field : this.fields.values() )
		//{
		//	if ( field.isSynthetic &&
		//			field.name.startsWith(
		//					// TODO Names-Anfang pruefen, stimmt dies immer ?
		//					"this$" ) )
		//	{
		//		outerField = field;
		//		break;
		//	}
		//}
		//
		//if ( outerField == null )
		//{
		//	throw new IllegalStateException( "no outer class" );
		//	// eventuell: return null;
		//}
		//
		//return outerField.declaringClassFqn;

		return new Fqn(
				this.fqn.packageName ,
				this.fqn.dottedName.getOuterClassName() );
	}

	/**
	 * @return ob diese Klasse eine ussere Klasse hat
	 */
	public boolean hasOuterClass()
	{
		//if (  /*this.isTopLevel*/( ! this.isInnerOrNested() ) ||
		//		this.isStatic )
		//{
		//	return false;
		//}
		//
		////final AsmFieldInfo outerField =
		////		this.fields.get(
		////				// TODO Prfen, ob dies immer stimmt
		////				// TODO Konstante
		////				"this$0" );
		////
		////if ( outerField == null )
		////{
		////	// TODO ist dies berhaupt mglich/erlaubt
		////	return false;
		////}
		////
		////return true;
		//return ! this.fqn.dottedName.isTopLevel();
		return this.isInner();
	}

	/**
	 * @param asmClasspathInfo AsmClasspathInfo
	 * @return Typ-Parameter dieser Klasse einschliesslich der Type-Parameter einer usseren Klasse dieser Klasse, falls diese Klasse eine ussere Klasse hat
	 */
	public ImmutableMap<String, AsmAbstractTypeParam> getTypeparamsIncludeOuterClass(
			final AsmClasspathInfo asmClasspathInfo )
	{
		final HashMap<String, AsmAbstractTypeParam> result = new HashMap<>();

		if ( this.hasOuterClass() )
		{
			final AsmClassInfo outerClassInfo = asmClasspathInfo.getClassInfoMandantory( getOuterClassFqn() );

			final ImmutableMap<String,AsmAbstractTypeParam> outerClassTypeparams = outerClassInfo.getTypeparamsIncludeOuterClass( asmClasspathInfo );

			if ( outerClassTypeparams != null )
			{
				for ( Entry<String, AsmAbstractTypeParam> outerClassTypeparamEntry : outerClassTypeparams.entrySet() )
				{
					result.put(
							outerClassTypeparamEntry.getKey() ,
							outerClassTypeparamEntry.getValue() );
				}
			}
		}

		if ( this.typeParams != null )
		{
			for ( AsmAbstractTypeParam typeParam : this.typeParams )
			{
				result.put(
						typeParam.name ,
						typeParam );
			}
		}

		if ( result.isEmpty() )
		{
			return null;
		}

		return ImmutableMap.create( result );
	}

	/**
	 * Methode zum Suchen eines Felds,
	 * welches in einer Methode dieser
	 * Klasse sichtbar ist.
	 * <br/>
	 * Also entweder in der gleichen Klasse
	 * private oder mit anderem Modifier,
	 * in einer Oberklasse protected oder
	 * default, wenn die Klasse im
	 * gleichen Package liegt oder
	 * protected oder public.
	 * 
	 * @param fieldNameToTest Name des gesuchten Felds
	 * @param asmClasspathInfo ASM-Classpath-Information
	 * @return ob vorhanden
	 */
	@Deprecated
	public boolean hasFieldVisibleInOwnMethod(
			final String fieldNameToTest ,
			final AsmClasspathInfo asmClasspathInfo )
	{
		final AsmFieldInfo fieldInfo =
				getFieldVisibleInOwnMethod(
						fieldNameToTest ,
						asmClasspathInfo );

		final boolean result = fieldInfo != null;

		return result;
	}

	public boolean hasFieldVisibleForCaller(
			final String fieldNameToTest ,
			//final PackageName callerPackageName ,
			//final SourceFileNameWithoutPath callerSourceFileNameWithoutPath ,
			final FqnAndSourceFileNameTuple callerFqnAndSourceTpl ,
			final AsmClasspathInfo asmClasspathInfo )
	{
		if ( fieldNameToTest.equals( "class" ) ||
				fieldNameToTest.equals( "super" ) ||
				fieldNameToTest.equals( "this" ) )
		{
			// reservierter Name
			return false;
		}

		final AsmFieldInfo fieldVisibleForCaller =
				getFieldVisibleForCaller(
						fieldNameToTest ,
						callerFqnAndSourceTpl ,
						asmClasspathInfo );

		return fieldVisibleForCaller != null;
	}

	public AsmFieldInfo getFieldVisibleForCaller(
			final String fieldNameToTest ,
			final FqnAndSourceFileNameTuple callerFqnAndSourceTpl ,
			final AsmClasspathInfo asmClasspathInfo )
	{
		if ( this.fields.containsKey(
				fieldNameToTest ) )
		{
			final AsmFieldInfo thisFieldInfo = this.fields.get( fieldNameToTest );

			if ( thisFieldInfo.isPublic )
			{
				return thisFieldInfo;
			}

			if ( thisFieldInfo.isProtected )
			{
				if ( callerFqnAndSourceTpl.fqn.packageName.equals( this.fqn.packageName ) )
					// protected erlaubt Sichtbarkeit im gleichen Package
				{
					return thisFieldInfo;
				}

				final AsmClassInfo callerClassInfo =
						asmClasspathInfo.getClassInfoMandantory(
								callerFqnAndSourceTpl.fqn );

				if ( callerClassInfo.isSuperClassOrImplementedInterface(
						fqn ,
						asmClasspathInfo ) )
				{
					return thisFieldInfo;
				}

				if ( isOuterClassSubClassOfThis(callerClassInfo, asmClasspathInfo) )
				{
					return thisFieldInfo;
				}
			}

			if ( ! ( thisFieldInfo.isProtected || thisFieldInfo.isPublic || thisFieldInfo.isPrivate ) )
				// default, package-scope
			{
				//if ( callerFqnAndSourceTpl.fqn != null && callerFqnAndSourceTpl.fqn.packageName.equals( this.fqn.packageName ) )
				if ( callerFqnAndSourceTpl.packageName.equals( this.fqn.packageName ) )
				{
					return thisFieldInfo;
				}
			}

			if ( callerFqnAndSourceTpl.fqn != null &&
					AsmClasspathInfo.isSameSourceFile(
					this ,
					//callerPackageName
					callerFqnAndSourceTpl.fqn.packageName ,
					//callerSourceFileNameWithoutPath
					callerFqnAndSourceTpl.sourceFileNameWithoutPath ) )
				// in der gleichen Quell-Datei sind alle Felder sichtbar
			{
				return thisFieldInfo;
			}
		}

		// super-Klasse
		if ( this.superType != null )
		{
			final AsmFieldInfo tmpResult =
					asmClasspathInfo.getClassInfoMandantory(
							this.superType.fqn )
					.getFieldVisibleForCaller(
							fieldNameToTest ,
							callerFqnAndSourceTpl ,
							asmClasspathInfo );

			if ( tmpResult != null )
			{
				return tmpResult;
			}
			//throw new RuntimeException( "not implemented" );
		}

		// super-Interfaces
		if ( this.interfaceTypes != null )
		{
			for ( final AsmClassType interfaceType : this.interfaceTypes )
			{
				final AsmFieldInfo tmpResult =
						asmClasspathInfo.getClassInfoMandantory(
								interfaceType.fqn )
						.getFieldVisibleForCaller(
								fieldNameToTest ,
								callerFqnAndSourceTpl ,
								asmClasspathInfo );

				if ( tmpResult != null )
				{
					return tmpResult;
				}
			}
			//throw new RuntimeException( "not implemented" );
		}

		// TODO 08.08.2016 muss anders geregelt wrden: Sichtbarkeit aufgrund gleicher Quell-Datei
		//final ImmutableArray<AsmClassInfo> classesInSameSourceFile =
		//		getOtherClassesInSameSourceFile(
		//				asmClasspathInfo );
		//
		//for ( final AsmClassInfo classInSameSourceFile : classesInSameSourceFile )
		//{
		//	if ( classInSameSourceFile.fields != null )
		//	{
		//		final AsmFieldInfo tmpResult =
		//				classInSameSourceFile.fields.get(
		//						fieldNameToTest );
		//
		//		if ( tmpResult != null )
		//		{
		//			return tmpResult;
		//		}
		//	}
		//}

		// TODO sichtbarkeit aus anonymer oder lokaler Klasse loese ich anders, damit beim An-Navigieren gleich die richtige Scope-Expression gesetzt wird
		//if ( this.isInnerOrNested() && ( ! this.isStatic ) )
		//	// anonyme oder lokale Klasse
		//{
		//	final AsmFieldInfo tmpResult =
		//			asmClasspathInfo.getClassInfoMandantory(
		//					this.getOuterClassFqn() )
		//			.getFieldVisibleForCaller(
		//					fieldNameToTest ,
		//					callerFqnAndSourceTpl ,
		//					asmClasspathInfo );
		//
		//	if ( tmpResult != null )
		//	{
		//		return tmpResult;
		//	}
		//}
		
		return null;
	}

	public AsmFieldInfo getStaticFieldInSameSourcefile(
			final String fieldNameToTest ,
			final FqnAndSourceFileNameTuple callerFqnAndSourceTpl ,
			final AsmClasspathInfo asmClasspathInfo )
	{
		final ImmutableArray<AsmClassInfo> classesInSameSourceFile =
				getOtherClassesInSameSourceFile(
						asmClasspathInfo );
		
		for ( final AsmClassInfo classInSameSourceFile : classesInSameSourceFile )
		{
			if ( classInSameSourceFile.fields != null )
			{
				final AsmFieldInfo tmpResult =
						classInSameSourceFile.fields.get(
								fieldNameToTest );
		
				if ( tmpResult != null &&
						tmpResult.isStatic )
				{
					return tmpResult;
				}
			}
		}
		
		return null;
	}

	/**
	 * Wenn die uebergebene Klasse eine innere Klasse ist,
	 * darf sie Felder sehen/lesen/schreiben, die von
	 * ihrer aeusseren Klasse gesehen werden darf.
	 * 
	 * @param classInfoToTest
	 * @param asmClasspathInfo
	 * @return ob eine aeussere Klasse der uebergebenen Klasse eine Unterklasse dieser Klasse ist
	 */
	private boolean isOuterClassSubClassOfThis(
			final AsmClassInfo classInfoToTest ,
			final AsmClasspathInfo asmClasspathInfo )
	{
		if ( classInfoToTest.hasOuterClass() )
		{
			final AsmClassInfo outerClassInfo =
					asmClasspathInfo.getClassInfoMandantory(
							classInfoToTest.getOuterClassFqn() );

			if ( outerClassInfo.isSuperClassOrImplementedInterface(
					fqn ,
					asmClasspathInfo ) )
			{
				return true;
			}

			// recursive call for multi inner classes
			return isOuterClassSubClassOfThis(
					outerClassInfo ,
					asmClasspathInfo );
		}
		return false;
	}
	
	/**
	 * Methode zum Suchen eines Felds,
	 * welches in einer Methode dieser
	 * Klasse sichtbar ist.
	 * <br/>
	 * Also entweder in der gleichen Klasse
	 * private oder mit anderem Modifier,
	 * in einer Oberklasse protected oder
	 * default, wenn die Klasse im
	 * gleichen Package liegt oder
	 * protected oder public.
	 * 
	 * @param fieldNameToTest Name des gesuchten Felds
	 * @param asmClasspathInfo 
	 * @return ob vorhanden
	 */
	@Deprecated
	public /*boolean*/ AsmFieldInfo getFieldVisibleInOwnMethod(
			final String fieldNameToTest ,
			final AsmClasspathInfo asmClasspathInfo )
	{
		// TODO die writable-Eigenschaft (nicht final) wird nicht geprft, Unit-Test machen
		if ( this.fields.containsKey(
				fieldNameToTest ) )
		{
			return /*true*/this.fields.get( fieldNameToTest );
		}

		// TODO Sichtbarkeit aus Oberklasse oder aufgrund gleicher Quell-Datei wird noch nicht beachtet
		if ( this.superType != null )
		{
			final AsmFieldInfo fieldInfoFromSuperClass =
					//this.superClassTuple.classInfoLazy.get()
					asmClasspathInfo.getClassInfoMandantory( this.superType.fqn )
					/*hasWritableFieldVisibleInSubClass*/
					.getFieldVisibleInSubClass(
							fieldNameToTest ,
							//this ,
							this );

			if ( fieldInfoFromSuperClass != null )
			{
				return /*true*/fieldInfoFromSuperClass;
			}
		}

		// TODO in Interfaces knnen auch Felder stecken, allerdings nur final
		if ( this.interfaceTypes != null )
		{
			for ( final AsmClassType interfaceType : this.interfaceTypes )
			{
				final AsmFieldInfo fieldInfoFromSuperInterface =
						asmClasspathInfo.getClassInfoMandantory( interfaceType.fqn )
						/*hasWritableFieldVisibleInSubClass*/
						.getFieldVisibleInSubClass(
								fieldNameToTest ,
								//this ,
								this );

				if ( fieldInfoFromSuperInterface != null )
				{
					return /*true*/fieldInfoFromSuperInterface;
				}
			}
			//throw new RuntimeException( "not implemented" );
		}

		// TODO sichtbarkeit aus anonymer oder lokaler Klasse loese ich anders, damit beim An-Navigieren gleich die richtige Scope-Expression gesetzt wird
		//if ( this.hasOuterClass() )
		//{
		//	final AsmClassInfo outerClassInfo = asmClasspathInfo.getClassInfoMandantory( getOuterClassFqn() );
		//
		//	final AsmFieldInfo fieldInfoFromOuterClass =
		//			asmClasspathInfo.getClassInfoMandantory( outerClassInfo.fqn )
		//			.getFieldVisibleInSubClass(
		//					fieldNameToTest ,
		//					// Sichtbar von aeusserer Klasse
		//					outerClassInfo );
		//
		//	if ( fieldInfoFromOuterClass != null )
		//	{
		//		return /*true*/fieldInfoFromOuterClass;
		//	}
		//}

		return /*false*/null;
	}

	/**
	 * Methode zum Prfen, ob diese Klasse ein
	 * Feld hat, welches in der bergebenen
	 * {@link AsmClassInfo} accessorClassInfo
	 * sichtbar ist.
	 * <br/>
	 * Dabei kann die Sichtbarkeit durch
	 * die Ableitung der {@link AsmClassInfo} accessorClassInfo
	 * von {@link AsmClassInfo} subClassInfo
	 * bei Protected, default oder public
	 * Feldern mglich sein.
	 * 
	 * @param fieldNameToTest Name des zu prfenden Feldes
	 * @param subClassInfo in Vererbungshierarchie zwischenliegende Klasse
	 * @param accessorClassInfo in dieser Klasse soll auf das Feld zugegriffen werden
	 * @return
	 */
	public /*boolean*/AsmFieldInfo getFieldVisibleInSubClass(
			final String fieldNameToTest ,
			//final ClassInfo subClassInfo ,
			final AsmClassInfo accessorClassInfo )
	{
		final AsmFieldInfo infoOfFieldToTest =
				this.fields.get( fieldNameToTest );

		if ( infoOfFieldToTest != null )
		{
			if ( infoOfFieldToTest.isProtected ||
					infoOfFieldToTest.isPublic )
			{
				return /*true*/infoOfFieldToTest;
			}

			if ( ( ! infoOfFieldToTest.isPrivate ) &&
					this./*classPackage*/fqn.packageName.equals(
							accessorClassInfo./*classPackage*/fqn.packageName ) )
			{
				return /*true*/infoOfFieldToTest;
			}

			if ( AsmClasspathInfo.isSameSourceFile(
					this ,
					accessorClassInfo ) )
				// in der gleichen Quell-Datei sind alle Felder sichtbar
			{
				return /*true*/infoOfFieldToTest;
			}
		}

		return /*false*/null;
	}


	public boolean hasMethodVisibleForCaller(
			final String methodNameToTest ,
			//final PackageName callerPackageName ,
			//final SourceFileNameWithoutPath callerSourceFileNameWithoutPath ,
			final Fqn callerFqn ,
			final AsmClasspathInfo asmClasspathInfo )
	{
		final ImmutableArray<? extends AsmMethodInfo> methodsVisibleForCaller =
				getMethodsVisibleInClass(
						methodNameToTest ,
						asmClasspathInfo.getClassInfoMandantory(
								callerFqn ) ,
						asmClasspathInfo );

		return methodsVisibleForCaller != null && ( ! methodsVisibleForCaller.isEmpty() );
	}

	@Deprecated
	public boolean hasMethodVisibleInOwnMethod(
			final String methodNameToTest ,
			final AsmClasspathInfo asmClasspathInfo )
	{
		final ImmutableArray<? extends AsmMethodInfo> methodInfos =
				getMethodsVisibleInOwnMethod(
						methodNameToTest ,
						asmClasspathInfo );

		final boolean result = methodInfos != null && ( ! methodInfos.isEmpty() );

		return result;
	}

	@Deprecated
	// Achtung, hier koennen auch Methoden aus anderen Klassen zurueckgegeben werden
	public ImmutableArray<? extends AsmMethodInfo> getMethodsVisibleInOwnMethod(
			final String methodNameToGet ,
			final AsmClasspathInfo asmClasspathInfo )
	{
		final ExtendedArrayList<AsmMethodInfo> result = new ExtendedArrayList<>();

		if ( this.methods.containsKey(
				methodNameToGet ) )
		{
			result.addAll(
					this.methods.get( methodNameToGet ) );
		}

		if ( this.superType == null )
			// Sichtbarkeit aus Oberklasse
		{
			if ( ! this.fqn.equals( Fqn.JAVA_LANG_OBJECT_FQN ) )
				// avoid stack overflow
			{
				final List<? extends AsmMethodInfo> methodInfosFromSuperClass =
						asmClasspathInfo.getClassInfoMandantory( Fqn.JAVA_LANG_OBJECT_FQN )
						.getMethodsVisibleInSubClass(
								methodNameToGet ,
								this ,
								asmClasspathInfo );

				if ( methodInfosFromSuperClass != null )
				{
					result.addAll(
							methodInfosFromSuperClass );
				}
			}
		}
		else
		{
			final List<? extends AsmMethodInfo> methodInfosFromSuperClass =
					asmClasspathInfo.getClassInfoMandantory( this.superType.fqn )
					.getMethodsVisibleInSubClass(
							methodNameToGet ,
							this ,
							asmClasspathInfo );

			if ( methodInfosFromSuperClass != null )
			{
				result.addAll(
						methodInfosFromSuperClass );
			}
		}

		if ( interfaceTypes != null )
			// in Interfaces knnen auch Methoden stecken
		{
			for ( final AsmClassType interfaceType : interfaceTypes )
			{
				final List<? extends AsmMethodInfo> methodInfosFromInterface =
						asmClasspathInfo.getClassInfoMandantory( interfaceType.fqn )
						.getMethodsVisibleInSubClass(
								methodNameToGet ,
								this ,
								asmClasspathInfo );

				if ( methodInfosFromInterface != null )
				{
					result.addAll(
							methodInfosFromInterface );
				}
			}
		}

		// Sichtbarkeit aufgrund gleicher Quell-Datei
		final ImmutableArray<AsmClassInfo> classesInSameSourceFile =
				getOtherClassesInSameSourceFile(
						asmClasspathInfo );

		for ( final AsmClassInfo classInSameSourceFile : classesInSameSourceFile )
		{
			final ImmutableArray<? extends AsmMethodInfo> methodsVisibleInClass =
					classInSameSourceFile.getMethodsVisibleInClass(
							methodNameToGet ,
							//callerClass
							this ,
							asmClasspathInfo );

			if ( methodsVisibleInClass != null )
			{
				result.addAll( methodsVisibleInClass );
			}
		}

		return ImmutableArray.create( result );
	}

	public List<? extends AsmMethodInfo> getMethodsVisibleInSubClass(
			final String methodNameToGet ,
			//final ClassInfo subClassInfo ,
			final AsmClassInfo accessorClassInfo ,
			final AsmClasspathInfo asmClasspathInfo )
	{
		final ExtendedArrayList<AsmMethodInfo> result = new ExtendedArrayList<>();

		final ImmutableArray<? extends AsmMethodInfo> thisMethodInfosOfToTest =
				this.methods.get(
						methodNameToGet );

		if ( thisMethodInfosOfToTest != null )
		{
			if ( thisMethodInfosOfToTest.get( 0 ).isProtected ||
					thisMethodInfosOfToTest.get( 0 ).isPublic )
			{
				result.addAll(
						thisMethodInfosOfToTest );
			}

			if ( ( ! thisMethodInfosOfToTest.get( 0 ).isPrivate ) &&
					this.fqn.packageName.equals(
							accessorClassInfo.fqn.packageName ) )
			{
				result.addAll(
						thisMethodInfosOfToTest );
			}

			if ( AsmClasspathInfo.isSameSourceFile(
					this ,
					accessorClassInfo ) )
				// in der gleichen Quell-Datei sind alle Felder sichtbar
			{
				result.addAll(
						thisMethodInfosOfToTest );
			}
		}

		// TODO Sichtbarkeit aufgrund gleicher Quell-Datei wird noch nicht beachtet

		if ( this.superType == null )
			// Sichtbarkeit aus Oberklasse
		{
			if ( ! this.fqn.equals( Fqn.JAVA_LANG_OBJECT_FQN ) )
				// avoid stack overflow
			{
				final List<? extends AsmMethodInfo> methodInfosFromSuperClass =
						asmClasspathInfo.getClassInfoMandantory( Fqn.JAVA_LANG_OBJECT_FQN )
						.getMethodsVisibleInSubClass(
								methodNameToGet ,
								this ,
								asmClasspathInfo );

				if ( methodInfosFromSuperClass != null )
				{
					result.addAll(
							methodInfosFromSuperClass );
				}
			}
		}
		else
		{
			final List<? extends AsmMethodInfo> methodInfosFromSuperClass =
					//this.superClassTuple.classInfoLazy.get()
					asmClasspathInfo.getClassInfoMandantory( this.superType.fqn )
					// TODO nicht korrekt implementiert
					.getMethodsVisibleInSubClass(
							methodNameToGet ,
							this ,
							asmClasspathInfo );

			if ( methodInfosFromSuperClass != null )
			{
				result.addAll(
						methodInfosFromSuperClass );
			}
		}

		// in Interfaces knnen auch Methoden stecken
		if ( interfaceTypes != null )
		{
			for ( final AsmClassType interfaceType : interfaceTypes )
			{
				final List<? extends AsmMethodInfo> methodInfosFromInterface =
						//interfaceTpl.classInfoLazy.get()
						asmClasspathInfo.getClassInfoMandantory( interfaceType.fqn )
						.getMethodsVisibleInSubClass(
								methodNameToGet ,
								this ,
								asmClasspathInfo );

				if ( methodInfosFromInterface != null )
				{
					result.addAll(
							methodInfosFromInterface );
				}
			}
		}

		return result;
	}

	// TODO namen vereinheitlichen: getMethodsVisibleInClass - getFieldVisibleForCaller
	public ImmutableArray<? extends AsmMethodInfo> getMethodsVisibleInClass(
			final String methodName ,
			final AsmClassInfo callerClass ,
			final AsmClasspathInfo asmClasspathInfo )
	{
		final ExtendedArrayList<AsmMethodInfo> result = new ExtendedArrayList<>();

		// TODO die writable-Eigenschaft (nicht final) wird nicht geprft, Unit-Test machen
		if ( this.methods.containsKey(
				methodName ) )
		{
			result.addAll(
					this.methods.get( methodName ) );
		}

		// TODO Sichtbarkeit aus Oberklasse oder aufgrund gleicher Quell-Datei wird noch nicht beachtet
		if ( this.superType == null )
		{
			if ( ! this.fqn.equals( Fqn.JAVA_LANG_OBJECT_FQN ) )
				// avoid stack overflow
			{
				final List<? extends AsmMethodInfo> methodInfosFromSuperClass =
						asmClasspathInfo.getClassInfoMandantory( Fqn.JAVA_LANG_OBJECT_FQN )
						// TODO nicht korrekt implementiert
						.getMethodsVisibleInSubClass(
								methodName ,
								this ,
								asmClasspathInfo );

				if ( methodInfosFromSuperClass != null )
				{
					result.addAll(
							methodInfosFromSuperClass );
				}
			}
		}
		else
		{
			final List<? extends AsmMethodInfo> methodInfosFromSuperClass =
					asmClasspathInfo.getClassInfoMandantory( this.superType.fqn )
					// TODO nicht korrekt implementiert
					.getMethodsVisibleInSubClass(
							methodName ,
							this ,
							asmClasspathInfo );

			if ( methodInfosFromSuperClass != null )
			{
				result.addAll(
						methodInfosFromSuperClass );
			}
		}

		// in Interfaces knnen auch Methoden stecken
		if ( interfaceTypes != null )
		{
			for ( final AsmClassType interfaceType : interfaceTypes )
			{
				final List<? extends AsmMethodInfo> methodInfosFromInterface =
						//interfaceTpl.classInfoLazy.get()
						asmClasspathInfo.getClassInfoMandantory( interfaceType.fqn )
						.getMethodsVisibleInSubClass(
								methodName ,
								this ,
								asmClasspathInfo );

				if ( methodInfosFromInterface != null )
				{
					result.addAll(
							methodInfosFromInterface );
				}
			}
		}

		return ImmutableArray.<AsmMethodInfo>create( result );
	}

	//public ImmutableArray<AsmMethodAndResolvedTypeargsTuple> getMethodsWithTypeargsVisibleInClass(
	//		final String methodName ,
	//		// Typ-Argumente der konkreten Instanz dieser Klasse
	//		final ImmutableMap<String, AsmType> typeargsNameToBoundMap ,
	//		final AsmClassInfo callerClass ,
	//		final AsmClasspathInfo asmClasspathInfo )
	//{
	//	final ExtendedArrayList<AsmMethodAndResolvedTypeargsTuple> result = new ExtendedArrayList<>();
	//
	//	// TODO die writable-Eigenschaft (nicht final) wird nicht geprft, Unit-Test machen
	//	if ( this.methods.containsKey(
	//			methodName ) )
	//	{
	//		//result.addAll( this.methods.get( methodName ) );
	//		final ImmutableArray<? extends AsmMethodInfo> methods = this.methods.get( methodName );
	//
	//		for ( final AsmMethodInfo method : methods )
	//		{
	//			result.add(
	//					new AsmMethodAndResolvedTypeargsTuple(
	//							method ,
	//							typeargsNameToBoundMap ) );
	//		}
	//	}
	//
	//	// TODO Sichtbarkeit aus Oberklasse oder aufgrund gleicher Quell-Datei wird noch nicht beachtet
	//	if ( this.superType == null )
	//	{
	//		if ( ! this.fqn.equals( Fqn.JAVA_LANG_OBJECT_FQN ) )
	//			// avoid stack overflow
	//		{
	//			final ImmutableArray<AsmMethodAndResolvedTypeargsTuple> methodTuplesFromSuperClass =
	//					asmClasspathInfo.getClassInfoMandantory( Fqn.JAVA_LANG_OBJECT_FQN )
	//					// TODO nicht korrekt implementiert
	//					.getMethodsWithTypeargsVisibleInClass(
	//							methodName ,
	//							// typeargsNameToBoundMap 
	//							//getBoundResolvedTypeParamMap( typeargsNameToBoundMap )
	//							// java.lang.Object nutzt keine Generics
	//							null ,
	//							// callerClass
	//							this ,
	//							asmClasspathInfo );
	//
	//			if ( methodTuplesFromSuperClass != null )
	//			{
	//				result.addAll(
	//						methodTuplesFromSuperClass );
	//			}
	//		}
	//	}
	//	else
	//	{
	//		final ImmutableArray<AsmMethodAndResolvedTypeargsTuple> methodTuplesFromSuperClass =
	//				asmClasspathInfo.getClassInfoMandantory( this.superType.fqn )
	//				// TODO nicht korrekt implementiert
	//				.getMethodsWithTypeargsVisibleInClass(
	//						methodName ,
	//						// typeargsNameToBoundMap 
	//						getBoundResolvedTypeParamMap( typeargsNameToBoundMap ) ,
	//						// callerClass
	//						this ,
	//						asmClasspathInfo );
	//
	//		if ( methodTuplesFromSuperClass != null )
	//		{
	//			result.addAll(
	//					methodTuplesFromSuperClass );
	//		}
	//	}
	//
	//	// in Interfaces knnen auch Methoden stecken
	//	if ( interfaceTypes != null )
	//	{
	//		for ( final AsmClassType interfaceType : interfaceTypes )
	//		{
	//			final ImmutableArray<AsmMethodAndResolvedTypeargsTuple> methodTuplesFromInterface =
	//					asmClasspathInfo.getClassInfoMandantory( interfaceType.fqn )
	//					.getMethodsWithTypeargsVisibleInClass(
	//							methodName ,
	//							// typeargsNameToBoundMap 
	//							getBoundResolvedTypeParamMap() ,
	//							// callerClass
	//							this ,
	//							asmClasspathInfo );
	//
	//			if ( methodTuplesFromInterface != null )
	//			{
	//				result.addAll(
	//						methodTuplesFromInterface );
	//			}
	//		}
	//	}
	//
	//	return ImmutableArray.<AsmMethodAndResolvedTypeargsTuple>create( result );
	//}

	public ImmutableArray<AsmBoundResolvedMethodAndOriginalMethodTuple> getBoundresolvedMethodsVisibleInClass(
			final AsmClassInfo inheritClassInfo ,
			final ImmutableMap<String, ? extends AsmType> inheritanceTypeArgs ,
			final String methodName ,
			// Typ-Argumente der konkreten Instanz dieser Klasse
			//final ImmutableMap<String, ? extends AsmType> typeargsNameToBoundMap ,
			final AsmClassInfo callerClass ,
			final AsmClasspathInfo asmClasspathInfo )
	{
		final ExtendedArrayList<AsmBoundResolvedMethodAndOriginalMethodTuple> result = new ExtendedArrayList<>();

		// TODO die writable-Eigenschaft (nicht final) wird nicht geprft, Unit-Test machen
		if ( this.methods.containsKey(
				methodName ) )
		{
			//result.addAll( this.methods.get( methodName ) );
			final ImmutableArray<? extends AsmMethodInfo> methods = this.methods.get( methodName );

			for ( final AsmMethodInfo method : methods )
			{
				result.add(
						new AsmBoundResolvedMethodAndOriginalMethodTuple(
								method.resolveBounds(
										//typeargsNameToBoundMap
										//this.getBoundResolvedTypeParamMap( typeargsNameToBoundMap )
										inheritClassInfo ,
										inheritanceTypeArgs ,
										asmClasspathInfo ) ,
								method ) );
			}
		}

		// TODO Sichtbarkeit aus Oberklasse oder aufgrund gleicher Quell-Datei wird noch nicht beachtet
		if ( this.superType == null )
		{
			if ( ! this.fqn.equals( Fqn.JAVA_LANG_OBJECT_FQN ) )
				// avoid stack overflow
			{
				final ImmutableArray<AsmBoundResolvedMethodAndOriginalMethodTuple> methodTuplesFromSuperClass =
						asmClasspathInfo.getClassInfoMandantory( Fqn.JAVA_LANG_OBJECT_FQN )
						.getBoundresolvedMethodsVisibleInClass(
								inheritClassInfo ,
								// java.lang.Object nutzt keine Generics
								null ,
								methodName ,
								// typeargsNameToBoundMap 
								//getBoundResolvedTypeParamMap( typeargsNameToBoundMap )
								// java.lang.Object nutzt keine Generics
								//null ,
								// callerClass
								this ,
								asmClasspathInfo );

				if ( methodTuplesFromSuperClass != null )
				{
					result.addAll(
							methodTuplesFromSuperClass );
				}
			}
		}
		else
		{
			final AsmClassInfo superAsmClassInfo = asmClasspathInfo.getClassInfoMandantory( this.superType.fqn );

			final ImmutableArray<AsmBoundResolvedMethodAndOriginalMethodTuple> methodTuplesFromSuperClass =
					superAsmClassInfo
					.getBoundresolvedMethodsVisibleInClass(
							inheritClassInfo ,
							//inheritanceTypeArgs
							//this.superType.resolveBound( inheritanceTypeArgs ).getEntireTypeArgs( asmClasspathInfo ) ,
							inheritanceTypeArgs ,
							methodName ,
							// typeargsNameToBoundMap 
							//superAsmClassInfo.getBoundResolvedTypeParamMap(
							//		// TODO uebergebene Typ-Argumente: typeargsNameToBoundMap
							//		//this.superType.getEntireTypeArgs( asmClasspathInfo )
							//		//superAsmClassInfo.getBoundResolvedTypeParamMap(
							//				this.getBoundResolvedTypeParamMap(
							//						typeargsNameToBoundMap ) )
							// callerClass
							this ,
							asmClasspathInfo );

			if ( methodTuplesFromSuperClass != null )
			{
				result.addAll(
						methodTuplesFromSuperClass );
			}
		}

		// in Interfaces knnen auch Methoden stecken
		if ( interfaceTypes != null )
		{
			for ( final AsmClassType interfaceType : interfaceTypes )
			{
				final AsmClassInfo interfaceAsmClassInfo = asmClasspathInfo.getClassInfoMandantory( interfaceType.fqn );

				final ImmutableArray<AsmBoundResolvedMethodAndOriginalMethodTuple> methodTuplesFromInterface =
						interfaceAsmClassInfo.getBoundresolvedMethodsVisibleInClass(
								inheritClassInfo ,
								inheritanceTypeArgs ,
								methodName ,
								// typeargsNameToBoundMap 
								//interfaceAsmClassInfo.getBoundResolvedTypeParamMap(
								//		// TODO uebergebene Typ-Argumente: typeargsNameToBoundMap
								//		interfaceType.getEntireTypeArgs(
								//				asmClasspathInfo ) ) ,
								// callerClass
								this ,
								asmClasspathInfo );

				if ( methodTuplesFromInterface != null )
				{
					result.addAll(
							methodTuplesFromInterface );
				}
			}
		}

		// TODO 10.08.2016 ich glaube folgender Block muss, weg methoden aus der gleichen Quell-Datei koennen zwar sichtbar sein, aber ein Objekt der Klasse mus im Scope sein oder die Methode ist static
		// Sichtbarkeit aufgrund gleicher Quell-Datei
		final ImmutableArray<AsmClassInfo> classesInSameSourceFile =
				getOtherClassesInSameSourceFile(
						asmClasspathInfo );

		for ( final AsmClassInfo classInSameSourceFile : classesInSameSourceFile )
		{
			if ( classInSameSourceFile.methods != null )
			{
				final ImmutableArray<AsmBoundResolvedMethodAndOriginalMethodTuple> methodTuplesForSameSourceFile =
						classInSameSourceFile.getBoundresolvedMethods(
								inheritClassInfo ,
								inheritanceTypeArgs ,
								methodName ,
								// typeargsNameToBoundMap 
								//classInSameSourceFile.getBoundResolvedTypeParamMap( typeargsNameToBoundMap ) ,
								null ,
								asmClasspathInfo );

				if ( methodTuplesForSameSourceFile != null )
				{
					//result.addAll( methodTuplesForSameSourceFile );
					for ( final AsmBoundResolvedMethodAndOriginalMethodTuple methodTupleForSameSourceFile : methodTuplesForSameSourceFile )
					{
						if ( methodTupleForSameSourceFile.boundResolvedMethod.isStatic )
							// ich vermute, dies gilt nur bei statischen Methoden
						{
							result.add( methodTupleForSameSourceFile );
						}
					}
				}
			}
		}

		return ImmutableArray.create( result );
	}

	private ImmutableArray<AsmBoundResolvedMethodAndOriginalMethodTuple> getBoundresolvedMethods(
			final AsmClassInfo inheritClassInfo ,
			final ImmutableMap<String, ? extends AsmType> inheritanceTypeArgs ,
			final String methodName ,
			// Typ-Argumente der konkreten Instanz dieser Klasse
			final ImmutableMap<String, ? extends AsmType> typeargsNameToBoundMap ,
			final AsmClasspathInfo asmClasspathInfo )
	{
		final ExtendedArrayList<AsmBoundResolvedMethodAndOriginalMethodTuple> result = new ExtendedArrayList<>();

		// TODO die writable-Eigenschaft (nicht final) wird nicht geprft, Unit-Test machen
		if ( this.methods.containsKey(
				methodName ) )
		{
			//result.addAll( this.methods.get( methodName ) );
			final ImmutableArray<? extends AsmMethodInfo> methods = this.methods.get( methodName );

			for ( final AsmMethodInfo method : methods )
			{
				result.add(
						new AsmBoundResolvedMethodAndOriginalMethodTuple(
								method.resolveBounds(
										//typeargsNameToBoundMap
										//getBoundResolvedTypeParamMap( typeargsNameToBoundMap )
										inheritClassInfo ,
										inheritanceTypeArgs ,
										asmClasspathInfo ) ,
								method ) );
			}
		}

		return ImmutableArray.create( result );
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public String toString()
	{
		final String typeParamStr;
		if ( this.typeParams != null && ( ! this.typeParams.isEmpty() ) )
		{
			final SeparatedStrBuff buff = new SeparatedStrBuff( ", " );

			for ( final AsmAbstractTypeParam typeParam : this.typeParams )
			{
				buff.add( String.valueOf( typeParam ) );
			}

			typeParamStr = "<" + buff + ">";
		}
		else
		{
			typeParamStr = "";
		}

		return this.fqn.toFqnStr() + typeParamStr;
	}

	public ImmutableArray<AsmClassInfo> getOtherClassesInSameSourceFile(
			final AsmClasspathInfo asmClasspathInfo )
	{
		final ImmutableArray<AsmClassInfo> allClassesForSourceFile =
				asmClasspathInfo.getAllClassesForSourceFile(
						this.fqn.packageName ,
						sourceFileName );

		final List<AsmClassInfo> result = new ArrayList<>();

		if ( allClassesForSourceFile != null )
		{
			for ( final AsmClassInfo asmClassForSourceFile : allClassesForSourceFile )
				// filter not this
			{
				if ( ! asmClassForSourceFile.equals( this ) )
				{
					result.add( asmClassForSourceFile );
				}
			}
		}

		return ImmutableArray.create( result );
	}

	public boolean isAssignableFrom(
			//final String fqnStrOfClassToTest ,
			final Fqn fqnOfClassToTest ,
			final AsmClasspathInfo asmClasspathInfo )
	{
		//if ( this.fqn.toFqnStr().equals( fqnStrOfClassToTest ) )
		if ( this.fqn.equals( fqnOfClassToTest ) )
		{
			return true;
		}

		//if ( this.fqn.toFqnStr().equals( "java.lang.Object" ) )
		if ( this.fqn.equals( Fqn.JAVA_LANG_OBJECT_FQN ) )
		{
			return true;
		}

		final AsmClassInfo classInfoToTest =
				asmClasspathInfo.getClassInfoMandantory(
						fqnOfClassToTest );

		return 
				classInfoToTest.isSuperClassOrImplementedInterface(
						this.fqn/*.toFqnStr()*/ ,
						asmClasspathInfo );
	}

	public boolean isSubClassOf(
			final Fqn fqnToTest ,
			final AsmClasspathInfo asmClasspathInfo )
	{
		return asmClasspathInfo.getClassInfoMandantory( fqnToTest ).isSuperClassOrImplementedInterface(fqn, asmClasspathInfo);
	}

	public boolean isSuperClassOrImplementedInterface(
			final Fqn fqnToTest ,
			final AsmClasspathInfo asmClasspathInfo )
	{
		if ( superType != null )
		{
			if ( superType.fqn.equals( fqnToTest ) )
			{
				return true;
			}

			if ( asmClasspathInfo.getClassInfoMandantory( this.superType.fqn )
					.isSuperClassOrImplementedInterface(
							fqnToTest ,
							asmClasspathInfo ) )
			{
				return true;
			}
		}

		if ( interfaceTypes != null )
		{
			for ( final AsmClassType interfaceType : interfaceTypes )
			{
				if ( interfaceType.fqn.equals( fqnToTest ) )
				{
					return true;
				}

				if ( asmClasspathInfo.getClassInfoMandantory( interfaceType.fqn )
						.isSuperClassOrImplementedInterface(
								fqnToTest ,
								asmClasspathInfo ) )
				{
					return true;
				}
			}
		}

		return false;
	}

	public Map<String, ? extends AsmAbstractTypeParam> getTypeParamMap()
	{
		if ( this.typeParams == null )
		{
			return Collections.<String, AsmAbstractTypeParam>emptyMap();
		}

		final Map<String, AsmExtendsTypeParam> result = new HashMap<>();

		for ( final AsmAbstractTypeParam typeParam : this.typeParams )
		{
			if ( typeParam instanceof AsmExtendsTypeParam )
			{
				result.put(
						typeParam.name ,
						(AsmExtendsTypeParam) typeParam );
			}
		}

		return result;
	}

	/**
	 * @return die Bounds der Typ-Parameter
	 */
	public ImmutableMap<String, AsmType> getBoundResolvedTypeParamMap()
	{
		if ( this.typeParams == null )
		{
			//return Collections.<String, AsmAbstractTypeParam>emptyMap();
			return null;
		}

		final Map<String, AsmType> result = new HashMap<>();

		for ( final AsmAbstractTypeParam typeParam : this.typeParams )
		{
			if ( typeParam instanceof AsmExtendsTypeParam )
			{
				result.put(
						typeParam.name ,
						( (AsmExtendsTypeParam) typeParam ).boundType.resolveBound(
								//typeargsNameToBoundMap
								null ) );
			}
		}

		return ImmutableMap.create( result );
	}

	/**
	 * @param strongerTypeParamBounds staerkere Bounds, die durch Vererbung oder Instanziierung festgelegt wurden
	 * @return die Bounds der Typ-Parameter
	 */
	public ImmutableMap<String, ? extends AsmType> getBoundResolvedTypeParamMap(
			final ImmutableMap<String, ? extends AsmType> strongerTypeParamBounds )
	{
		if ( this.typeParams == null )
		{
			//return Collections.<String, AsmAbstractTypeParam>emptyMap();
			//return null;
			return strongerTypeParamBounds;
		}

		final Map<String, AsmType> result = new HashMap<>();

		for ( final AsmAbstractTypeParam typeParam : this.typeParams )
		{
			if ( strongerTypeParamBounds != null &&
					strongerTypeParamBounds.containsKey( typeParam.name ) )
			{
				result.put(
						typeParam.name ,
						strongerTypeParamBounds.get( typeParam.name ) );
			}
			else if ( typeParam instanceof AsmExtendsTypeParam )
			{
				result.put(
						typeParam.name ,
						( (AsmExtendsTypeParam) typeParam ).boundType );
			}
		}

		return ImmutableMap.create( result );
	}

	public AsmClassType getThisType()
	{
		return new AsmClassType(
				fqn ,
				TypeArgsUtil.convertTypeParamsToTypeArgs( typeParams ) );
	}

	//public AsmClassType getSuperType()
	//{
	//	if ( this.superType == null )
	//	{
	//		return null;
	//	}
	//
	//	return new AsmClassType(
	//			this.superType.fqn ,
	//			// TODO
	//			this.superType.getEntireTypeArgs().get( 0 ) );
	//}

	public AsmType getBoundForTypeParamName(
			final String typeParamName ,
			final ImmutableArray<? extends AsmType> typeArgs ,
			final AsmClasspathInfo asmClasspathInfo )
	{
		if ( ! CollectionUtil.isNullOrEmpty( this.typeParams ) )
		{
			//for ( final AsmAbstractTypeParam typeParam : this.typeParams )
			for ( int index = 0 ; index < this.typeParams.size() ; index++ )
			{
				final AsmAbstractTypeParam typeParam = this.typeParams.get( index );

				if ( typeParam.name.equals( typeParamName ) )
				{
					if ( typeArgs != null && typeArgs.size() > index )
					{
						return typeArgs.get( index );
					}
					if ( typeParam instanceof AsmExtendsTypeParam )
					{
						return ((AsmExtendsTypeParam) typeParam).boundType;
					}
					return AsmClassType.JAVA_LANG_OBJECT;
				}
			}
		}

		if ( this.superType != null )
		{
			final AsmType typeParamFromSuperClass =
					asmClasspathInfo.getClassInfoMandantory( this.superType.fqn )
					.getBoundForTypeParamName(
							typeParamName ,
							this.superType.typeArgs ,
							asmClasspathInfo );

			if ( typeParamFromSuperClass != null )
			{
				return typeParamFromSuperClass;
			}
		}

		if ( interfaceTypes != null )
		{
			for ( final AsmClassType interfaceType : interfaceTypes )
			{
				final AsmType typeParamFromInterface =
						asmClasspathInfo.getClassInfoMandantory( interfaceType.fqn )
						.getBoundForTypeParamName(
								typeParamName ,
								interfaceType.typeArgs ,
								asmClasspathInfo );

				if ( typeParamFromInterface != null )
				{
					return typeParamFromInterface;
				}
			}
		}

		return null;
	}

	public AsmFieldInfo getOuterClassInstanceField(
			final AsmClasspathInfo asmClasspathInfo )
	{
		if ( this.isStatic || this.isTopLevel() )
		{
			return null;
		}
		
		AsmFieldInfo asmField = null;
		int fieldIndex = 0;
		do
		{
				asmField =
						this.getFieldVisibleInOwnMethod(
								// reservierter Feld-Name fuer outer
								"this$" + fieldIndex ,
								asmClasspathInfo );

			fieldIndex++;

			if ( fieldIndex > 1024 )
			{
				//throw new RuntimeException();
				return null;
			}
		}
		while ( asmField == null ||
				( ! asmField.isSynthetic ) 
				//|| ( ! ( (ClassType) staticRefType ).entireFqn( asmClasspathInfo ).equals( asmField.declaringClassFqn ) )
				);
		
		return asmField;
	}
	
}