package de.heinerkuecker.asm_improved;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.jar.JarInputStream;

import org.objectweb.asm.ClassReader;

import de.heinerkuecker.asm_improved.asm_dom.AsmDomClass;
import de.heinerkuecker.asm_improved.asm_dom.AsmDomClassVisitor;
import de.heinerkuecker.asm_improved.util.DottedName;
import de.heinerkuecker.asm_improved.util.Fqn;
import de.heinerkuecker.asm_improved.util.GetTypesafeMap;
import de.heinerkuecker.asm_improved.util.ImmutableArray;
import de.heinerkuecker.asm_improved.util.PackageName;
import de.heinerkuecker.asm_improved.util.SourceFileNameWithoutPath;

/**
 * Klasse zum Kapseln der
 * <a href="http://asm.ow2.org">ASM</a>-Informationen
 * aller *.class-Dateien aus dem Classpath.
 * 
 * @author Heiner K&uuml;cker
 */
public final class AsmClasspathInfo
{
    /**
     * Kriterium f&uuml;r von der Analyse auszuschliessende Klassen.
     */
    private static final ImmutableArray<? extends String> EXCLUDE_NAME_ENDSWITH_STR_ARR = ImmutableArray.<String>create( /*"package-info"*/ );

    /**
     * Betriebsart(Modus), sofort Scannen oder bei Bedarf Scannen.
     */
    public enum Mode
    {
    	/**
    	 * Scannen der Informationen aus den *.class-Dateien bei Bedarf (on demand).
    	 */
    	LAZY ,

    	/**
    	 * Sofortiges Scannen der Informationen aus den *.class-Dateien.
    	 */
    	EAGER;
    }

    /**
     * sofort oder bei Bedarf scannen.
     */
    public final Mode mode;

    /**
     * Ob das Scannen des Classpath im Modus {@link Mode#EAGER} fertig ist,
     * wenn nicht, kann beim rekursiven Zugriff eine gewnschte Klasse
     * noch nicht gescannt worden sein.
     */
    private enum ScanState
    {
    	WORK ,

    	READY;
    }
    
    private ScanState state;

    /**
     * Jar-Files (oder andere Archive)
     * und Verzeichnisse des Classpath.
     */
    public final /*ImmutableArray<? extends File>*/ClasspathFiles classpathFiles;
    
    /**
     * Packages
     */
    private final GetTypesafeMap<String, AsmPackageInfo> packageMap = GetTypesafeMap.create( new HashMap<String, AsmPackageInfo>() );

	/**
	 * Konstruktor mit oder ohne sofortigem Start
	 * des Scannens der zu scannenden
	 * *.class- bzw *.jar-Dateien.
	 * 
	 * @param mode see {@link #mode}
	 * @param classpathFileArr Array mit *.jar-Dateien, *.class-Dateien oder Verzeichnissen,
	 * in denen *.class-Dateien erwartet werden
	 */
	public AsmClasspathInfo(
			final Mode mode ,
			final File... classpathFileArr )
	{
		if ( mode == null )
		{
			throw new IllegalArgumentException(
					"mode is null" );
		}

		this.mode = mode;
		this.classpathFiles =
				//ImmutableArray.create( classpathFileArr )
				new ClasspathFiles( classpathFileArr );

		if ( mode == Mode.EAGER )
        {
			state = ScanState.WORK;

	        for ( final File classpathFile : classpathFileArr )
	        {
	            if ( classpathFile.isDirectory() )
	            {
	                scanBinDirRecursive(
	                		classpathFile );
	            }
	            else if ( classpathFile.getName().endsWith( ".jar" ) )
	            {
	        		scanJarFile(
	        				classpathFile );
	            }
	            else if ( classpathFile.getName().endsWith( ".class" ) )
	            {
	            	scanClassFile(
	            			classpathFile );
	            }
	            else
	            {
	            	throw new IllegalArgumentException(
	            			"jar file or class file or directory with class files expected: " +
	            			classpathFile );
	            }
	        }

			state = ScanState.READY;
        }
	}

	public boolean containsPackage(
			final String packageName )
	{
		if ( this.mode != Mode.EAGER || this.state != ScanState.READY )
		{
			throw new IllegalStateException(
					"class path scan not complete" );
		}

		final AsmPackageInfo classPackage =
			packageMap.get(
					packageName );

		return classPackage != null;
	}

	public boolean containsClassInfo(
			final String fullQualifiedClassName )
	{
		final AsmClassInfo testClassInfo =
				getClassInfoOptional(
						fullQualifiedClassName );

		return testClassInfo != null;
	}

	public boolean containsClassInfo(
			final Fqn fqn )
	{
		final AsmClassInfo classInfoOptional = getClassInfoOptional( fqn );

		return classInfoOptional != null;
	}

	/**
	 * 
	 * @param packageName
	 * @param simpleClassName
	 * @return
	 * @deprecated Strings sind fr package und Klassen-Name nicht eindeutig
	 */
	@Deprecated
	public AsmClassInfo getClassInfoMandantory(
			final String packageName ,
			final String simpleClassName )
	{
		final AsmPackageInfo classPackage =
			packageMap.get(
					packageName );

		if ( classPackage == null )
		{
			throw new IllegalArgumentException(
					"no classPackage for packageName " +
					packageName );
		}

		final AsmClassInfo classInfo =
				classPackage.classMap.get(
				simpleClassName );

		if ( classInfo == null )
		{
			throw new IllegalArgumentException(
					"no classInfo for " +
					packageName + ":" +
					simpleClassName );
		}

		return classInfo;
	}

	public AsmClassInfo getClassInfoMandantory(
			final String fullQualifiedClassName )
	{
		final AsmClassInfo classInfoOptional =
				getClassInfoOptional(
						fullQualifiedClassName );

		if ( classInfoOptional == null )
			// doppelte Absicherung, eigentlich drfte dies nicht vorkommen
		{
			throw new IllegalArgumentException(
					"no classInfo for " +
					fullQualifiedClassName );
		}

		return classInfoOptional;
	}

	public AsmClassInfo getClassInfoMandantory(
			final Fqn fqn )
	{
		final AsmClassInfo classInfo =
				getClassInfoOptional(
						fqn );

		if ( classInfo == null )
		{
			throw new IllegalArgumentException(
					"no ClassInfo for " +
					fqn );
		}

		return classInfo;
	}

	/**
	 * 
	 * @param fqn full qualified name
	 * @return ClassInfo or <code>null</code>
	 */
	public AsmClassInfo getClassInfoOptional(
			final Fqn fqn )
	{
		if ( mode == Mode.EAGER && ! ( state == ScanState.READY ) )
		{
			throw new IllegalStateException(
					"class path scan not complete" );
		}
		
		final AsmPackageInfo classPackage =
				packageMap.get(
						fqn.packageName.value );

		if ( classPackage == null )
		{
			if ( mode == Mode.EAGER )
			{
				return null;
			}

			return relaodAndGetClassInfoOptional( fqn );
		}

		final AsmClassInfo classInfo =
				classPackage.classMap.get(
						fqn.dottedName.toString() );

		if ( classInfo == null )
		{
			if ( mode == Mode.EAGER )
			{
				return null;
			}

			return relaodAndGetClassInfoOptional( fqn );
		}

		return classInfo;
	}

	/**
	 * @param fqn full qualified name
	 * @return reloaded ClassInfo
	 */
	private AsmClassInfo relaodAndGetClassInfoOptional(
			final Fqn fqn )
    {
		for ( final File classpathFile : classpathFiles.fileArr )
		{
			if ( classpathFile.isDirectory() )
			{
				final File classFile =
						new File(
								classpathFile.getPath() +
								File.separator +
								fqn.toClassFileFqnStr() +
								".class" );

				if ( classFile.exists() )
				{
					return scanClassFile(
							classFile );
					//break;
					//return getClassInfoOptional( fqn );
				}
			}
			else if ( classpathFile.isFile() )
			{
				try ( JarFile jarFile = new JarFile( classpathFile ) )
				{
					final String jarEntryName =
							fqn.toClassFileFqnStr() +
							".class";

					final JarEntry jarEntry =
							jarFile.getJarEntry(
									jarEntryName );

					if ( jarEntry != null )
					{
						try ( InputStream jarInputStream = jarFile.getInputStream( jarEntry ) )
						{
							return scanClassContentStream(
									classpathFile ,
									jarInputStream
									//isRtJar
									);
						}

						//break;
						//return getClassInfoOptional( fqn );
					}
				}
				catch ( IOException iox )
				{
					System.err.println( iox.getMessage() );
					iox.printStackTrace();
					throw new RuntimeException( iox );
				}
			}
			else
			{
				throw new IllegalStateException(
						String.valueOf( classpathFile ) );
			}
		}
		return null;
    }

	/**
	 * 
	 * @param fullQualifiedClassName full qualified name
	 * @return ClassInfo or <code>null</code>
	 */
	public AsmClassInfo getClassInfoOptional(
			final String fullQualifiedClassName )
	{
		if ( fullQualifiedClassName.contains( "." ) )
		{
			final List<? extends String> noDotFullQualifiedClassNameParts =
					Fqn.createNoDotFullQualifiedClassNameParts(
							fullQualifiedClassName );

			return getClassInfoOptional(
					noDotFullQualifiedClassNameParts );
		}
		else
		{
			return getClassInfoOptional(
					new Fqn(
							new PackageName( "" ) ,
							DottedName.valueOfDotSeparatedStr( fullQualifiedClassName ) ) );
		}
	}

	/**
	 * @param noDotFullQualifiedClassNameParts
	 * @return ClassInfo or <code>null</code>
	 */
    public AsmClassInfo getClassInfoOptional(
            final List<? extends String> noDotFullQualifiedClassNameParts )
    {
	    for ( int packageNameSimpleOrInnerOrNestedClassNameSeparatorIndex = 0 ;
	    		packageNameSimpleOrInnerOrNestedClassNameSeparatorIndex < noDotFullQualifiedClassNameParts.size() ;
	    		packageNameSimpleOrInnerOrNestedClassNameSeparatorIndex++ )
	    	// weil der vollqualifizierte Name aus einem Package und einem Punkt-getrenntem Namen einer inneren/geschachtelten Klasse/Interface bestehen kann, alle Kombinationen ausprobieren
	    {
	    	final String tmpPackageName =
	    			Fqn.joinWithDotSeparatorChar(
	    					noDotFullQualifiedClassNameParts.subList(
	    							//fromIndex
	    							0 ,
	    							//toIndex
	    							packageNameSimpleOrInnerOrNestedClassNameSeparatorIndex ) );

	    	final String tmpClassName =
	    			Fqn.joinWithDotSeparatorChar(
	    					noDotFullQualifiedClassNameParts.subList(
	    							//fromIndex
	    							packageNameSimpleOrInnerOrNestedClassNameSeparatorIndex ,
	    							//toIndex
	    							noDotFullQualifiedClassNameParts.size() ) );

	    	final Fqn tmpFqn =
	    			new Fqn(
	    					new PackageName( tmpPackageName ) ,
	    					DottedName.valueOfDotSeparatedStr( tmpClassName ) );

    		final AsmClassInfo classInfoOptional =
    				getClassInfoOptional(
    						tmpFqn );

    		if ( classInfoOptional != null )
	    	{
	    		return classInfoOptional;
	    	}
	    }
		return null;
    }

    /**
     * @param topLevelFqn {@link Fqn} of specified top level class
     * @return all inner or nested classes of specified top level class
     */
    public ImmutableArray<AsmClassInfo> getAllInnerOrNestedClassesFor(
    		final Fqn topLevelFqn )
    {
		if ( this.mode != Mode.EAGER || this.state != ScanState.READY )
		{
			throw new IllegalStateException(
					"class path scan not complete" );
		}

		//if ( ! getClassInfoMandantory( topLevelFqn ).isTopLevel() )
		//{
		//	throw new IllegalArgumentException( "no top level class " + topLevelFqn );
		//}

    	final List<AsmClassInfo> result = new ArrayList<>();

    	//System.out.println( topLevelFqn );
    	
    	final AsmPackageInfo classPackage =
    			packageMap.get(
				( topLevelFqn.packageName != null
					? topLevelFqn.packageName.value
					: "" ) );

		if ( classPackage == null )
		{
			throw new IllegalArgumentException(
					"no classPackage for class " +
							topLevelFqn );
		}

    	final Iterable<? extends AsmClassInfo> classInfosInSamePackage =
    			classPackage.classMap.values();

    	for ( final AsmClassInfo classInfoInSamePackage : classInfosInSamePackage )
    	{
			if ( ! classInfoInSamePackage.fqn.dottedName.isTopLevel() &&
					classInfoInSamePackage.fqn.dottedName.parts.get( 0 ).equals(
							topLevelFqn.dottedName.parts.get( 0 ) ) )
			{
				result.add(
						classInfoInSamePackage );
			}
		}
    	
    	return ImmutableArray.create( result );
    }

    public ImmutableArray<AsmClassInfo> getAllClassesForSourceFile(
    		final PackageName packageName ,
    		final String sourceFileName )
    {
    	if ( this.mode != Mode.EAGER || this.state != ScanState.READY )
    	{
    		throw new IllegalStateException(
    				"class path scan not complete" );
    	}

    	final List<AsmClassInfo> result = new ArrayList<>();

    	final Iterable<? extends AsmClassInfo> classInfosInSamePackage =
    			packageMap.get(
    					packageName.value ).classMap.values();

    	if ( classInfosInSamePackage != null )
    	{
    		for ( final AsmClassInfo classInfoInSamePackage : classInfosInSamePackage )
    		{
    			if ( classInfoInSamePackage != null &&
    					classInfoInSamePackage.sourceFileName != null &&
    					classInfoInSamePackage.sourceFileName.equals(
    					sourceFileName ) )
    			{
    				result.add(
    						classInfoInSamePackage );
    			}
    		}
    	}

    	if ( result.isEmpty() )
    	{
    		return null;
    	}
    	
    	return ImmutableArray.create( result );
    }

	private void scanJarFile(
			final File jarFile )
	{
		//final boolean isRtJar = jarFile.getName().equals( "rt.jar" );
		String jarEntryName = null;

		try ( InputStream jarFileInStream =
				new FileInputStream(
						jarFile );
				JarInputStream jarInputStream =
						new JarInputStream(
								jarFileInStream ) )
		{
			JarEntry jarEntry;
			while ( ( jarEntry = jarInputStream.getNextJarEntry() ) != null )
			{
				//System.out.println( jarEntry.getName() );
				//if ( ( ! isRtJar ) ||
				//		jarEntry.getName().startsWith( "java/lang/" ) ||
				//		jarEntry.getName().startsWith( "java/util/" ) ||
				//		jarEntry.getName().startsWith( "java/io/" ) ||
				//		jarEntry.getName().startsWith( "java/net/" ) )
				//	// TODO wegen Dauer Test aus rt.jar nur die Pakete java.lang.*, java.util.* und java.io.* auslesen
				//	// TODO mal ordentliche Lsung machen
                {
	                if ( jarEntry.getName().endsWith( ".class" ) &&
	                		( ! hasExcludeEndswithStr( jarEntry.getName() ) ) )
	                {
	    				//System.out.println( "jarEntry: " + jarEntry.getName() );
	                	jarEntryName = jarEntry.getName();

	                	scanClassContentStream(
	                			jarFile ,
	                			jarInputStream
	                			//isRtJar
	                			);
	                }
                }
			}
		}
		catch ( final Exception iox )
		{
			//System.err.println( iox.getMessage() );
			//iox.printStackTrace();
			throw new RuntimeException(
					"error scanning " +
					jarEntryName +
					" in jar file " +
					jarFile ,
					iox );
		}
	}

    private AsmClassInfo scanClassFile(
            final File classFile )
    {
    	//System.out.println( new Exception().getStackTrace()[ 0 ] + " " + classFile );

    	try  ( InputStream classInputStream = new FileInputStream( classFile ) )
    	{
    		return scanClassContentStream(
    				classFile ,
    				classInputStream
    				//isRtJar: false
    				);
    	}
    	catch ( Exception e )
    	{
    		System.err.println(e.getMessage());
    		e.printStackTrace();

    		throw new RuntimeException(
    				"error scanning " + classFile ,
    				e );
    	}
    }

    private static boolean hasExcludeEndswithStr(
            final String pFilename )
    {
    	for ( String excludeNameEndswithStr : EXCLUDE_NAME_ENDSWITH_STR_ARR )
    	{
    		if ( pFilename.endsWith( excludeNameEndswithStr + ".class" ) )
    		{
    			return true;
    		}
    		//if ( pFilename.contains( "$" ) )
    		//	// es handelt sich um eine innere Klasse
    		//{
    		//	final String lFileNameToDollarSign =
    		//			pFilename.substring(
    		//					0 ,
    		//					pFilename.lastIndexOf( '$' ) );
    		//
    		//	if ( lFileNameToDollarSign.endsWith( excludeNameEndswithStr ) )
    		//	{
    		//		return true;
    		//	}
    		//}
    	}
    	return false;
    }

    private void scanBinDirRecursive(
        final File binDir )
    {
        final File[] files = binDir.listFiles();

        for ( final File file : files )
        {
            if ( file.isDirectory() )
            {
                scanBinDirRecursive(
                		file );
            }
            else if ( file.getName().endsWith( ".class" ) &&
                ( ! hasExcludeEndswithStr( file.getName() ) ) )
            {
                scanClassFile(
                		file );
            }
        }
    }

	private AsmClassInfo scanClassContentStream(
			final File classpathFile ,
			final InputStream classInputStream )
	throws IOException
	{
		final AsmDomClassVisitor classVisitor = new AsmDomClassVisitor();
		
		final ClassReader classReader =
				new ClassReader(
						classInputStream );

		classReader.accept(
				classVisitor ,
				0 );

		final AsmDomClass domClass = classVisitor.domClass;

		final AsmClassInfo classInfo =
				AsmDomConverter.convertClass(
						domClass ,
						classpathFile ,
						this.classpathFiles );

		final Fqn classFqn = classInfo.fqn;

		AsmPackageInfo classPackage;
		if ( ( classPackage = packageMap.get( classFqn.packageName.toString() ) ) == null )
			// Package anlegen, wenn nicht vorhanden
		{
			classPackage =
					new AsmPackageInfo(
							classFqn.packageName );

			packageMap.put(
					classFqn.packageName.toString() ,
					classPackage );
		}

		if ( false && // TODO schalter, ob ambigious class ignoriert werden kann
				classPackage.classMap.containsKey( /*simpleClassName*/classFqn.dottedName.toString() ) )
			// Fehlermeldung, wenn eine Klasse zweimal/mehrmals im Classpath enthalten ist
		{
			if ( ! classPackage.classMap.get( /*simpleClassName*/classFqn.dottedName.toString() ).file.equals( classpathFile ) )
				// Mglicherweise wurde eine innere/geschachtelte Klasse mit ihrem Dollar-Zeichen-getrenntem Namen gesucht.
				// In diesem Fall keine Exception werfen
			{
				throw new IllegalStateException(
						"ambigious class " +
								///*packageName*/classVisitor.fqn.packageName + "." +
								///*simpleClassName*/classVisitor.fqn.dottedName
								classFqn );
			}
		}
		else
		{
			classPackage.classMap.put(
					classFqn.dottedName.toString() ,
					classInfo );
		}

		return classInfo;
	}

	static boolean isSameSourceFile(
			final AsmClassInfo classInfo1 ,
			final AsmClassInfo classInfo2 )
	{
		if ( 
				// nur im gleichen Package gltig, sonst kann der Name der Java-Datei in einem Classpath in unterschiedlichen Packages mehrmals auftauchen
				classInfo1.fqn.packageName.equals( classInfo2.fqn.packageName ) &&
				classInfo1.sourceFileName != null &&
				classInfo2.sourceFileName != null &&
				classInfo1.sourceFileName.equals(
						classInfo2.sourceFileName ) )
		{
			return true;
		}
		return false;
	}

	static boolean isSameSourceFile(
			final AsmClassInfo classInfo1 ,
	        final PackageName otherPackageName ,
			final SourceFileNameWithoutPath otherSourceFileNameWithoutPath )
	{
		if ( 
				// nur im gleichen Package gltig, sonst kann der Name der Java-Datei in einem Classpath in unterschiedlichen Packages mehrmals auftauchen
				( ( classInfo1.fqn.packageName == null && otherPackageName == null ) ||
						classInfo1.fqn.packageName.equals( otherPackageName ) ) &&
				classInfo1.sourceFileName != null &&
				otherSourceFileNameWithoutPath != null &&
				classInfo1.sourceFileName.equals(
						otherSourceFileNameWithoutPath.value ) )
		{
			return true;
		}
		return false;
	}

	/**
	 * @return
	 */
	public Iterable<? extends AsmClassInfo> getAllClassInfos()
	{
		if ( this.mode != Mode.EAGER || this.state != ScanState.READY )
		{
			throw new IllegalStateException(
					"class path scan not complete" );
		}

		final Iterable<? extends AsmPackageInfo> packages = this.packageMap.values();

		return new Iterable<AsmClassInfo>()
		{

			@Override
			public Iterator<AsmClassInfo> iterator()
			{
				return new Iterator<AsmClassInfo>()
				{
					final Iterator<? extends AsmPackageInfo> packagesIterator = packages.iterator();

					Iterator<? extends AsmClassInfo> classInfoIterator;
					
					@Override
					public boolean hasNext()
					{
						while ( classInfoIterator == null || ( ! classInfoIterator.hasNext() ) )
							// ueber leere Packages laufen
						{
							if ( ! packagesIterator.hasNext() )
							{
								return false;
							}
							
							classInfoIterator = packagesIterator.next().getAllClass().iterator();
						}

						return classInfoIterator.hasNext();
					}

					@Override
					public AsmClassInfo next()
					{
						if ( classInfoIterator == null || ( ! classInfoIterator.hasNext() ) )
						{
							classInfoIterator = packagesIterator.next().getAllClass().iterator();
						}

						return classInfoIterator.next();
					}
				};
			}
		};
	}

}
