package de.heinerkuecker.asm_improved;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.TreeMap;

import de.heinerkuecker.asm_improved.type.AsmArray;
import de.heinerkuecker.asm_improved.type.AsmTopLevelOrInnerClassType;
import de.heinerkuecker.asm_improved.type.AsmClassType;
import de.heinerkuecker.asm_improved.type.AsmType;
import de.heinerkuecker.asm_improved.type.generics.arg.AsmExtendsWildcard;
import de.heinerkuecker.asm_improved.type.generics.arg.AsmSuperWildcard;
import de.heinerkuecker.asm_improved.type.generics.arg.AsmUnboundWildcard;
import de.heinerkuecker.asm_improved.type.generics.param.AsmAbstractTypeParam;
import de.heinerkuecker.asm_improved.type.generics.param.AsmExtendsTypeParam;
import de.heinerkuecker.asm_improved.type.generics.param.AsmUnboundedTypeParam;
import de.heinerkuecker.asm_improved.type.generics.var.AsmTypevar;
import de.heinerkuecker.asm_improved.type.member.AsmInnerClassType;
import de.heinerkuecker.asm_improved.type.primitive.AsmBoolean;
import de.heinerkuecker.asm_improved.type.primitive.AsmByte;
import de.heinerkuecker.asm_improved.type.primitive.AsmChar;
import de.heinerkuecker.asm_improved.type.primitive.AsmDouble;
import de.heinerkuecker.asm_improved.type.primitive.AsmFloat;
import de.heinerkuecker.asm_improved.type.primitive.AsmInt;
import de.heinerkuecker.asm_improved.type.primitive.AsmLong;
import de.heinerkuecker.asm_improved.type.primitive.AsmShort;
import de.heinerkuecker.asm_improved.util.Fqn;
import de.heinerkuecker.asm_improved.util.ImmutableArray;

/**
 * Parser f&uuml;r Type-Informationen,
 * mit und ohne Generics,
 * anhand der von
 * <a href="asm.ow2.org">ASM</a> gelieferten
 * Strings.
 * 
 * @author Heiner K&uuml;cker
 */
final class AsmDescriptionAndSignatureParser
{
	/**
	 * Konstruktor private,
	 * Instanziieren verboten,
	 * Util-Klasse.
	 */
	private AsmDescriptionAndSignatureParser()
	{
		super();
	}

	static MethodTypeparamsAndReturnAndParamsTuple parseMethodDescriptionAndSignature(
            final String desc ,
            final String signature ,
			final TreeMap<String, ? extends String> innerClassFileNameToDotSeparatedNameMap )
	{
		final String strToParse =
				determineDescOrSignatureAndRemoveExceptions(
						desc ,
						signature );

		final ImmutableArray<? extends AsmAbstractTypeParam> typeParams;
		final ImmutableArray<? extends AsmType> paramTypes;
		final int returnTypeStartIndex;
		if ( strToParse.startsWith( "()" ) )
			// keine Parameter
		{
			typeParams = null;
			paramTypes = null;
			returnTypeStartIndex = "()".length();
		}
		else if ( strToParse.startsWith( "(" ) )
		{
			typeParams = null;

			final ArrayList<AsmType> tmpParamTypes = new ArrayList<>();

			String paramsTypeStrToParse =
					strToParse.substring(
							//beginIndex: fhrende Klammer ueberspringen
							1 ,
							//endIndex
							strToParse.lastIndexOf( ')' ) );

			while ( ! paramsTypeStrToParse.isEmpty() )
			{
				final String paramTypeStrToParse =
						takeParamTypeStrFromParamsTypeStr(
								paramsTypeStrToParse );

				final AsmType paramType =
						parseTypeStr(
								paramTypeStrToParse ,
								innerClassFileNameToDotSeparatedNameMap );

				tmpParamTypes.add( paramType );
				
				paramsTypeStrToParse =
						paramsTypeStrToParse.substring(
								//beginIndex
								paramTypeStrToParse.length() );
			}

			tmpParamTypes.trimToSize();
			paramTypes =
					//Collections.unmodifiableList( tmpParamTypes )
					ImmutableArray.create( tmpParamTypes );
			
			returnTypeStartIndex = strToParse.lastIndexOf( ')' ) + 1;
		}
		else if ( strToParse.startsWith( "<" ) )
			// Typ-Parameter
		{
			final int typeParamsEndIndex =
					AsmDescAndSignatureParserUtil.getAfterGtPosRecursiv(
							strToParse ,
							// damit die Methode getAfterGtPosRecursiv funktioniert, muss das beginnende kleiner-als-Zeichen verbraucht worden sein TODO evtl umbauen
							0 + 1 );

			final String typeParamsStr =
					strToParse.substring(
							//beginIndex
							0 ,
							typeParamsEndIndex );

			typeParams =
					parseTypeParams(
							typeParamsStr ,
							innerClassFileNameToDotSeparatedNameMap );

			final String strToParseWithoutTypeParams =
					strToParse.substring(
							typeParamsEndIndex );

			final MethodTypeparamsAndReturnAndParamsTuple tplWithoutTypeparams =
					parseMethodDescriptionAndSignature(
							//desc
							null ,
							//signature
							strToParseWithoutTypeParams ,
							innerClassFileNameToDotSeparatedNameMap );

			return new MethodTypeparamsAndReturnAndParamsTuple(
					typeParams ,
					tplWithoutTypeparams.returnType ,
					tplWithoutTypeparams.paramTypes );
		}
		else
		{
			throw new IllegalArgumentException( strToParse );
		}

		final String returnTypeStrToParse =
				strToParse.substring(
						returnTypeStartIndex );

		final AsmType returnType =
				parseTypeStr(
						returnTypeStrToParse ,
						innerClassFileNameToDotSeparatedNameMap );

		return new MethodTypeparamsAndReturnAndParamsTuple(
				typeParams ,
				returnType ,
				paramTypes );
	}

	/**
	 * Entscheiden, ob desc oder signature geparst werden soll.
	 * 
	 * @param desc
	 * @param signature
	 * @return zu parsender Type-String
	 */
    private static String determineDescOrSignatureAndRemoveExceptions(
            final String desc ,
            final String signature )
    {
	    final String strToParse;
		{
			final String strToParseWithExceptions;
			if ( signature == null )
			{
				// alte Welt ohne Generics
				strToParseWithExceptions = desc;
			}
			else
			{
				// beim Auftreten von Generics sind in signature alle Informationen aus desc auch enthalten
				strToParseWithExceptions = signature;
			}
			
			if ( strToParseWithExceptions.contains( "^" ) )
				// checked Exceptions in der Signatur enthalten
			{
				strToParse =
						strToParseWithExceptions.substring(
								//beginIndex
								0 ,
								//endIndex
								strToParseWithExceptions.indexOf( '^' ) );
			}
			else
			{
				strToParse = strToParseWithExceptions;
			}
		}
	    return strToParse;
    }
	
	private static ImmutableArray<? extends AsmAbstractTypeParam> parseTypeParams(
			final String typeParamsStr ,
			final TreeMap<String, ? extends String> innerClassFileNameToDotSeparatedNameMap )
	{
		if ( ! typeParamsStr.startsWith( "<" ) )
		{
			throw new IllegalArgumentException( typeParamsStr );
		}

		String restTypeParamsStr =
				typeParamsStr.substring(
						//beginIndex: spitze Klammer (kleiner-als) am Anfang ueberspringen
						1 ,
						//endIndex: endende spitze Klammer (groesser-als) ueberspringen
						typeParamsStr.length() - 1 );

		final ArrayList<AsmAbstractTypeParam> result = new ArrayList<>();

		while ( ! restTypeParamsStr.isEmpty() )
		{
			final int colonPos = restTypeParamsStr.indexOf( ':' );

			final String typeParamName =
					restTypeParamsStr.substring(
							//beginIndex
							0 ,
							//endIndex
							colonPos );

			final int currTypParamEndIndex =
					AsmDescAndSignatureParserUtil.getTypeEndPos(
							restTypeParamsStr ,
							//startPos
							colonPos + 1 ,
							//primitivePossible 
							false );

			final String currTypParamStrToParse =
					restTypeParamsStr.substring(
							//beginIndex
							colonPos + 1 ,
							//endIndex
							currTypParamEndIndex + 1 );

			final AsmAbstractTypeParam currTypParam;
			if ( currTypParamStrToParse.equals( "Ljava/lang/Object;" ) )
				// seltsamerweise steht in ASM dieser Typ nach dem Namen des Typ-Parameters, wenn dieser keinen Bound hat
			{
				currTypParam = new AsmUnboundedTypeParam( typeParamName );
			}
			else if ( currTypParamStrToParse.startsWith( "L" ) ||
					currTypParamStrToParse.startsWith( "T" ) )
				// dies bedeuted auch extends
				// signature: <K:Ljava/lang/Enum<TK;>;>(Ljava/lang/Class<TK;>;)[TK;
				// EnumMap: private static <K extends Enum<K>> K[] getKeyUniverse(Class<K> keyType)
			{
				final AsmType currTypParamBoundType =
						parseTypeStr(
								currTypParamStrToParse ,
								innerClassFileNameToDotSeparatedNameMap );

				currTypParam =
						new AsmExtendsTypeParam(
								typeParamName ,
								currTypParamBoundType );
			}
			else if ( currTypParamStrToParse.startsWith( ":" ) )
				// zwei Doppelpunkte bedeuten extends
			{
				final AsmType currTypParamBoundType =
						parseTypeStr(
								currTypParamStrToParse.substring(
										//beginIndex: fuehrenden Doppelpunkt entfernen
										1 ) ,
								innerClassFileNameToDotSeparatedNameMap );

				currTypParam =
						new AsmExtendsTypeParam(
								typeParamName ,
								currTypParamBoundType );
			}
			else
			{
				throw new IllegalArgumentException( typeParamsStr );
			}

			result.add( currTypParam );

			restTypeParamsStr =
					restTypeParamsStr.substring(
							//beginIndex
							currTypParamEndIndex + 1 );
		}

		//result.trimToSize();
		//return Collections.unmodifiableList( result );
		return ImmutableArray.create( result );
	}

	private static String takeParamTypeStrFromParamsTypeStr(
			final String paramsTypeStr )
	{
		int startPosForArrayPrefix = 0;

		while ( paramsTypeStr.charAt( startPosForArrayPrefix ) == '[' )
			// bei mehrdimensionalen Arrays stehen vor dem Element-Type oeffnende eckige Klammern in der Anzahl der Array-Dimansionen
		{
			startPosForArrayPrefix++;
		}

		if ( paramsTypeStr.charAt( startPosForArrayPrefix ) == 'L' ||
				// der Parameter ist eine Klasse
				paramsTypeStr.charAt( startPosForArrayPrefix ) == 'T' )
			// der Parameter ist eine Typ-Variable
		{
			final int endIndex =
					AsmDescAndSignatureParserUtil.getTypeEndPos(
							paramsTypeStr ,
							startPosForArrayPrefix + 1 ,
							//primitivePossible
							false ) + 1;

			return paramsTypeStr.substring(
					//beginIndex
					0 ,
					endIndex );
		}

		return paramsTypeStr.substring(
				//beginIndex
				0 ,
				//endIndex
				startPosForArrayPrefix + 1 );
	}

	private static AsmType parseTypeStr(
			final String typeStr ,
			final TreeMap<String, ? extends String> innerClassFileNameToDotSeparatedNameMap )
	{
    	// die interne Codierung desc ist zu erkennen im Quellcode von org.objectweb.asm.Type.getDescriptor(StringBuffer, Class<?>)
    	switch ( typeStr )
        {
        case "I":
			return AsmInt.INSTANCE;
        case "Z":
			return AsmBoolean.INSTANCE;
        case "B":
			return AsmByte.INSTANCE;
        case "C":
			return AsmChar.INSTANCE;
        case "S":
        	return AsmShort.INSTANCE;
        case "J":
			return AsmLong.INSTANCE;
        case "D":
        	return AsmDouble.INSTANCE;
        case "F":
        	return AsmFloat.INSTANCE;
        case "V":
        	// void
			return null;
        default:
        	if ( typeStr.startsWith( "L" ) )
        		// eine Klasse
        	{
        		final ImmutableArray<? extends AsmType> typeArgs;
        		final int classSlashFqnEndIndex;
        		if ( ! typeStr.contains( "<" ) )
        			// keine Generics
        		{
        			classSlashFqnEndIndex =
        					//endIndex: hinten steht ein Semikolon
        					typeStr.length() - 1;

        			typeArgs = null;
        		}
        		else
        		{
        			classSlashFqnEndIndex =
        					//endIndex: hinten steht ein '<'-Zeichen
        					typeStr.indexOf( '<' );

        			final ArrayList<AsmType> tmpTypeArgs = new ArrayList<>();

        			int currTypeArgPos = classSlashFqnEndIndex + 1;

        			while ( typeStr.charAt( currTypeArgPos ) != '>' )
        			{
        				final String currTypeArgStrToParse;
        				if ( typeStr.charAt( currTypeArgPos ) == '*' )
        					// Sonderfall unbound-wildcard
        				{
            				currTypeArgStrToParse = "*";
        				}
        				else
        				{
            				final int currTypeArgEndIndex =
            						AsmDescAndSignatureParserUtil.getTypeEndPos(
            								typeStr ,
            								currTypeArgPos ,
            								//primitivePossible
            								true ) + 1;

            				currTypeArgStrToParse =
            						typeStr.substring(
            								//beginIndex
            								currTypeArgPos ,
            								//endIndex
            								currTypeArgEndIndex );
        				}

        				final AsmType currTypeArg =
        						parseTypeStr(
        								currTypeArgStrToParse ,
        								innerClassFileNameToDotSeparatedNameMap );

        				tmpTypeArgs.add( currTypeArg );

        				currTypeArgPos +=
        						currTypeArgStrToParse.length();
        			}

        			tmpTypeArgs.trimToSize();
        			typeArgs = /*Collections.unmodifiableList*/ImmutableArray.create( tmpTypeArgs );
        		}

        		final String rawName =
        				typeStr.substring(
        						//beginIndex: vorn steht L
        						1 ,
        						classSlashFqnEndIndex );

        		//final String slashToDotReplacedFqn = rawName.replace( '/' , '.' );
        		//final String packageName;
        		//final String simpleName;
        		//if ( slashToDotReplacedFqn.contains( "." ) )
        		//{
        		//	packageName =
        		//			slashToDotReplacedFqn.substring(
        		//					0 ,
        		//					slashToDotReplacedFqn.lastIndexOf( '.' ) );
        		//
        		//	simpleName =
        		//			slashToDotReplacedFqn.substring(
        		//					slashToDotReplacedFqn.lastIndexOf( '.' ) + 1 );;
        		//}
        		//else
        		//{
        		//	packageName = "";
        		//	simpleName = slashToDotReplacedFqn;
        		//}

        		final Fqn fqn =
                        //new Fqn(
                        //      new PackageName( packageName ) ,
                        //      DottedName.valueOfDotSeparatedStr( simpleName ) ) ,
                        InnerOrNestedClassNameResolver.resolve(
                                //dollarSeparatedRawName
                                rawName ,
                                innerClassFileNameToDotSeparatedNameMap );

        		return new AsmClassType(
        		        fqn ,
        				typeArgs );
        	}
        	else if ( typeStr.startsWith( "[" ) )
        		// ein Array
        	{
        		return new AsmArray(
        				parseTypeStr(
        						// fuehrende Klammer ueberspringen
        						typeStr.substring( 1 ) ,
        						innerClassFileNameToDotSeparatedNameMap ) );
        	}
        	else if ( typeStr.startsWith( "+" ) )
        		// ein extends-Wildcard
        	{
        		return new AsmExtendsWildcard(
        				parseTypeStr(
        						// fuehrendes Plus-Zeichen ueberspringen
        						typeStr.substring( 1 ) ,
        						innerClassFileNameToDotSeparatedNameMap ) );
        	}
        	else if ( typeStr.startsWith( "*" ) )
        		// ein unbound-Wildcard
        	{
        		return AsmUnboundWildcard.INSTANCE;
        	}
        	else if ( typeStr.startsWith( "-" ) )
				// super-wildcard
				// signature: (Ljava/util/function/Consumer<-TT;>;)V
				// java.util.Optional<T>: public void ifPresent(Consumer<? super T> consumer)
			{
        		return new AsmSuperWildcard(
        				parseTypeStr(
        						// fuehrendes Minus-Zeichen ueberspringen
        						typeStr.substring( 1 ) ,
        						innerClassFileNameToDotSeparatedNameMap ) );
			}
        	else if ( typeStr.startsWith( "T" ) )
        		// eine Typ-Variable
        	{
        		return parseTypvar(
        				typeStr ,
        				innerClassFileNameToDotSeparatedNameMap );
        	}
        	else
        	{
        		throw new IllegalArgumentException( typeStr );
        	}
        }
	}

	///**
	// * Semikolon suchen unter Beachtung
	// * Schachtelung Generics.
	// * 
	// * @param str
	// * @param startPos
	// * @return
	// */
	//static int getSemicolonPosUnderConsiderationGenerics(
	//		final String str ,
	//		final int startPos ,
	//		// bei Typ-Variablen koennte der Name der Typ-Variablen dem Code eines primitive entsprechen
	//		final boolean primitivePossible )
	//{
	//	if ( primitivePossible &&
	//			( str.charAt( startPos ) == 'I' ||
	//			str.charAt( startPos ) == 'I' ||
	//			str.charAt( startPos ) == 'Z' ||
	//			str.charAt( startPos ) == 'B' ||
	//			str.charAt( startPos ) == 'C' ||
	//			str.charAt( startPos ) == 'S' ||
	//			str.charAt( startPos ) == 'J' ||
	//			str.charAt( startPos ) == 'D' ||
	//			str.charAt( startPos ) == 'F' ||
	//			str.charAt( startPos ) == 'V' ) )
	//		// ein Primitive hat Semikolon
	//		// erstes Zeichen des Typs ist auch letztes Zeichen des Typs
	//	{
	//		return startPos;
	//	}
	//	
	//	if ( str.charAt( startPos ) == '[' )
	//		// Array
	//	{
	//		// rekursiver Aufruf, damit die Logik mit den Primitiven erneut wirken kann
	//		// java.util.Calendar, Field: cachedLocaleData
	//		// Ljava/util/concurrent/ConcurrentMap<Ljava/util/Locale;[I>;
	//		return getSemicolonPosUnderConsiderationGenerics(
	//				str ,
	//				startPos + 1 ,
	//				primitivePossible );
	//	}
	//
	//	int currPos = startPos;
	//
	//	while ( str.charAt( currPos ) != ';' )
	//	{
	//		if ( str.charAt( currPos ) == '<' )
	//			// spitze Klammer Generics
	//		{
	//			currPos =
	//					getAfterGtPosRecursiv(
	//							str ,
	//							// damit die Methode getAfterGtPosRecursiv funktioniert, muss das beginnende kleiner-als-Zeichen verbraucht worden sein TODO evtl umbauen
	//							currPos + 1 );
	//		}
	//		else
	//		{
	//			currPos++;
	//		}
	//	}
	//
	//	return currPos;
	//}

	///**
	// * Groesser-Als-Zeichen suchen
	// * mit moeglicher Schachtelung
	// * Kleiner-Als-Zeichen zu
	// * Groesser-Als-Zeichen.
	// * 
	// * @param str
	// * @param startPos
	// * @return
	// */
	//private static int getAfterGtPosRecursiv(
	//		final String str ,
	//		final int startPos )
	//{
	//	int currPos = startPos;
	//
	//	while ( str.charAt( currPos ) != '>' )
	//	{
	//		if ( str.charAt( currPos ) == '<' )
	//			// spitze Klammer Generics
	//		{
	//			currPos =
	//					getAfterGtPosRecursiv(
	//							str ,
	//							currPos + 1 );
	//		}
	//		else
	//		{
	//			currPos++;
	//		}
	//	}
	//
	//	// '>' nicht noch einmal pruefen
	//	return currPos + 1;
	//}
	
	public static AsmType parseFieldDescriptionAndSignature(
			final String desc ,
			final String signature ,
			final TreeMap<String, ? extends String> innerClassFileNameToDotSeparatedNameMap )
	{
		final String strToParse =
				determineDescOrSignatureAndRemoveExceptions(
						desc ,
						signature );

		final AsmType fieldType =
				parseTypeStr(
						strToParse ,
						innerClassFileNameToDotSeparatedNameMap );

		return fieldType;
	}

    static ClassTypeparamsAndInheritTypeArgsTuple parseClassTypeparamsAndInheritTypeArgsTupel(
            final String signature ,
			final TreeMap<String, ? extends String> innerClassFileNameToDotSeparatedNameMap )
    {
    	if ( signature == null || signature.isEmpty() )
        {
	        return null;
        }

        final String typeParamParsedRestSignature;
    	final ImmutableArray<? extends AsmAbstractTypeParam> typeParams;
    	if ( signature.startsWith( "<" ) )
        {
    		final String typeParamStrToParse =
    				signature.substring(
    						//beginIndex
    						0 ,
    						//endIndex
    						AsmDescAndSignatureParserUtil.getAfterGtPosRecursiv(
    								signature ,
    								1 ) );

    		//System.out.println( typeParamStrToParse );

    		typeParams =
    				parseTypeParams(
    						typeParamStrToParse ,
    						innerClassFileNameToDotSeparatedNameMap );

    		typeParamParsedRestSignature =
    				signature.substring(
    						typeParamStrToParse.length() );
        }
    	else
    	{
    		typeParams = null;
    		typeParamParsedRestSignature = signature;
    	}

        final String superTypeArgsParsedRestSignature;
    	final ImmutableArray<? extends AsmType> superTypeArgs;
    	{
    		final int semicolonPos = typeParamParsedRestSignature.indexOf( ';' );
    		final int leftAnglePos = typeParamParsedRestSignature.indexOf( '<' );
    		
    		if ( leftAnglePos > - 1 &&
    				leftAnglePos < semicolonPos )
    			// es gibt ein '<'-Zeichen und dieses liegt vor dem Semikolon (ein Semikolon msste immer auftauchen)
    		{
    			final int afterGtPosRecursiv =
    					AsmDescAndSignatureParserUtil.getAfterGtPosRecursiv(
    							typeParamParsedRestSignature ,
    							leftAnglePos + 1 );

    			final String superTypeArgsStrToParse =
    					typeParamParsedRestSignature.substring(
    							//beginIndex
    							leftAnglePos ,
    							//endIndex
    							afterGtPosRecursiv );

    			superTypeArgs =
    					parseTypeArgs(
    							superTypeArgsStrToParse ,
    							innerClassFileNameToDotSeparatedNameMap );

    			superTypeArgsParsedRestSignature =
    					typeParamParsedRestSignature.substring(
    							afterGtPosRecursiv + 1 );
    		}
    		else
    		{
    			superTypeArgs = null;
    			superTypeArgsParsedRestSignature =
    					typeParamParsedRestSignature.substring(
    							semicolonPos + 1 );
    		}
    	}

    	//final Map<String, List<? extends AsmType>> interfaceTypeArgs;
    	final List<ImmutableArray<? extends AsmType>> interfaceTypeArgs;
    	if ( superTypeArgsParsedRestSignature.isEmpty() )
    	{
    		interfaceTypeArgs = null;
    	}
    	else
    	{
    		//final Map<String, List<? extends AsmType>> tmpInterfaceTypeArgs = new HashMap<>();
    		final List<ImmutableArray<? extends AsmType>> tmpInterfaceTypeArgs = new ArrayList<>();

    		String interfaceTypeArgsRestStr = superTypeArgsParsedRestSignature;

    		while ( ! interfaceTypeArgsRestStr.isEmpty() )
    		{
    			//System.out.println( interfaceTypeArgsRestStr );

    			final int semicolonPos = interfaceTypeArgsRestStr.indexOf( ';' );
        		final int leftAnglePos = interfaceTypeArgsRestStr.indexOf( '<' );

        		if ( leftAnglePos > - 1 &&
        				leftAnglePos < semicolonPos )
        			// es gibt ein '<'-Zeichen und dieses liegt vor dem Semikolon (ein Semikolon msste immer auftauchen)
        		{
        			final int afterGtPosRecursiv =
        					AsmDescAndSignatureParserUtil.getAfterGtPosRecursiv(
        							interfaceTypeArgsRestStr ,
        							leftAnglePos + 1 );

        			final String currInterfaceTypeArgsStrToParse =
        					interfaceTypeArgsRestStr.substring(
        							//beginIndex
        							leftAnglePos ,
        							//endIndex
        							afterGtPosRecursiv );

        			final String interfaceClassNamePartStr =
        					interfaceTypeArgsRestStr.substring(
        							//beginIndex
        							1 ,
        							//endIndex
        							leftAnglePos );

        			final ImmutableArray<? extends AsmType> currInterfaceTypeArgs =
        					parseTypeArgs(
        							currInterfaceTypeArgsStrToParse ,
        							innerClassFileNameToDotSeparatedNameMap );

        			//final String interfaceClassFqName = interfaceClassNamePartStr.replace( '/' , '.' );
        			//tmpInterfaceTypeArgs.put( interfaceClassFqName , currInterfaceTypeArgs );
        			tmpInterfaceTypeArgs.add( currInterfaceTypeArgs );

        			interfaceTypeArgsRestStr =
        					interfaceTypeArgsRestStr.substring(
        							afterGtPosRecursiv + 1 );
        		}
        		else
        		{
        			// Einhaltung Position absichern
        			tmpInterfaceTypeArgs.add( null );

        			interfaceTypeArgsRestStr =
        					interfaceTypeArgsRestStr.substring(
        							semicolonPos + 1 );
        		}
    		}

    		interfaceTypeArgs =
    				Collections./*unmodifiableMap*/unmodifiableList(
    						tmpInterfaceTypeArgs );
    	}
    	
    	return new ClassTypeparamsAndInheritTypeArgsTuple(
    			typeParams ,
    			superTypeArgs ,
    			ImmutableArray.create( interfaceTypeArgs ) );
    }

	private static ImmutableArray<? extends AsmType> parseTypeArgs(
			final String typeArgsStr ,
			final TreeMap<String, ? extends String> innerClassFileNameToDotSeparatedNameMap )
	{
		if ( ! typeArgsStr.startsWith( "<" ) )
		{
			throw new IllegalArgumentException( typeArgsStr );
		}

		String restTypeArgsStr =
				typeArgsStr.substring(
						//beginIndex: spitze Klammer (kleiner-als) am Anfang ueberspringen
						1 ,
						//endIndex: endende spitze Klammer (groesser-als) ueberspringen
						typeArgsStr.length() - 1 );

		final ArrayList<AsmType> result = new ArrayList<>();

		while ( ! restTypeArgsStr.isEmpty() )
		{
			final AsmType currTypArg;
			final int currTypArgEndIndex;
			final int primitiveArrayEndIndex;
			if ( ( primitiveArrayEndIndex =
					getPrimitiveArrayEndIndex(
							restTypeArgsStr ) ) > 0 )
				// als Type-Argument kann ein Primitiv-Array auftauchen, siehe java.util.stream.Node.OfDouble
			{
				currTypArgEndIndex = primitiveArrayEndIndex;
			}
			else
			{
				currTypArgEndIndex =
						AsmDescAndSignatureParserUtil.getTypeEndPos(
								restTypeArgsStr ,
								//startPos
								0 ,
								//primitivePossible 
								false ) + 1;
			}

			final String currTypArgStrToParse =
					restTypeArgsStr.substring(
							//beginIndex
							0 ,
							//endIndex
							currTypArgEndIndex );

			currTypArg =
					parseTypeStr(
							currTypArgStrToParse ,
							innerClassFileNameToDotSeparatedNameMap );

			result.add( currTypArg );

			restTypeArgsStr =
					restTypeArgsStr.substring(
							//beginIndex
							currTypArgEndIndex );
		}

		//result.trimToSize();
		//return Collections.unmodifiableList( result );
		return ImmutableArray.<AsmType>create( result );
	}

	/**
	 * Feststellen, ob ein Primitiv-Array codiert wurde und End-Index zurckgeben
	 * @param typeStrWithPossiblePrimitiveArrTypeAtStart Type-String mit eventuellem codiertem Primitiv-Array am Anfang
	 * @return der End-Index (Basis 0, rechts aussen liegen) des Primitiv-Arrays oder -1, wenn Primitiv-Array codiert wurde
	 */
	private static int getPrimitiveArrayEndIndex(
			final String typeStrWithPossiblePrimitiveArrTypeAtStart )
	{
		//if ( ! typeStrWithPossiblePrimitiveArrTypeAtStart.startsWith( "[" ) )
		//{
		//	return -1;
		//}

		int currPos = 0;

		while ( true )
		{
			final char currChar =
					typeStrWithPossiblePrimitiveArrTypeAtStart.charAt(
							currPos );

			if ( currChar == '[' )
				// nchste Array-Dimension
			{
				currPos++;
			}
			else if (
					currChar == 'I' ||
					currChar == 'Z' ||
					currChar == 'B' ||
					currChar == 'C' ||
					currChar == 'S' ||
					currChar == 'J' ||
					currChar == 'D' ||
					currChar == 'F' )
				// Kenn-Buchstabe fr Primitive
			{
				return currPos + 1;
			}
			else
			{
				return -1;
			}
		}
	}

	private static AsmType parseTypvar(
			final String typeStr ,
			final TreeMap<String, ? extends String> innerClassFileNameToDotSeparatedNameMap )
	{
		final String restTypeStr =
				typeStr.substring(
						//beginIndex: T ueberspringen
						1 ,
						//endIndex: endendes Semikolon weglassen
						typeStr.length() - 1 );

		AsmType result = null;

		final List<? extends TypeNameAndTypeArgsStrTuple> separatedInnerOrNestedClassParts =
				separateInnerOrNestedClassParts(
						restTypeStr );

		// der Name der Typ-Variable enthlt einen Punkt, scheinbar ist der Typ ein inner/nested type of type variable
		// Testfall: TT.InnerIf<Ljava/lang/Integer;>;
		// Testfall: TT.InnerIf.InnerInnerIf<Ljava/lang/Integer;>;
		for ( final TypeNameAndTypeArgsStrTuple part : separatedInnerOrNestedClassParts )
		{
			if ( result == null )
			{
				// es gab noch keinen outer/enclosing type
				result = new AsmTypevar( part.typeName );
			}
			else
			{
				final ImmutableArray<? extends AsmType> typeArgs;
				if ( part.typeArgsStr != null )
				{
					typeArgs =
							parseTypeArgs(
									part.typeArgsStr ,
									innerClassFileNameToDotSeparatedNameMap );
				}
				else
				{
					typeArgs = null;
				}

				result =
						//new AsmMemberClassType(
						//		//outerOrEnclosingType
						//		result ,
						//		//relativName
						//		DottedName.valueOfNoDotSeparatedStr( part.typeName ) ,
						//		typeArgs );
						//new AsmClassType(
						//		new Fqn(
						//				( (AsmClassType) result ).fqn.packageName ,
						//				DottedName.valueOfDotSeparatedStr( ( (AsmClassType) result ).fqn.dottedName.toClassFileName() + "." + part.typeName ) ) , 
						//		typeArgs );
						new AsmInnerClassType(
								//outerOrEnclosingType
								(AsmTopLevelOrInnerClassType) result ,
								//relativName
								//DottedName.valueOfNoDotSeparatedStr(
								//new JavaIdentifier(
										part.typeName ,
								typeArgs );

			}
		}
		
		return result;
		//while ( ! restTypeStr.isEmpty() )
		//{
		//	if ( ! restTypeStr.contains( "." ) )
		//	{
		//		if ( result == null )
		//		{
		//			// es gab keinen outer/enclosing type
		//			result = new AsmTypevar( restTypeStr );
		//		}
		//		else
		//		{
		//			result =
		//					new AsmMemberClassType(
		//							//outerOrEnclosingType
		//							result ,
		//							restTypeStr ,
		//							typeArgs );		
		//		}
		//	}
		//	else
		//	{
		//		// der Name der Typ-Variable enthlt einen Punkt, scheinbar ist der Typ ein inner/nested type of type variable
		//		// Testfall: TT.InnerIf<Ljava/lang/Integer;>;
		//		// Testfall: TT.InnerIf.InnerInnerIf<Ljava/lang/Integer;>;
		//		final String typevarNameOfTypevarInnerOrNestedType =
		//				restTypeStr.substring(
		//						//beginIndex
		//						0 ,
		//						//endIndex
		//						restTypeStr.indexOf( '.' ) );
		//
		//		final AsmTypevar asmTypevarWithInnerOrNestedType =
		//				new AsmTypevar(
		//						typevarNameOfTypevarInnerOrNestedType );
		//
		//		final String typevarInnerOrNestedTypStr =
		//				restTypeStr.substring(
		//						//beginIndex
		//						typevarNameOfTypevarInnerOrNestedType.length() + 1 );
		//
		//		final String typevarInnerOrNestedTypeName;
		//		final List<? extends AsmType> typeArgs;
		//		if ( typevarInnerOrNestedTypStr.contains( "<" ) )
		//		{
		//			typevarInnerOrNestedTypeName =
		//					typevarInnerOrNestedTypStr.substring(
		//							//beginIndex
		//							0 ,
		//							//endIndex
		//							typevarInnerOrNestedTypStr.indexOf( '<' ) );
		//
		//			final String typeArgsStr =
		//					typevarInnerOrNestedTypStr.substring(
		//							//beginIndex
		//							typevarInnerOrNestedTypStr.indexOf( '<' ) );
		//
		//			typeArgs =
		//					parseTypeArgs(
		//							typeArgsStr );
		//		}
		//		else
		//		{
		//			typevarInnerOrNestedTypeName = typevarInnerOrNestedTypStr;
		//			typeArgs = null;
		//		}
		//	}
		//}
		//return result;
	}

	// Testfall: TT.InnerIf<Ljava/lang/Integer;>;
	// Testfall: TT.InnerIf.InnerInnerIf<Ljava/lang/Integer;>;
	// fhrendes 'TT.' bereits abgeschnitten
	private static List<? extends TypeNameAndTypeArgsStrTuple> separateInnerOrNestedClassParts(
			final String strToSeparate )
	{
		if ( ! strToSeparate.contains( "." ) )
		{
			return Arrays.asList(
					separateTypeNameAndTypeArgs(
							strToSeparate ) );
		}

		String restStrToSeparate = strToSeparate;

		final ArrayList<TypeNameAndTypeArgsStrTuple> result = new ArrayList<>();

		while ( ! restStrToSeparate.isEmpty() )
		{
			final int dotPos = restStrToSeparate.indexOf( '.' );
			final int leftAngleBracePos = restStrToSeparate.indexOf( '<' );

			if ( dotPos < 0 )
				// es gibt kein '.'
			{
				result.add(
						separateTypeNameAndTypeArgs(
								restStrToSeparate ) );

				restStrToSeparate = "";
			}
			else if (
					// es gibt ein '.'
					// es gibt kein '<'
					leftAngleBracePos < 0 ||
					// oder das '<' liegt hinter dem Punkt
					dotPos < leftAngleBracePos )
			{
				result.add(
						separateTypeNameAndTypeArgs(
								restStrToSeparate.substring(
										//beginIndex
										0 ,
										//endIndex
										dotPos ) ) );

				restStrToSeparate =
						restStrToSeparate.substring(
								//beginIndex
								dotPos + 1 );
			}
			else
				// es gibt ein '.'
				// es gibt ein '<'
				// das '<' liegt vor dem Punkt
			{
				final int afterGtPosRecursiv =
						AsmDescAndSignatureParserUtil.getAfterGtPosRecursiv(
								restStrToSeparate ,
								leftAngleBracePos + 1 );

				result.add(
						separateTypeNameAndTypeArgs(
								restStrToSeparate.substring(
										//beginIndex
										0 ,
										//endIndex
										afterGtPosRecursiv ) ) );

				restStrToSeparate =
						restStrToSeparate.substring(
								//beginIndex: trennenden Punkt berspringen
								afterGtPosRecursiv + 1 );
			}
		}

		return result;
	}

	private static TypeNameAndTypeArgsStrTuple separateTypeNameAndTypeArgs(
			final String strToSep )
	{
		if ( ! strToSep.contains( "<" ) )
		{
			return new TypeNameAndTypeArgsStrTuple(
					strToSep ,
					//typeArgsStr
					null );
		}

		final int leftAngleBracePos = strToSep.indexOf( '<' );
		
		return new TypeNameAndTypeArgsStrTuple(
				//typeName
				strToSep.substring(
						//beginIndex
						0 ,
						//endIndex
						leftAngleBracePos ) ,
				//typeArgsStr
				strToSep.substring(
						//beginIndex
						leftAngleBracePos ) );
	}

}
