package de.heinerkuecker.asm_improved;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;

import de.heinerkuecker.asm_improved.annotation_use.AsmAnnotationUse;
import de.heinerkuecker.asm_improved.annotation_use.AsmEnumValueUse;
import de.heinerkuecker.asm_improved.asm_dom.AsmDomAnnotation;
import de.heinerkuecker.asm_improved.asm_dom.AsmDomAnnotationAnnotationValue;
import de.heinerkuecker.asm_improved.asm_dom.AsmDomAnnotationDefault;
import de.heinerkuecker.asm_improved.asm_dom.AsmDomAnnotationEnumValue;
import de.heinerkuecker.asm_improved.asm_dom.AsmDomAnnotationPrimitiveValue;
import de.heinerkuecker.asm_improved.asm_dom.AsmDomAnnotationValue;
import de.heinerkuecker.asm_improved.asm_dom.AsmDomArray;
import de.heinerkuecker.asm_improved.asm_dom.AsmDomClass;
import de.heinerkuecker.asm_improved.asm_dom.AsmDomField;
import de.heinerkuecker.asm_improved.asm_dom.AsmDomInnerClass;
import de.heinerkuecker.asm_improved.asm_dom.AsmDomMethod;
import de.heinerkuecker.asm_improved.asm_dom.AsmDomOuterClass;
import de.heinerkuecker.asm_improved.asm_dom.AsmDomTypeAnnotation;
import de.heinerkuecker.asm_improved.type.AsmArray;
import de.heinerkuecker.asm_improved.type.AsmClassType;
import de.heinerkuecker.asm_improved.type.AsmTopLevelOrInnerClassTypeOrTypevar;
import de.heinerkuecker.asm_improved.type.AsmType;
import de.heinerkuecker.asm_improved.type.AsmTypePath;
import de.heinerkuecker.asm_improved.type.generics.param.AsmAbstractTypeParam;
import de.heinerkuecker.asm_improved.type.generics.param.AsmExtendsTypeParam;
import de.heinerkuecker.asm_improved.type.generics.param.AsmIntersectionBound;
import de.heinerkuecker.asm_improved.type.generics.param.AsmUnboundedTypeParam;
import de.heinerkuecker.asm_improved.type.primitive.AsmBoolean;
import de.heinerkuecker.asm_improved.type.primitive.AsmByte;
import de.heinerkuecker.asm_improved.type.primitive.AsmChar;
import de.heinerkuecker.asm_improved.type.primitive.AsmInt;
import de.heinerkuecker.asm_improved.type.primitive.AsmLong;
import de.heinerkuecker.asm_improved.type.primitive.AsmShort;
import de.heinerkuecker.asm_improved.util.CollectionUtil;
import de.heinerkuecker.asm_improved.util.EmptyArrays;
import de.heinerkuecker.asm_improved.util.Fqn;
import de.heinerkuecker.asm_improved.util.ImmutableArray;
import de.heinerkuecker.asm_improved.util.ImmutableMap;
import de.heinerkuecker.asm_improved.util.PutAddCollMap;

/**
 * Umwandler von {@link AsmDomClass}-Objekten
 * und untergeordneten Objekten zu
 * {@link AsmClassInfo}-Objekten.
 * 
 * @author Heiner K&uuml;cker
 */
public final class AsmDomConverter
{

	/**
	 * Konstruktor private,
	 * Instanziieren verboten,
	 * Util-Klasse.
	 */
	private AsmDomConverter()
	{
		super();
	}

	/**
	 * Erzeugen der von anderen Methoden ben&ouml;tigten
	 * Treemap mit den Namens-Mappings f&uuml;r innere und geschachtelte Klassen.
	 * 
	 * @param domClass DOM-Objekt
	 * @return Treemap mit den Namens-Mappings f&uuml;r innere und geschachtelte Klassen
	 */
	private static TreeMap<String, String> convertToInnerClassFileNameToDotSeparatedNameMap(
			final AsmDomClass domClass
			//final ClasspathFiles classpathFiles
			)
	{
		final TreeMap<String, String> innerClassFileNameToDotSeparatedNameMap = new TreeMap<>();

		for ( final AsmDomInnerClass domInnerClass : domClass.innerClasses )
		{
			final String name = domInnerClass.name;
			final String outerName = domInnerClass.outerName;
			final String innerName = domInnerClass.innerName;

			//if ( name != null && name.contains( "ClassWithAnonymusClass_1" ) )
			//{
			//	System.out.println( "debug break" );
			//}

			if (
					// bei anonymen inneren Klassen sind outerName und innerName null
					outerName != null &&
					innerName != null )
			{
				final String dotSeparatedInnerOrNestedClassRawName =
						outerName + "." + innerName;

				if ( name.length() != dotSeparatedInnerOrNestedClassRawName.length() )
					// die Namens-Strings mit Dollarzeichen und mit Punkt mssen gleich lang sein
				{
					throw new IllegalStateException(
							name +
							" " +
							dotSeparatedInnerOrNestedClassRawName );
				}

				//System.out.println( "name " + name );
				//System.out.println( "dotSeparatedInnerOrNestedClassRawName " + dotSeparatedInnerOrNestedClassRawName );

				innerClassFileNameToDotSeparatedNameMap.put(
						name ,
						dotSeparatedInnerOrNestedClassRawName );
			}
			else
				// scheinbar eine anonyme innere Klasse
			{
				final String dotSeparatedInnerOrNestedClassRawName =
						name.substring( 0 , name.lastIndexOf( '$' ) ) + 
						"." +
						name.substring( name.lastIndexOf( '$' ) + 1 );

				innerClassFileNameToDotSeparatedNameMap.put(
						name ,
						dotSeparatedInnerOrNestedClassRawName );

			}
		}

		//System.out.println( innerClassFileNameToDotSeparatedNameMap.descendingMap() );

		return innerClassFileNameToDotSeparatedNameMap;
	}

	static AsmClassInfo convertClass(
			final AsmDomClass domClass ,
			final File classpathFile ,
			final ClasspathFiles classpathFiles )
	{
		//if ( //domClass.classRawName.startsWith( "com/oneandone/access/mobile/mss/scheduler/impl/PersistentTimerCallbackMXBean" ) ||
		//		domClass.classRawName.contains( "ClassWithInnerClassWithGenerics_0" ) )
		//{
		//	System.out.println( "debug break" );
		//}

		final TreeMap<String, String> innerClassFileNameToDotSeparatedNameMap =
				convertToInnerClassFileNameToDotSeparatedNameMap(
						domClass
						//classpathFiles
						);

		final Fqn classFqn = 
				InnerOrNestedClassNameResolver.resolve(
						domClass.classRawName ,
						innerClassFileNameToDotSeparatedNameMap );

		final boolean isPublic;
		final boolean isFinal;
		final boolean isInterface;
		final boolean isAbstract;
		// TODO isSynthetic
		final boolean isAnnotation;
		final boolean isEnum;
		final boolean isPrivate;
		final boolean isProtected;
		final boolean isStatic;
		//final boolean isInnerOrNested;

		final AsmDomInnerClass selfAsmDomInnerClass =
				getSelfAsmDomInnerClass(
						domClass.classRawName ,
						domClass.innerClasses );

		//if ( domClass.innerClasses.isEmpty() )
		if ( selfAsmDomInnerClass == null )
		{
			isPublic = ( domClass.access & org.objectweb.asm.Opcodes.ACC_PUBLIC ) != 0;
			//isFinal = ( domClass.access & org.objectweb.asm.Opcodes.ACC_FINAL ) != 0;
			//isInterface = ( domClass.access & org.objectweb.asm.Opcodes.ACC_INTERFACE ) != 0;
			//isAbstract = ( domClass.access & org.objectweb.asm.Opcodes.ACC_ABSTRACT ) != 0;
			//isAnnotation = ( domClass.access & org.objectweb.asm.Opcodes.ACC_ANNOTATION ) != 0;
			//isEnum = ( domClass.access & org.objectweb.asm.Opcodes.ACC_ENUM ) != 0;

			isPrivate = false;
			isProtected = false;
			isStatic = true;
			//isInnerOrNested = false;
		}
		else
		{
			//if ( domClass.innerClasses.size() != 1 )
			//{
			//	//throw new IllegalArgumentException( domClass.classRawName );
			//	System.out.println( "debug break" );
			//}

			// see https://docs.oracle.com/javase/specs/jvms/se8/html/jvms-4.html#jvms-4.7.6
			//final AsmDomInnerClass asmDomInnerClass = domClass.innerClasses.get( 0 );

			isPublic = ( selfAsmDomInnerClass.access & org.objectweb.asm.Opcodes.ACC_PUBLIC ) != 0;
			//isFinal = ( asmDomInnerClass.access & org.objectweb.asm.Opcodes.ACC_FINAL ) != 0;
			//isInterface = ( asmDomInnerClass.access & org.objectweb.asm.Opcodes.ACC_INTERFACE ) != 0;
			//isAbstract = ( asmDomInnerClass.access & org.objectweb.asm.Opcodes.ACC_ABSTRACT ) != 0;
			//isAnnotation = ( asmDomInnerClass.access & org.objectweb.asm.Opcodes.ACC_ANNOTATION ) != 0;
			//isEnum = ( asmDomInnerClass.access & org.objectweb.asm.Opcodes.ACC_ENUM ) != 0;

			isPrivate = ( selfAsmDomInnerClass.access & org.objectweb.asm.Opcodes.ACC_PRIVATE ) != 0;
			isProtected = ( selfAsmDomInnerClass.access & org.objectweb.asm.Opcodes.ACC_PROTECTED ) != 0;

			// bytecode ist chaoten c programmierer idioten scheisse
			//isStatic = ( asmDomInnerClass.access & org.objectweb.asm.Opcodes.ACC_STATIC ) != 0;
			//isStatic = domClass.outerClasses.isEmpty();
			// hier gilt immer static, wird aber bei top-level-klassen nicht explizit gesetzt
			//isStatic = ! classFqn.isTopLevel();
			isStatic = ( selfAsmDomInnerClass.access & org.objectweb.asm.Opcodes.ACC_STATIC ) != 0;

			//isInnerOrNested = true;
		}
		isFinal = ( domClass.access & org.objectweb.asm.Opcodes.ACC_FINAL ) != 0;
		isInterface = ( domClass.access & org.objectweb.asm.Opcodes.ACC_INTERFACE ) != 0;
		isAbstract = ( domClass.access & org.objectweb.asm.Opcodes.ACC_ABSTRACT ) != 0;
		isAnnotation = ( domClass.access & org.objectweb.asm.Opcodes.ACC_ANNOTATION ) != 0;
		isEnum = ( domClass.access & org.objectweb.asm.Opcodes.ACC_ENUM ) != 0;

		if ( ! domClass.outerClasses.isEmpty() )
			// Punkt-Name aus outerClass besorgen
		{
			if ( domClass.outerClasses.size() != 1 )
			{
				throw new IllegalArgumentException(
						domClass.classRawName );
			}

			final AsmDomOuterClass asmDomOuterClass = domClass.outerClasses.get( 0 );

			innerClassFileNameToDotSeparatedNameMap.put(
					domClass.classRawName ,
					asmDomOuterClass.owner +
					"." +
					domClass.classRawName.substring( asmDomOuterClass.owner.length() + 1 ) );
		}

		//if ( classFqn.toFqnStr().equals( "de.heinerkuecker.parser.java8.test.class_.inner.ClassWithInnerClassWithGenerics_0.Inner" ) )
		//{
		//	System.out.println( classFqn + " static " + ( domClass.access & org.objectweb.asm.Opcodes.ACC_STATIC ) + " " + domClass.access );
		//}
		//if ( classFqn.toFqnStr().startsWith( "de.heinerkuecker.parser.java8.test.field.from_expression.outer_class.ClassWithAnonymusClass_1$" ) )
		//{
		//	System.out.println( "debug break" );
		//}

		final Fqn superFqn;
		if (
				// ASM liefert seltsamerweise bei Interfaces die Oberklasse java.lang.Object
				isInterface ||
				domClass.superName == null )
		{
			superFqn = null;
		}
		else
		{
			superFqn =
					InnerOrNestedClassNameResolver.resolve(
							//dollarSeparatedRawName
							domClass.superName ,
							innerClassFileNameToDotSeparatedNameMap );
		}

		final FilteredForNonStaticAndSortedForOuterNameLengthReverseAsmDomInnerClasses filteredForNonStaticAndSortedForOuterNameLengthReverseAsmDomInnerClasses =
				AsmInnerClassTypeCorrector.filterForNonStaticAndSortInnerClassesForOuterNameLengthReverse(
						domClass.innerClasses );
		
		final ClassTypeparamsAndInheritTypeArgsTuple classTypeparamsAndInheritTypeArgsTpl =
				AsmClassDescAndSignatureParser.parseClassTypeparamsAndInheritTypeArgsTupel(
						domClass.signature ,
						new HashMap<String, AsmExtendsTypeParam>() ,
						classpathFiles ,
						innerClassFileNameToDotSeparatedNameMap ,
						filteredForNonStaticAndSortedForOuterNameLengthReverseAsmDomInnerClasses );

		final Map<String, AsmExtendsTypeParam> extendsTypeParamMap;
		if ( classTypeparamsAndInheritTypeArgsTpl == null )
		{
			extendsTypeParamMap = Collections.<String, AsmExtendsTypeParam>emptyMap();
		}
		else
		{
			extendsTypeParamMap = classTypeparamsAndInheritTypeArgsTpl.getExtendsTypeParamMap();
		}

		final ImmutableArray<? extends AsmAbstractTypeParam> typeParamsWithTypeAnnotations;
		if ( classTypeparamsAndInheritTypeArgsTpl == null ||
				classTypeparamsAndInheritTypeArgsTpl.typeParams == null ||
				classTypeparamsAndInheritTypeArgsTpl.typeParams.isEmpty() )
		{
			typeParamsWithTypeAnnotations = null;
		}
		else
		{
			final List<AsmAbstractTypeParam> tmpTypeParams = new ArrayList<>();
			for ( int typeParamIndex = 0 ; typeParamIndex < classTypeparamsAndInheritTypeArgsTpl.typeParams.size() ; typeParamIndex++ )
			{
				final AsmAbstractTypeParam classTypeParam = classTypeparamsAndInheritTypeArgsTpl.typeParams.get( typeParamIndex );

				final List<AsmDomTypeAnnotation> currTypeParamTypeAnnotations =
						CollectionUtil.getAtIndexOrNullIfIndexOutOfBound(
								domClass.typeParamTypeAnnotations ,
								typeParamIndex );

				final Map<AsmTypePath, ImmutableArray<AsmAnnotationUse>> convertedTypeParamTypeAnnotations =
						convertTypeAnnotations(
								currTypeParamTypeAnnotations ,
								extendsTypeParamMap ,
								classpathFiles ,
								innerClassFileNameToDotSeparatedNameMap ,
								filteredForNonStaticAndSortedForOuterNameLengthReverseAsmDomInnerClasses );

				if ( classTypeParam instanceof AsmUnboundedTypeParam )
				{
					tmpTypeParams.add(
							new AsmUnboundedTypeParam(
									convertedTypeParamTypeAnnotations
									// einziger value
									//.values().iterator().next()
									.get( AsmTypePath.NULL.INSTANCE ) ,
									classTypeParam.name ) );
				}
				else if ( classTypeParam instanceof AsmExtendsTypeParam )
				{
					final AsmExtendsTypeParam classExtendsTypeParam = (AsmExtendsTypeParam) classTypeParam;

					final AsmType classExtendsTypeParamBoundType = classExtendsTypeParam.boundType;

					final List<List<AsmDomTypeAnnotation>> currTypeParamBoundsTypeAnnotations =
							CollectionUtil.getAtIndexOrNullIfIndexOutOfBound(
									domClass.typeParamBoundTypeAnnotations ,
									typeParamIndex );

					if ( classExtendsTypeParamBoundType instanceof AsmIntersectionBound )
					{
						final AsmIntersectionBound classIntersectionTypeParamBoundType = (AsmIntersectionBound) classExtendsTypeParamBoundType;

						final List<AsmType> intersectedTypesWithTypeAnnotations = new ArrayList<>();
						for ( int intersectedTypeIndex = 0 ; intersectedTypeIndex < classIntersectionTypeParamBoundType.boundTypes.size() ; intersectedTypeIndex++ )
						{
							final AsmType intersectedType = classIntersectionTypeParamBoundType.boundTypes.get( intersectedTypeIndex );

							final List<AsmDomTypeAnnotation> currTypeParamBoundTypeAnnotations =
									CollectionUtil.getNullSafeAtIndexOrNullIfIndexOutOfBound(
											currTypeParamBoundsTypeAnnotations ,
											intersectedTypeIndex );

							final Map<AsmTypePath, ImmutableArray<AsmAnnotationUse>> convertedTypeParamBoundTypeAnnotations =
									convertTypeAnnotations(
											currTypeParamBoundTypeAnnotations ,
											extendsTypeParamMap ,
											classpathFiles ,
											innerClassFileNameToDotSeparatedNameMap ,
											filteredForNonStaticAndSortedForOuterNameLengthReverseAsmDomInnerClasses );

							final AsmType intersectedTypeWithTypeAnnotations =
									intersectedType.cloneAddTypeAnnotations(
											convertedTypeParamBoundTypeAnnotations ,
											AsmTypePath.NULL.INSTANCE );
							
							intersectedTypesWithTypeAnnotations.add(
									intersectedTypeWithTypeAnnotations );
						}

						tmpTypeParams.add(
								new AsmExtendsTypeParam(
										convertedTypeParamTypeAnnotations.get( AsmTypePath.NULL.INSTANCE ) ,
										classTypeParam.name ,
										new AsmIntersectionBound(
												ImmutableArray.create( intersectedTypesWithTypeAnnotations ) ) ) );
					}
					else
					{
						final int typeParamBoundIndex = 0;

						final List<AsmDomTypeAnnotation> currTypeParamBoundTypeAnnotations =
								CollectionUtil.getNullSafeAtIndexOrNullIfIndexOutOfBound(
										currTypeParamBoundsTypeAnnotations ,
										typeParamBoundIndex );

						final Map<AsmTypePath, ImmutableArray<AsmAnnotationUse>> convertedTypeParamBoundTypeAnnotations =
								convertTypeAnnotations(
										currTypeParamBoundTypeAnnotations ,
										extendsTypeParamMap ,
										classpathFiles ,
										innerClassFileNameToDotSeparatedNameMap ,
										filteredForNonStaticAndSortedForOuterNameLengthReverseAsmDomInnerClasses );

						final AsmType classExtendsTypeParamBoundTypeWithTypeAnnotations =
								classExtendsTypeParamBoundType.cloneAddTypeAnnotations(
										convertedTypeParamBoundTypeAnnotations ,
										AsmTypePath.NULL.INSTANCE );

						tmpTypeParams.add(
								new AsmExtendsTypeParam(
										convertedTypeParamTypeAnnotations.get( AsmTypePath.NULL.INSTANCE ) ,
										classTypeParam.name ,
										classExtendsTypeParamBoundTypeWithTypeAnnotations ) );
					}
				}
				else
				{
					throw new IllegalStateException( String.valueOf( classTypeParam ) );
				}
			}

			typeParamsWithTypeAnnotations = ImmutableArray.create( tmpTypeParams );
		}

		final ImmutableArray<? extends ImmutableArray<? extends AsmType>> interfaceTypeArgs;
		if ( classTypeparamsAndInheritTypeArgsTpl == null )
		{
			interfaceTypeArgs = null;
		}
		else
		{
			interfaceTypeArgs = classTypeparamsAndInheritTypeArgsTpl.interfaceTypeArgs;
		}

		final AsmClassType[] interfaceTypes;
		if ( domClass.interfaceRawNames == null )
		{
			interfaceTypes = null;
		}
		else
		{
			interfaceTypes = new AsmClassType[ domClass.interfaceRawNames.length ];

			for ( int interfaceIndex = 0 ; interfaceIndex < domClass.interfaceRawNames.length ; interfaceIndex++ )
			{
				final String interfaceRawName = domClass.interfaceRawNames[ interfaceIndex ];

				final Fqn interfaceFqn =
						InnerOrNestedClassNameResolver.resolve(
								interfaceRawName ,
								innerClassFileNameToDotSeparatedNameMap );

				final AsmClassType interfaceTypeWithoutTypeAnnotations =
						new AsmClassType(
								interfaceFqn ,
								// typeArgs
								( interfaceTypeArgs != null ? interfaceTypeArgs.get( interfaceIndex ) : null ) );

				final List<AsmDomTypeAnnotation> currInterfaceTypeAnnotations =
						CollectionUtil.getAtIndexOrNullIfIndexOutOfBound(
								domClass.interfaceTypeAnnotations ,
								interfaceIndex );

				final Map<AsmTypePath, ImmutableArray<AsmAnnotationUse>> convertedTypeAnnotations =
						convertTypeAnnotations(
								currInterfaceTypeAnnotations ,
								extendsTypeParamMap ,
								classpathFiles ,
								innerClassFileNameToDotSeparatedNameMap ,
								filteredForNonStaticAndSortedForOuterNameLengthReverseAsmDomInnerClasses );

				interfaceTypes[ interfaceIndex ] =
						interfaceTypeWithoutTypeAnnotations.cloneAddTypeAnnotations(
								convertedTypeAnnotations ,
								AsmTypePath.NULL.INSTANCE );
			}
		}

		final ImmutableMap<String, ? extends AsmFieldInfo> fields =
				convertFields(
						domClass.fields ,
						classFqn ,
						extendsTypeParamMap ,
						classpathFiles ,
						innerClassFileNameToDotSeparatedNameMap ,
						filteredForNonStaticAndSortedForOuterNameLengthReverseAsmDomInnerClasses );

		final ImmutableMap<String, ? extends ImmutableArray<? extends AsmMethodInfo>> methods =
				convertMethods(
						domClass.methods ,
						classFqn ,
						extendsTypeParamMap ,
						classpathFiles ,
						innerClassFileNameToDotSeparatedNameMap ,
						filteredForNonStaticAndSortedForOuterNameLengthReverseAsmDomInnerClasses );

		final AsmClassType superType;
		if ( isInterface ||
				classFqn.equals( Fqn.JAVA_LANG_OBJECT_FQN ) )
		{
			superType = null;
		}
		else
		{
			final Map<AsmTypePath, ImmutableArray<AsmAnnotationUse>> convertedTypeAnnotations =
					convertTypeAnnotations(
							domClass.superTypeAnnotations ,
							extendsTypeParamMap ,
							classpathFiles ,
							innerClassFileNameToDotSeparatedNameMap ,
							filteredForNonStaticAndSortedForOuterNameLengthReverseAsmDomInnerClasses );

			final AsmClassType superTypeWithoutTypeAnnotations =
					new AsmClassType(
							superFqn ,
							( classTypeparamsAndInheritTypeArgsTpl == null 
							? null 
									: classTypeparamsAndInheritTypeArgsTpl.superTypeArgs ) );
			
			superType =
					superTypeWithoutTypeAnnotations.cloneAddTypeAnnotations(
							convertedTypeAnnotations ,
							AsmTypePath.NULL.INSTANCE );
		}

		final ImmutableArray<? extends AsmAnnotationUse> annotations =
				convertAnnotations(
						domClass.annotations ,
						extendsTypeParamMap ,
						classpathFiles ,
						innerClassFileNameToDotSeparatedNameMap ,
						filteredForNonStaticAndSortedForOuterNameLengthReverseAsmDomInnerClasses );

		final Fqn enclosingClassFqn;
		final String enclosingMethodName;

		if ( domClass.outerClasses.isEmpty() )
		{
			enclosingClassFqn = null;
			enclosingMethodName = null;
		}
		else
		{
			if ( domClass.outerClasses.size() != 1 )
			{
				throw new IllegalStateException();
			}

			enclosingClassFqn =
					InnerOrNestedClassNameResolver.resolve(
							//dollarSeparatedRawName
							domClass.outerClasses.get( 0 ).owner ,
							innerClassFileNameToDotSeparatedNameMap );

			enclosingMethodName = domClass.outerClasses.get( 0 ).name;
		}

		final AsmClassInfo classInfo =
				new AsmClassInfo(
						classpathFile ,
						classFqn ,
						isPrivate ,
						isProtected ,
						isPublic ,
						isAbstract ,
						isStatic ,
						isFinal ,
						isInterface ,
						isEnum ,
						isAnnotation ,
						//classFqn.isTopLevel() ,
						//isInnerOrNested ,
						typeParamsWithTypeAnnotations ,
						superType ,
						ImmutableArray.<AsmClassType>create( interfaceTypes ) ,
						fields ,
						methods ,
						domClass.sourceFileName ,
						annotations ,
						enclosingClassFqn ,
						enclosingMethodName );

		return classInfo;
	}

	private static AsmDomInnerClass getSelfAsmDomInnerClass(
			final String classRawName ,
			final List<AsmDomInnerClass> innerClasses )
	{
		for ( final AsmDomInnerClass asmDomInnerClass : innerClasses )
		{
			if ( classRawName.equals( asmDomInnerClass.name ) )
			{
				return asmDomInnerClass;
			}
		}
		return null;
	}

	private static ImmutableMap<String, ? extends ImmutableArray<? extends AsmMethodInfo>> convertMethods(
			final Iterable<? extends AsmDomMethod> domMethods ,
			final Fqn declaringClassFqn ,
			// die Type-Parameter einer Methode gelten nur fr die aktuelle Methode, deshalb ? extends
			final Map<String, ? extends AsmExtendsTypeParam> extendsTypeParamMap ,
			final ClasspathFiles classpathFiles ,
			final TreeMap<String, String> innerClassFileNameToDotSeparatedNameMap ,
			final FilteredForNonStaticAndSortedForOuterNameLengthReverseAsmDomInnerClasses innerClasses )
	{
		final PutAddCollMap<String, List<AsmMethodInfo>, AsmMethodInfo> methodsPutAddCollMap =
				new PutAddCollMap<String, List<AsmMethodInfo>, AsmMethodInfo>( new HashMap<>() )
		{
			@Override
			protected List<AsmMethodInfo> createSub()
			{
				return new ArrayList<>();
			}
		};

		for ( final AsmDomMethod domMethod : domMethods )
		{
			final AsmMethodInfo methodInfo =
					convertMethod(
							domMethod ,
							declaringClassFqn ,
							extendsTypeParamMap ,
							classpathFiles ,
							innerClassFileNameToDotSeparatedNameMap ,
							innerClasses );

			methodsPutAddCollMap.putSub(
					methodInfo.name ,
					methodInfo );
		}

		final Map<String, ImmutableArray<? extends AsmMethodInfo>> methodsWithImmutableArray = new HashMap<>();
		for ( final Map.Entry<String, List<AsmMethodInfo>> methodsPerNameEntry : methodsPutAddCollMap.map.entrySet() )
		{
			methodsWithImmutableArray.put(
					methodsPerNameEntry.getKey() ,
					ImmutableArray.<AsmMethodInfo>create( methodsPerNameEntry.getValue() ) );
		}

		return ImmutableMap.<String, ImmutableArray<? extends AsmMethodInfo>>create( methodsWithImmutableArray );
	}

	private static AsmMethodInfo convertMethod(
			final AsmDomMethod domMethod ,
			final Fqn declaringClassFqn ,
			// die Type-Parameter einer Methode gelten nur fr die aktuelle Methode, deshalb ? extends
			final Map<String, ? extends AsmExtendsTypeParam> readOnlyExtendsTypeParamMap ,
			final ClasspathFiles classpathFiles ,
			final TreeMap<String, String> innerClassFileNameToDotSeparatedNameMap ,
			final FilteredForNonStaticAndSortedForOuterNameLengthReverseAsmDomInnerClasses innerClasses )
	{
		final boolean isAbstract =  ( domMethod.access & org.objectweb.asm.Opcodes.ACC_ABSTRACT ) != 0;

		final boolean isFinal =  ( domMethod.access & org.objectweb.asm.Opcodes.ACC_FINAL ) != 0;

		final boolean isPublic = ( domMethod.access & org.objectweb.asm.Opcodes.ACC_PUBLIC ) != 0;

		final boolean isProtected = ( domMethod.access & org.objectweb.asm.Opcodes.ACC_PROTECTED ) != 0;

		final boolean isPrivate = ( domMethod.access & org.objectweb.asm.Opcodes.ACC_PRIVATE ) != 0;

		final boolean isStatic = ( domMethod.access & org.objectweb.asm.Opcodes.ACC_STATIC ) != 0;

		final boolean isSynchronized = ( domMethod.access & org.objectweb.asm.Opcodes.ACC_SYNCHRONIZED ) != 0;

		final boolean isSynthetic = ( domMethod.access & org.objectweb.asm.Opcodes.ACC_SYNTHETIC ) != 0;

		final boolean isNative = ( domMethod.access & org.objectweb.asm.Opcodes.ACC_NATIVE ) != 0;

		final boolean isBridge = ( domMethod.access & org.objectweb.asm.Opcodes.ACC_BRIDGE ) != 0;

		final boolean hasVarargs = ( domMethod.access & org.objectweb.asm.Opcodes.ACC_VARARGS ) != 0;

		// die Type-Parameter einer Methode gelten nur fr die aktuelle Methode
		final Map<String, AsmExtendsTypeParam> extendsTypeParamMap = new HashMap<>( readOnlyExtendsTypeParamMap );

		final MethodTypeparamsAndReturnAndParamsAndExceptionsTuple methodTypeparamsAndReturnAndParamsAndExceptionsTpl =
				AsmMethodDescAndSignatureParser.parseMethodDescriptionAndSignature(
						domMethod.desc ,
						domMethod.signature ,
						domMethod.exceptions ,
						extendsTypeParamMap ,
						classpathFiles ,
						innerClassFileNameToDotSeparatedNameMap ,
						innerClasses );

		final ImmutableArray<? extends AsmAnnotationUse> annotations =
				convertAnnotations(
						domMethod.annotations ,
						extendsTypeParamMap ,
						classpathFiles ,
						innerClassFileNameToDotSeparatedNameMap ,
						innerClasses );

		// falls es sich um eine Methode in einer Annotatuions-Deklaration handelt
		final Object annotationDefaultValue =
				convertAnnotationDefault(
						domMethod.annotationDefault ,
						methodTypeparamsAndReturnAndParamsAndExceptionsTpl.returnType ,
						extendsTypeParamMap ,
						classpathFiles ,
						innerClassFileNameToDotSeparatedNameMap ,
						innerClasses );

		final List<AsmMethodParamInfo> tmpParams = new ArrayList<>();
		if ( methodTypeparamsAndReturnAndParamsAndExceptionsTpl.paramTypes != null )
		{
			for ( int parameterIndex = 0 ; parameterIndex < methodTypeparamsAndReturnAndParamsAndExceptionsTpl.paramTypes.size() ; parameterIndex++ )
			{
				final AsmType paramTypeWithoutTypeAnnotations =
						methodTypeparamsAndReturnAndParamsAndExceptionsTpl.paramTypes.get( parameterIndex );

				final List<AsmDomTypeAnnotation> currFormalParamTypeAnnotations =
						CollectionUtil.getAtIndexOrNullIfIndexOutOfBound(
								domMethod.formalParamTypeAnnotations ,
								parameterIndex );

				final Map<AsmTypePath, ImmutableArray<AsmAnnotationUse>> convertedTypeAnnotations =
						convertTypeAnnotations(
								currFormalParamTypeAnnotations ,
								extendsTypeParamMap ,
								classpathFiles ,
								innerClassFileNameToDotSeparatedNameMap ,
								innerClasses );

				final AsmType paramTypeWithTypeAnnotations =
						paramTypeWithoutTypeAnnotations.cloneAddTypeAnnotations(
								convertedTypeAnnotations ,
								AsmTypePath.NULL.INSTANCE );

				final List<AsmDomAnnotation> domAnnotationList =
						CollectionUtil.getAtIndexOrNullIfIndexOutOfBound(
								domMethod.parameterAnnotationListList ,
								parameterIndex );

				final ImmutableArray<? extends AsmAnnotationUse> paramAnnotations;
				if ( domAnnotationList != null )
				{
					paramAnnotations =
							convertAnnotations(
									domAnnotationList ,
									extendsTypeParamMap ,
									classpathFiles ,
									innerClassFileNameToDotSeparatedNameMap ,
									innerClasses );
				}
				else
				{
					paramAnnotations = null;
				}

				tmpParams.add(
						new AsmMethodParamInfo(
								paramTypeWithTypeAnnotations ,
								paramAnnotations ) );
			}
		}

		final AsmType returnTypeWithoutTypeAnnotations = methodTypeparamsAndReturnAndParamsAndExceptionsTpl.returnType;

		final AsmType returnTypeWithTypeAnnotations;
		if ( returnTypeWithoutTypeAnnotations != null )
		{
			final Map<AsmTypePath, ImmutableArray<AsmAnnotationUse>> convertedTypeAnnotations =
					convertTypeAnnotations(
							domMethod.returnTypeAnnotations ,
							extendsTypeParamMap ,
							classpathFiles ,
							innerClassFileNameToDotSeparatedNameMap ,
							innerClasses );

			returnTypeWithTypeAnnotations =
					returnTypeWithoutTypeAnnotations.cloneAddTypeAnnotations(
							convertedTypeAnnotations ,
							AsmTypePath.NULL.INSTANCE );
		}
		else
		{
			returnTypeWithTypeAnnotations = null;
		}

		final ImmutableArray<? extends AsmTopLevelOrInnerClassTypeOrTypevar> exceptionsWithTypeAnnotations;
		if ( methodTypeparamsAndReturnAndParamsAndExceptionsTpl.exceptions == null ||
				methodTypeparamsAndReturnAndParamsAndExceptionsTpl.exceptions.isEmpty() )
		{
			exceptionsWithTypeAnnotations = null;
		}
		else
		{
			final List<AsmTopLevelOrInnerClassTypeOrTypevar> tmpExceptionWithTypeAnnotations = new ArrayList<>();
			for ( int exceptionIndex = 0 ; exceptionIndex < methodTypeparamsAndReturnAndParamsAndExceptionsTpl.exceptions.size() ; exceptionIndex++ )
			{
				final AsmTopLevelOrInnerClassTypeOrTypevar exceptionWithoutTypeAnnotation = methodTypeparamsAndReturnAndParamsAndExceptionsTpl.exceptions.get( exceptionIndex );

				// temporaere Map zum Erfuellen der Parameter der Methode AsmType#cloneAddTypeAnnotations
				final Map<AsmTypePath, ImmutableArray<AsmAnnotationUse>> typeAnnotationsMap =
						new HashMap<>();

				List<AsmDomTypeAnnotation> currThrowsTypeAnnotations =
						CollectionUtil.getAtIndexOrNullIfIndexOutOfBound(
								domMethod.throwsTypeAnnotations ,
								exceptionIndex );

				Map<AsmTypePath, ImmutableArray<AsmAnnotationUse>> convertedTypeAnnotations =
						convertTypeAnnotations(
								currThrowsTypeAnnotations ,
								extendsTypeParamMap ,
								classpathFiles ,
								innerClassFileNameToDotSeparatedNameMap ,
								innerClasses );

				final ImmutableArray<AsmAnnotationUse> typeAnnotations;
				if ( convertedTypeAnnotations != null && ! convertedTypeAnnotations.isEmpty() )
				{
					typeAnnotations = convertedTypeAnnotations
							// einziger value
							.values().iterator().next();
				}
				else
				{
					typeAnnotations = null;
				}

				typeAnnotationsMap.put(
						AsmTypePath.NULL.INSTANCE ,
						typeAnnotations );

				tmpExceptionWithTypeAnnotations.add(
						(AsmTopLevelOrInnerClassTypeOrTypevar) exceptionWithoutTypeAnnotation.cloneAddTypeAnnotations(
								typeAnnotationsMap ,
								AsmTypePath.NULL.INSTANCE ) );
			}
			exceptionsWithTypeAnnotations =
					ImmutableArray.create( tmpExceptionWithTypeAnnotations );
		}

		final AsmMethodInfo methodInfo =
				new AsmMethodInfo(
						declaringClassFqn ,
						isAbstract ,
						isFinal ,
						isPrivate ,
						isProtected ,
						isPublic ,
						isStatic ,
						isSynchronized ,
						isSynthetic ,
						isNative ,
						isBridge ,
						domMethod.name ,
						domMethod.desc ,
						// TODO hier mssen noch typ parameter dran
						methodTypeparamsAndReturnAndParamsAndExceptionsTpl.typeParams ,
						returnTypeWithTypeAnnotations ,
						ImmutableArray.<AsmMethodParamInfo>create( tmpParams ) ,
						hasVarargs ,
						exceptionsWithTypeAnnotations ,
						annotations ,
						annotationDefaultValue );

		return methodInfo;
	}

	private static ImmutableMap<String, ? extends AsmFieldInfo> convertFields(
			final Collection<? extends AsmDomField> domFields ,
			final Fqn declaringClassFqn ,
			final Map<String, ? extends AsmExtendsTypeParam> extendsTypeParamMap ,
			final ClasspathFiles classpathFiles ,
			final TreeMap<String, String> innerClassFileNameToDotSeparatedNameMap ,
			final FilteredForNonStaticAndSortedForOuterNameLengthReverseAsmDomInnerClasses innerClasses )
	{
		final Map<String, AsmFieldInfo> fields = new HashMap<>();

		for ( final AsmDomField domField : domFields )
		{
			final AsmFieldInfo field =
					convertField(
							domField ,
							declaringClassFqn ,
							extendsTypeParamMap ,
							classpathFiles ,
							innerClassFileNameToDotSeparatedNameMap ,
							innerClasses );

			fields.put(
					field.name ,
					field );
		}

		return ImmutableMap.<String, AsmFieldInfo>create( fields ) ;
	}

	private static AsmFieldInfo convertField(
			final AsmDomField domField ,
			final Fqn declaringClassFqn ,
			final Map<String, ? extends AsmExtendsTypeParam> extendsTypeParamMap ,
			final ClasspathFiles classpathFiles ,
			final TreeMap<String, String> innerClassFileNameToDotSeparatedNameMap ,
			final FilteredForNonStaticAndSortedForOuterNameLengthReverseAsmDomInnerClasses innerClasses )
	{
		final boolean isPublic = ( domField.access & org.objectweb.asm.Opcodes.ACC_PUBLIC ) != 0;

		final boolean isProtected = ( domField.access & org.objectweb.asm.Opcodes.ACC_PROTECTED ) != 0;

		final boolean isPrivate = ( domField.access & org.objectweb.asm.Opcodes.ACC_PRIVATE ) != 0;

		final boolean isStatic = ( domField.access & org.objectweb.asm.Opcodes.ACC_STATIC ) != 0;

		final boolean isFinal = ( domField.access & org.objectweb.asm.Opcodes.ACC_FINAL ) != 0;

		final boolean isTransient = ( domField.access & org.objectweb.asm.Opcodes.ACC_TRANSIENT ) != 0;

		final boolean isVolatile = ( domField.access & org.objectweb.asm.Opcodes.ACC_VOLATILE ) != 0;

		final boolean isSynthetic = ( domField.access & org.objectweb.asm.Opcodes.ACC_SYNTHETIC ) != 0;

		final AsmType fieldTypeWithoutTypeAnnotations =
				AsmFieldDescAndSignatureParser.parseFieldDescriptionAndSignature(
						domField.desc ,
						domField.signature ,
						extendsTypeParamMap ,
						classpathFiles ,
						innerClassFileNameToDotSeparatedNameMap ,
						innerClasses );

		final Map<AsmTypePath, ImmutableArray<AsmAnnotationUse>> convertedTypeAnnotations =
				convertTypeAnnotations(
						domField.typeAnnotations ,
						extendsTypeParamMap ,
						classpathFiles ,
						innerClassFileNameToDotSeparatedNameMap ,
						innerClasses );

		final AsmType fieldTypeWithTypeAnnotations =
				fieldTypeWithoutTypeAnnotations.cloneAddTypeAnnotations(
						convertedTypeAnnotations ,
						AsmTypePath.NULL.INSTANCE );

		final ImmutableArray<? extends AsmAnnotationUse> annotations =
				convertAnnotations(
						domField.annotations ,
						extendsTypeParamMap ,
						classpathFiles ,
						innerClassFileNameToDotSeparatedNameMap ,
						innerClasses );

		final AsmFieldInfo fieldInfo =
				new AsmFieldInfo(
						declaringClassFqn ,
						domField.name ,
						isPrivate ,
						isProtected ,
						isPublic ,
						isStatic ,
						isFinal ,
						isTransient ,
						isVolatile ,
						isSynthetic ,
						fieldTypeWithTypeAnnotations ,
						domField.initialValue ,
						annotations );

		return fieldInfo;
	}

	private static ImmutableArray<? extends AsmAnnotationUse> convertAnnotations(
			final Collection<? extends AsmDomAnnotation> domAnnotations ,
			final Map<String, ? extends AsmExtendsTypeParam> extendsTypeParamMap ,
			final ClasspathFiles classpathFiles ,
			final TreeMap<String, String> innerClassFileNameToDotSeparatedNameMap ,
			final FilteredForNonStaticAndSortedForOuterNameLengthReverseAsmDomInnerClasses innerClasses )
	{
		final List<AsmAnnotationUse> convertedAnnotations = new ArrayList<>();

		for ( final AsmDomAnnotation domAnnotation : domAnnotations )
		{
			convertedAnnotations.add(
					convertAnnotation(
							domAnnotation ,
							extendsTypeParamMap ,
							classpathFiles ,
							innerClassFileNameToDotSeparatedNameMap ,
							innerClasses ) );
		}

		if ( convertedAnnotations .isEmpty() )
		{
			// Speicher sparen
			return null;
		}
		
		return ImmutableArray.<AsmAnnotationUse>create( convertedAnnotations );
	}

	private static AsmAnnotationUse convertAnnotation(
			final AsmDomAnnotation domAnnotation ,
			final Map<String, ? extends AsmExtendsTypeParam> extendsTypeParamMap ,
			final ClasspathFiles classpathFiles ,
			final TreeMap<String, String> innerClassFileNameToDotSeparatedNameMap ,
			final FilteredForNonStaticAndSortedForOuterNameLengthReverseAsmDomInnerClasses innerClasses )
	{
		final AsmClassType annotationType =
				(AsmClassType) AsmDescAndSignatureParserUtil.parseTypeStr(
						domAnnotation.desc ,
						extendsTypeParamMap ,
						classpathFiles ,
						innerClassFileNameToDotSeparatedNameMap ,
						innerClasses );

		final Map<String, Object> annotationValues = new HashMap<>();
		for ( final AsmDomAnnotationValue annotationValue : domAnnotation.values )
		{
			annotationValues.put(
					annotationValue.name ,
					convertAnnotationValue(
							annotationValue ,
							extendsTypeParamMap ,
							classpathFiles ,
							innerClassFileNameToDotSeparatedNameMap ,
							innerClasses ) );
		}

		final ImmutableMap<String, Object> immutableAnnotationValues =
				ImmutableMap.create( annotationValues );

		return new AsmAnnotationUse(
				annotationType ,
				immutableAnnotationValues );
	}

	private static Object convertAnnotationValue(
			final AsmDomAnnotationValue domAnnotationValue ,
			final Map<String, ? extends AsmExtendsTypeParam> extendsTypeParamMap ,
			final ClasspathFiles classpathFiles ,
			final TreeMap<String, String> innerClassFileNameToDotSeparatedNameMap ,
			final FilteredForNonStaticAndSortedForOuterNameLengthReverseAsmDomInnerClasses innerClasses )
	{
		if ( domAnnotationValue instanceof AsmDomAnnotationPrimitiveValue )
		{
			final AsmDomAnnotationPrimitiveValue domAnnotationPrimitiveValue = (AsmDomAnnotationPrimitiveValue) domAnnotationValue;

			if ( domAnnotationPrimitiveValue.value == null )
			{
				throw new IllegalStateException();
			}

			if ( domAnnotationPrimitiveValue.value instanceof org.objectweb.asm.Type )
			{
				return AsmDescAndSignatureParserUtil.parseTypeStr(
						// TODO hier wird ein Objekt org.objectweb.asm.Type geliefert, knnte man direkt(ohne Umwandlung in String) in AsmType umwandeln
						String.valueOf( domAnnotationPrimitiveValue.value ) ,
						extendsTypeParamMap ,
						classpathFiles ,
						innerClassFileNameToDotSeparatedNameMap ,
						innerClasses );
			}

			return domAnnotationPrimitiveValue.value;
		}

		if ( domAnnotationValue instanceof AsmDomAnnotationEnumValue )
		{
			final AsmClassType enumType =
					(AsmClassType) AsmDescAndSignatureParserUtil.parseTypeStr(
							( (AsmDomAnnotationEnumValue) domAnnotationValue ).desc ,
							extendsTypeParamMap ,
							classpathFiles ,
							innerClassFileNameToDotSeparatedNameMap ,
							innerClasses );

			final AsmEnumValueUse enumValue =
					new AsmEnumValueUse(
							enumType ,
							( (AsmDomAnnotationEnumValue) domAnnotationValue ).value );

			return enumValue;

		}

		if ( domAnnotationValue instanceof AsmDomArray )
		{
			final AsmDomArray asmDomArray = (AsmDomArray) domAnnotationValue;

			final List<Object> arrayAnnotationValueList = new ArrayList<>();
			for ( final AsmDomAnnotationValue arrayDomAnnotationValue : asmDomArray.values )
			{
				arrayAnnotationValueList.add(
						convertAnnotationValue( 
								arrayDomAnnotationValue ,
								extendsTypeParamMap ,
								classpathFiles ,
								innerClassFileNameToDotSeparatedNameMap ,
								innerClasses ) );
			}

			return arrayAnnotationValueList.toArray();
		}

		if ( domAnnotationValue instanceof AsmDomAnnotationAnnotationValue )
		{
			final AsmDomAnnotationAnnotationValue asmDomAnnotationAnnotationValue = (AsmDomAnnotationAnnotationValue) domAnnotationValue;

			final AsmClassType annotationType =
					(AsmClassType) AsmDescAndSignatureParserUtil.parseTypeStr(
							asmDomAnnotationAnnotationValue.desc ,
							extendsTypeParamMap ,
							classpathFiles ,
							innerClassFileNameToDotSeparatedNameMap ,
							innerClasses );

			final Map<String, Object> annotationValues = new HashMap<>();
			for ( final AsmDomAnnotationValue annotationValue : asmDomAnnotationAnnotationValue.values )
			{
				annotationValues.put(
						annotationValue.name ,
						convertAnnotationValue(
								annotationValue ,
								extendsTypeParamMap ,
								classpathFiles ,
								innerClassFileNameToDotSeparatedNameMap ,
								innerClasses ) );
			}

			final ImmutableMap<String, Object> immutableAnnotationValues =
					ImmutableMap.create( annotationValues );

			return new AsmAnnotationUse(
					annotationType ,
					immutableAnnotationValues );
		}

		throw new IllegalArgumentException( String.valueOf( domAnnotationValue ) );
	}

	/**
	 * 
	 * @param domAnnotationDefault
	 * @param annotationMethodReturnType zustzliche Information zur korrekten Rckgabe des Default-Wertes
	 * @param innerClassFileNameToDotSeparatedNameMap
	 * @return
	 */
	private static Object convertAnnotationDefault(
			final AsmDomAnnotationDefault domAnnotationDefault ,
			final AsmType annotationMethodReturnType ,
			final Map<String, ? extends AsmExtendsTypeParam> extendsTypeParamMap ,
			final ClasspathFiles classpathFiles ,
			final TreeMap<String, String> innerClassFileNameToDotSeparatedNameMap ,
			final FilteredForNonStaticAndSortedForOuterNameLengthReverseAsmDomInnerClasses innerClasses )
	{
		if ( domAnnotationDefault == null )
		{
			return null;
		}

		if ( domAnnotationDefault.domPrimitiveValue != null )
		{
			if ( domAnnotationDefault.domPrimitiveValue.value == null )
			{
				throw new IllegalStateException();
			}

			if ( domAnnotationDefault.domPrimitiveValue.value instanceof org.objectweb.asm.Type )
			{
				return AsmDescAndSignatureParserUtil.parseTypeStr(
						// TODO hier wird ein Objekt org.objectweb.asm.Type geliefert, knnte man direkt(ohne Umwandlung in String) in AsmType umwandeln
						String.valueOf( domAnnotationDefault.domPrimitiveValue.value ) ,
						extendsTypeParamMap ,
						classpathFiles ,
						innerClassFileNameToDotSeparatedNameMap ,
						innerClasses );
			}

			return domAnnotationDefault.domPrimitiveValue.value;
		}

		if ( domAnnotationDefault.domAnnotationAnnotationValue != null )
		{
			return 
					convertAnnotationValue(
							domAnnotationDefault.domAnnotationAnnotationValue ,
							extendsTypeParamMap ,
							classpathFiles ,
							innerClassFileNameToDotSeparatedNameMap ,
							innerClasses );
		}

		if ( domAnnotationDefault.domArray != null )
		{
			if ( domAnnotationDefault.domArray.values.isEmpty() )
			{
				//System.out.println( ( new Exception() ).getStackTrace()[ 0 ] + " " + annotationMethodReturnType );

				final AsmArray annotationMethodReturnArrayType = (AsmArray) annotationMethodReturnType;

				// nur bei leerem Array wird diese Methode von ASM aufgerufen, bei nicht-leerem Array wird de.heinerkuecker.asm_improved.AnnotationDefaultAnnotationVisitor.visit(String, Object) mit einem primitiven Array aufgerufen
				if ( annotationMethodReturnArrayType.elementType.equals( AsmBoolean.INSTANCE ) )
				{
					return EmptyArrays.BOOLEAN;
				}
				if ( annotationMethodReturnArrayType.elementType.equals( AsmByte.INSTANCE ) )
				{
					return EmptyArrays.BYTE;
				}
				if ( annotationMethodReturnArrayType.elementType.equals( AsmChar.INSTANCE ) )
				{
					return EmptyArrays.CHAR;
				}
				if ( annotationMethodReturnArrayType.elementType.equals( AsmShort.INSTANCE ) )
				{
					return EmptyArrays.SHORT;
				}
				if ( annotationMethodReturnArrayType.elementType.equals( AsmInt.INSTANCE ) )
				{
					return EmptyArrays.INT;
				}
				if ( annotationMethodReturnArrayType.elementType.equals( AsmLong.INSTANCE ) )
				{
					return EmptyArrays.LONG;
				}
			}

			return 
					convertAnnotationValue(
							domAnnotationDefault.domArray ,
							extendsTypeParamMap ,
							classpathFiles ,
							innerClassFileNameToDotSeparatedNameMap ,
							innerClasses );
		}

		if ( domAnnotationDefault.domEnumValue != null )
		{
			return 
					convertAnnotationValue(
							domAnnotationDefault.domEnumValue ,
							extendsTypeParamMap ,
							classpathFiles ,
							innerClassFileNameToDotSeparatedNameMap ,
							innerClasses );
		}

		throw new IllegalArgumentException( String.valueOf( domAnnotationDefault ) );
	}

	private static Map<AsmTypePath, ImmutableArray<AsmAnnotationUse>> convertTypeAnnotations(
			final List<AsmDomTypeAnnotation> domTypeAnnotations ,
			final Map<String, ? extends AsmExtendsTypeParam> extendsTypeParamMap ,
			final ClasspathFiles classpathFiles ,
			final TreeMap<String, String> innerClassFileNameToDotSeparatedNameMap ,
			final FilteredForNonStaticAndSortedForOuterNameLengthReverseAsmDomInnerClasses innerClasses )
	{
		final ArraylistHashmap<AsmTypePath, AsmAnnotationUse> tmpListMap = new ArraylistHashmap<>();

		if ( domTypeAnnotations != null )
		{
			for ( final AsmDomTypeAnnotation domTypeAnnotation : domTypeAnnotations )
			{
				final AsmAnnotationUse convertedTypeAnnotation =
						convertTypeAnnotation(
								domTypeAnnotation ,
								extendsTypeParamMap ,
								classpathFiles ,
								innerClassFileNameToDotSeparatedNameMap ,
								innerClasses );

				tmpListMap.putListElement(
						domTypeAnnotation.typePath ,
						convertedTypeAnnotation );
			}
		}

		final Map<AsmTypePath, ImmutableArray<AsmAnnotationUse>> result = new HashMap<>();

		for ( final Entry<AsmTypePath, ArrayList<AsmAnnotationUse>> tmpListMapEntry : tmpListMap.entrySet() )
		{
			result.put(
					tmpListMapEntry.getKey() ,
					ImmutableArray.create(
							tmpListMapEntry.getValue() ) );
		}

		return result;
	}

	private static AsmAnnotationUse convertTypeAnnotation(
			final AsmDomTypeAnnotation domTypeAnnotation ,
			final Map<String, ? extends AsmExtendsTypeParam> extendsTypeParamMap ,
			final ClasspathFiles classpathFiles ,
			final TreeMap<String, String> innerClassFileNameToDotSeparatedNameMap ,
			final FilteredForNonStaticAndSortedForOuterNameLengthReverseAsmDomInnerClasses innerClasses )
	{
		final AsmClassType annotationType =
				(AsmClassType) AsmDescAndSignatureParserUtil.parseTypeStr(
						domTypeAnnotation.desc ,
						extendsTypeParamMap ,
						classpathFiles ,
						innerClassFileNameToDotSeparatedNameMap ,
						innerClasses );

		final Map<String, Object> annotationValues = new HashMap<>();
		for ( final AsmDomAnnotationValue annotationValue : domTypeAnnotation.values )
		{
			annotationValues.put(
					annotationValue.name ,
					convertAnnotationValue(
							annotationValue ,
							extendsTypeParamMap ,
							classpathFiles ,
							innerClassFileNameToDotSeparatedNameMap ,
							innerClasses ) );
		}

		final ImmutableMap<String, Object> immutableAnnotationValues =
				ImmutableMap.create(
						annotationValues );

		return new AsmAnnotationUse(
				annotationType ,
				immutableAnnotationValues );
	}

	@SuppressWarnings("serial")
	private static class ArraylistHashmap<K, V>
	extends HashMap<K, ArrayList<V>>
	{

		/* (non-Javadoc)
		 * @see java.util.HashMap#put(java.lang.Object, java.lang.Object)
		 */
		public void putListElement(
				final K key ,
				V value )
		{
			ArrayList<V> tmpList = get( key );

			if ( tmpList == null )
			{
				tmpList = new ArrayList<>();

				put(
						key ,
						tmpList );
			}

			tmpList.add( value );
		}

	}
}
