package de.heinerkuecker.asm_improved;

import de.heinerkuecker.asm_improved.annotation_use.AsmAnnotationUse;
import de.heinerkuecker.asm_improved.type.AsmType;
import de.heinerkuecker.asm_improved.util.Fqn;
import de.heinerkuecker.asm_improved.util.ImmutableArray;

/**
 * Informationen eines Felds.
 */
public final class AsmFieldInfo
{
	/**
	 * hier steht der vollqualifizierte Name der Klasse/Interface/Enum/Annotation,
	 * in welcher dieses Feld deklariert ist, statt die
	 * {@link AsmClassInfo}, damit eventuell nicht so viele
	 * Objekte im Speicher gehalten werden mssen.
	 */
	public final Fqn declaringClassFqn;

	public final String name;

	public final boolean isPrivate;

	public final boolean isProtected;

	public final boolean isPublic;

	public final boolean isStatic;

	public final boolean isFinal;

	public final boolean isTransient;

	public final boolean isVolatile;

	public final boolean isSynthetic;

	public final AsmType type;

	/**
	 * Annotationen.
	 */
	public final ImmutableArray<? extends AsmAnnotationUse> annotations;

	/**
     * The field's initial value. This parameter, which may be
     * <tt>null</tt> if the field does not have an initial value,
     * must be an {@link Integer}, a {@link Float}, a {@link Long}, a
     * {@link Double} or a {@link String} (for <tt>int</tt>,
     * <tt>float</tt>, <tt>long</tt> or <tt>String</tt> fields
     * respectively). <i>This parameter is only used for static
     * fields</i>. Its value is ignored for non static fields, which
     * must be initialized through bytecode instructions in
     * constructors or methods.
     * <br/>
     * See method {@link org.objectweb.asm.ClassVisitor#visitField(int, String, String, String, Object)} parameter value.
	 */
	public final Object initialValue;
	
	/**
	 * Konstruktor.
	 */
    AsmFieldInfo(
    		final Fqn declaringClassFqn ,
    		final String name ,
            boolean isPrivate ,
            boolean isProtected ,
            boolean isPublic ,
            final boolean isStatic ,
            final boolean isFinal ,
            final boolean isTransient ,
            final boolean isVolatile ,
            final boolean isSynthetic ,
            final AsmType type ,
            final Object initialValue ,
			final ImmutableArray<? extends AsmAnnotationUse> annotations )
    {
    	if ( declaringClassFqn == null )
    	{
    		throw new IllegalArgumentException(
    				"declaringClassFqn is null" );
    	}
    	this.declaringClassFqn = declaringClassFqn;
    	this.name = name;
        this.isPrivate = isPrivate;
        this.isProtected = isProtected;
        this.isPublic = isPublic;
        this.isStatic = isStatic;
        this.isFinal = isFinal;
        this.isTransient = isTransient;
        this.isVolatile = isVolatile;
        this.isSynthetic = isSynthetic;
        this.type = type;
        this.initialValue = initialValue;
		this.annotations = annotations;
    }

	/**
	 * @return default/package scope
	 */
	public boolean isDefault()
	{
		return ! ( this.isPublic || this.isProtected || this.isPrivate );
	}

	/**
     * {@inheritDoc}
     */
    @Override
    public String toString()
    {
    	//return
    	//		"FieldInfo [" +
    	//        "name=" + name + ", " +
    	//        "isPrivate=" + isPrivate + ", " +
    	//		"isProtected=" + isProtected + ", " +
    	//		"isPublic=" + isPublic + ", " +
    	//		"isStatic=" + isStatic + ", " +
    	//		"isFinal=" + isFinal + ", " +
    	//		"type=" + type +
    	//		"]";
        return
        		( isPrivate ? "private " : "" ) +
        		( isProtected ? "protected " : "" ) +
        		( isPublic ? "public " : "" ) +
        		( isStatic ? "static " : "" ) +
        		( isFinal ? "final " : "" ) +
        		type + " " +
        		name +
        		( initialValue != null ? " = " + initialValue : "" ) +
        		( this.annotations != null && ( ! this.annotations.isEmpty() ) ? ", annotations: " +  this.annotations : "" );
    }

}