package de.heinerkuecker.asm_improved;

import de.heinerkuecker.asm_improved.annotation_use.AsmAnnotationUse;
import de.heinerkuecker.asm_improved.annotation_use.AsmEnumValueUse;
import de.heinerkuecker.asm_improved.type.AsmTopLevelOrInnerClassTypeOrTypevar;
import de.heinerkuecker.asm_improved.type.AsmType;
import de.heinerkuecker.asm_improved.type.generics.param.AsmAbstractTypeParam;
import de.heinerkuecker.asm_improved.type.util.AsmTypeUtil;
import de.heinerkuecker.asm_improved.type.util.InheritTypeArgsUtil;
import de.heinerkuecker.asm_improved.type.util.InheritTypeArgsUtil.SuperClassesOrInterfaceInheritBoundReinforcedTypeArgMapOption;
import de.heinerkuecker.asm_improved.util.Fqn;
import de.heinerkuecker.asm_improved.util.ImmutableArray;
import de.heinerkuecker.asm_improved.util.ImmutableMap;

/**
 * Informationen zu einer Methode.
 * 
 * @author Heiner K&uuml;cker
 */
public class AsmMethodInfo
{
	/**
	 * Spezieller Methoden-Name fr Konstruktoren
	 * in ASM (im Class-File)
	 */
	public static final String CONSTRUCTOR_METHOD_NAME = "<init>";

	/**
	 * hier steht der vollqualifizierte Name der Klasse/Interface/Enum/Annotation,
	 * in welcher diese Methode deklariert ist, statt die
	 * {@link AsmClassInfo}, damit eventuell nicht so viele
	 * Objekte im Speicher gehalten werden mssen.
	 */
	public final Fqn declaringClassFqn;

	public final boolean isAbstract;

	public final boolean isFinal;

	public final boolean isPrivate;

	public final boolean isProtected;

	public final boolean isPublic;

	public final boolean isStatic;

	public final boolean isSynchronized;

	public final boolean isSynthetic;

	public final boolean isNative;

	public final boolean isBridge;

    public final String name;

    public final String desc;

	public final ImmutableArray<? extends AsmAbstractTypeParam> typeParams;

	/**
	 * Return-Typ der Methode.
	 * <code>null</code> bedeutet void.
	 */
	public final AsmType returnType;

	public final ImmutableArray<? extends AsmMethodParamInfo> params;

    public final boolean hasVarargs;

    public final ImmutableArray<? extends AsmTopLevelOrInnerClassTypeOrTypevar> exceptions;

    /**
     * Default-Value fr Methoden von Annotationen.
     */
    public final Object annotationDefaultValue;

	/**
	 * Annotationen.
	 */
	public final ImmutableArray<? extends AsmAnnotationUse> annotations;
    
    /**
     * Konstruktor.
     * 
	 * @param isPrivate
	 * @param isProtected
	 * @param isPublic
	 * @param name
	 * @param desc
	 */
    AsmMethodInfo(
    		final Fqn declaringClassFqn ,
    		final boolean isAbstract ,
    		final boolean isFinal ,
            boolean isPrivate ,
            boolean isProtected ,
            boolean isPublic ,
            final boolean isStatic ,
            final boolean isSynchronized ,
            final boolean isSynthetic ,
            final boolean isNative ,
            final boolean isBridge ,
            String name ,
            String desc ,
            final ImmutableArray<? extends AsmAbstractTypeParam> typeParams ,
            final AsmType returnType ,
            final ImmutableArray<? extends AsmMethodParamInfo> params ,
            final boolean hasVarargs ,
            final ImmutableArray<? extends AsmTopLevelOrInnerClassTypeOrTypevar> exceptions ,
			final ImmutableArray<? extends AsmAnnotationUse> annotations ,
			final Object annotationDefaultValue )
    {
    	if ( declaringClassFqn == null )
    	{
    		throw new IllegalArgumentException(
    				"declaringClassFqn is null" );
    	}
    	this.declaringClassFqn = declaringClassFqn;
    	this.isAbstract = isAbstract;
    	this.isFinal = isFinal;
        this.isPrivate = isPrivate;
        this.isProtected = isProtected;
        this.isPublic = isPublic;
        this.isStatic = isStatic;
        this.isSynchronized = isSynchronized;
        this.isSynthetic = isSynthetic;
        this.isNative = isNative;
        this.isBridge = isBridge;
        this.name = name;
        this.desc = desc;
        this.typeParams = typeParams;
        this.returnType = returnType;
        this.params = params;
        this.hasVarargs = hasVarargs;
        this.exceptions = exceptions;
        this.annotations = annotations;

        if ( annotationDefaultValue != null &&
        		! ( annotationDefaultValue instanceof AsmType ||
        				annotationDefaultValue instanceof String ||
        				annotationDefaultValue instanceof Character ||
        				annotationDefaultValue instanceof Boolean ||
        				annotationDefaultValue instanceof Byte ||
        				annotationDefaultValue instanceof Short ||
        				annotationDefaultValue instanceof Integer ||
        				annotationDefaultValue instanceof Long ||
        				annotationDefaultValue instanceof Float ||
        				annotationDefaultValue instanceof Double ||

        				annotationDefaultValue instanceof char[] ||
        				annotationDefaultValue instanceof boolean[] ||
        				annotationDefaultValue instanceof byte[] ||
        				annotationDefaultValue instanceof short[] ||
        				annotationDefaultValue instanceof int[] ||
        				annotationDefaultValue instanceof long[] ||
        				annotationDefaultValue instanceof float[] ||
        				annotationDefaultValue instanceof double[] ||

        				annotationDefaultValue instanceof Object[] ||
        				annotationDefaultValue instanceof AsmAnnotationUse ||
		        		annotationDefaultValue instanceof AsmEnumValueUse ) ) 
        {
			throw new IllegalArgumentException(
					String.valueOf( annotationDefaultValue ) +
					( annotationDefaultValue != null ? " " + annotationDefaultValue.getClass() : "" ) );
		}
        
        if ( annotationDefaultValue instanceof Object[] )
        {
			final Object[] annotationDefaultValueArr = (Object[]) annotationDefaultValue;

			for ( final Object annotationDefaultValueArrValue : annotationDefaultValueArr)
			{
		        if ( ! ( annotationDefaultValueArrValue instanceof AsmType ||
		        		annotationDefaultValueArrValue instanceof String ||
		        		annotationDefaultValueArrValue instanceof Character ||
		        		annotationDefaultValueArrValue instanceof Boolean ||
		        		annotationDefaultValueArrValue instanceof Byte ||
		        		annotationDefaultValueArrValue instanceof Short ||
		        		annotationDefaultValueArrValue instanceof Integer ||
		        		annotationDefaultValueArrValue instanceof Long ||
		        		annotationDefaultValueArrValue instanceof Float ||
		        		annotationDefaultValueArrValue instanceof Double ||
		        		annotationDefaultValueArrValue instanceof AsmAnnotationUse ||
		        		annotationDefaultValueArrValue instanceof AsmEnumValueUse ) ) 
		        {
					throw new IllegalArgumentException(
							String.valueOf( annotationDefaultValueArrValue ) +
							( annotationDefaultValueArrValue != null ? " " + annotationDefaultValueArrValue.getClass() : "" ) );
				}
			}
        }
        
        this.annotationDefaultValue = annotationDefaultValue;
    }

	/**
	 * @return count of parameters
	 */
    public int getParamCount()
    {
    	if ( this.params == null )
        {
            return 0;
        }

    	return this.params.size();
    }

	/**
	 * @return default/package scope
	 */
	public boolean isDefault()
	{
		return ! ( this.isPublic || this.isProtected || this.isPrivate );
	}

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString()
    {
    	//return "MethodInfo [declaringClassPackageName="
    	//        + declaringClassPackageName + ", declaringSimpleClassName="
    	//        + declaringSimpleClassName + ", isAbstract=" + isAbstract
    	//        + ", isPrivate=" + isPrivate + ", isProtected="
    	//        + isProtected + ", isPublic=" + isPublic + ", name=" + name
    	//        + ", desc=" + desc + ", typeParams=" + typeParams
    	//        + ", returnType=" + returnType + ", paramTypes="
    	//        + paramTypes + ", hasVarargs=" + hasVarargs + "]";
    	return
    			( this.typeParams != null && ( ! this.typeParams.isEmpty() ) ? String.valueOf( this.typeParams ) + " " : "" ) +
    			( this.returnType != null ? String.valueOf( this.returnType ) : "void" ) + " " +
    			this.declaringClassFqn + "#" +
    			this.name +
    			( this.params != null ? String.valueOf( this.params ) : "[]" );
    }

    public AsmMethodInfo resolveBounds(
			final AsmClassInfo inheritClassInfo ,
			final ImmutableMap<String, ? extends AsmType> inheritanceTypeArgs ,
			//final ImmutableMap<String, ? extends AsmType> inheritanceTypeParamNameToTypeArgMap ,
			final AsmClasspathInfo asmClasspathInfo
    		// Typ-Argumente der konkreten Instanz dieser Klasse
    		//final ImmutableMap<String, ? extends AsmType> typeargsNameToBoundMap
    		)
    {
    	final AsmClassInfo typeParamHavingClassInfo =
    			asmClasspathInfo.getClassInfoMandantory(
    					declaringClassFqn );

    	final SuperClassesOrInterfaceInheritBoundReinforcedTypeArgMapOption superClassesOrInterfaceInheritBoundReinforcedTypeArgMapOption =
    			InheritTypeArgsUtil.getSuperClassesOrInterfaceInheritBoundReinforcedTypeArgMap(
				inheritClassInfo ,
				inheritanceTypeArgs ,
				typeParamHavingClassInfo ,
				asmClasspathInfo );

    	final ImmutableMap<String, ? extends AsmType> typeargsNameToBoundMap;
    	if ( superClassesOrInterfaceInheritBoundReinforcedTypeArgMapOption.isPresent() )
    	{
			typeargsNameToBoundMap = superClassesOrInterfaceInheritBoundReinforcedTypeArgMapOption.getResult();
		}
    	else
    	{
			typeargsNameToBoundMap = null;
		}

    	final AsmType boundResolvedReturnType;
    	if ( returnType == null )
    	{
    		boundResolvedReturnType = null;
    	}
    	else
    	{
    		boundResolvedReturnType =
    				returnType.resolveBound(
    						typeargsNameToBoundMap );
    	}

    	final ImmutableArray<AsmMethodParamInfo> resolvedParamBounds =
    			AsmTypeUtil.resolveParamBounds(
    					params ,
    					typeargsNameToBoundMap );

    	return new AsmMethodInfo(
    			declaringClassFqn,
    			isAbstract,
    			isFinal,
    			isPrivate,
    			isProtected,
    			isPublic,
    			isStatic,
    			isSynchronized, 
    			isSynthetic,
    			isNative, 
    			isBridge,
    			this.name,
    			desc, 
    			typeParams,
    			boundResolvedReturnType ,
    			resolvedParamBounds ,
    			hasVarargs, 
    			exceptions, 
    			annotations, 
    			annotationDefaultValue );
    }

}