package de.heinerkuecker.asm_improved;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import de.heinerkuecker.asm_improved.type.AsmTopLevelOrInnerClassType;
import de.heinerkuecker.asm_improved.type.AsmClassType;
import de.heinerkuecker.asm_improved.type.AsmType;
import de.heinerkuecker.asm_improved.type.generics.param.AsmAbstractTypeParam;
import de.heinerkuecker.asm_improved.type.generics.param.AsmExtendsTypeParam;
import de.heinerkuecker.asm_improved.type.generics.param.AsmIntersectionBound;
import de.heinerkuecker.asm_improved.type.generics.param.AsmUnboundedTypeParam;
import de.heinerkuecker.asm_improved.type.generics.var.AsmTypevar;
import de.heinerkuecker.asm_improved.type.member.AsmInnerClassType;
import de.heinerkuecker.asm_improved.util.DottedName;
import de.heinerkuecker.asm_improved.util.Fqn;
import de.heinerkuecker.asm_improved.util.ImmutableArray;
import de.heinerkuecker.asm_improved.util.StringUtil;

/**
 * Parser f&uuml;r Type-Variable.
 * 
 * @author Heiner K&uuml;cker
 */
final class AsmTypevarTypeParser
{
	/**
	 * Konstruktor private,
	 * Instanziieren verboten,
	 * Util-Klasse.
	 */
	private AsmTypevarTypeParser()
	{
		super();
	}

	static AsmType parseTypvar(
			final String typeStr ,
			final Map<String, ? extends AsmExtendsTypeParam> extendsTypeParamMap ,
			final ClasspathFiles classpathFiles ,
			final TreeMap<String, ? extends String> innerClassFileNameToDotSeparatedNameMap ,
			final FilteredForNonStaticAndSortedForOuterNameLengthReverseAsmDomInnerClasses innerClasses )
	{
	    if ( ! typeStr.startsWith( "T" ) )
	    {
	        throw new IllegalArgumentException( typeStr );
	    }

		final String typeStrWithoutStartingTypeChar =
				typeStr.substring(
						//beginIndex: T ueberspringen
						1 ,
						//endIndex: endendes Semikolon weglassen
						typeStr.length() - 1 );

		AsmType result = null;

		final List<? extends TypeNameAndTypeArgsStrTuple> separatedInnerOrNestedClassParts =
				separateInnerOrNestedClassParts(
						typeStrWithoutStartingTypeChar );

		final List<? extends TypeNameAndTypeArgsStrTuple> separatedInnerOrNestedClassButJoinedIfNoTypeargsParts =
				joinInnerOrNestedClassPartsWithoutTypeArgs(
						separatedInnerOrNestedClassParts );

		// der Name der Typ-Variable enthlt einen Punkt, scheinbar ist der Typ ein inner/nested type of type variable
		// Testfall: TT.InnerIf<Ljava/lang/Integer;>;
		// Testfall: TT.InnerIf.InnerInnerIf<Ljava/lang/Integer;>;
		for ( TypeNameAndTypeArgsStrTuple part : separatedInnerOrNestedClassButJoinedIfNoTypeargsParts )
		{
			if ( result == null )
			{
				// es gab noch keinen outer/enclosing type
				result =
						new AsmTypevar(
								part.typeName );
			}
			else
			{
				final ImmutableArray<? extends AsmType> typeArgs;
				if ( part.typeArgsStr != null )
				{
					typeArgs =
							AsmTypeargsTypeParser.parseTypeArgs(
									part.typeArgsStr ,
									extendsTypeParamMap ,
									classpathFiles ,
									innerClassFileNameToDotSeparatedNameMap ,
									innerClasses );
				}
				else
				{
					typeArgs = null;
				}

				if ( result instanceof AsmTypevar )
				{
					final AsmExtendsTypeParam extendsTypeParam = extendsTypeParamMap.get( ( (AsmTypevar) result ).name );

					final AsmType boundType = extendsTypeParam.boundType;

					final AsmTopLevelOrInnerClassType matchingRelativType;
					if ( boundType instanceof AsmIntersectionBound )
					{
						final AsmIntersectionBound intersectionBound = (AsmIntersectionBound) boundType;

						final AsmTopLevelOrInnerClassType matchingRelativTypeOfIntersectionBound = 
								resolveIntersectionBoundOfTypevarRelativType(
										part ,
										intersectionBound ,
										typeArgs ,
										classpathFiles );

						matchingRelativType = matchingRelativTypeOfIntersectionBound;
					}
					else if ( boundType instanceof AsmTypevar )
					{
						final AsmTypevar boundTypevar = (AsmTypevar) boundType;

						final AsmExtendsTypeParam resolvedBoundTypeparam =
								// nur auf AsmExtendsTypeParam kann sinnvoll eine innere/geschachtelte Klasse angegeben werden
								(AsmExtendsTypeParam) resolveBoundTypevar(
										extendsTypeParamMap ,
										boundTypevar );

						if ( resolvedBoundTypeparam.boundType instanceof AsmIntersectionBound )
						{
							final AsmIntersectionBound intersectionBound = (AsmIntersectionBound) resolvedBoundTypeparam.boundType;

							final AsmTopLevelOrInnerClassType matchingRelativTypeOfIntersectionBound = 
									resolveIntersectionBoundOfTypevarRelativType(
											part ,
											intersectionBound ,
											typeArgs ,
											classpathFiles );

							matchingRelativType = matchingRelativTypeOfIntersectionBound;
						}
						else
						{
							if ( resolvedBoundTypeparam.boundType instanceof AsmClassType )
							{
								final AsmClassType classBoundType = (AsmClassType) resolvedBoundTypeparam.boundType;

								if ( classBoundType.typeArgs == null || classBoundType.typeArgs.isEmpty() )
								{
									matchingRelativType =
											new AsmClassType(
													createTypevarRelativeFqn(
															classBoundType ,
															part ) , 
													typeArgs );
								}
								else
								{
									throw new IllegalStateException(
											String.valueOf( classBoundType ) + ( classBoundType != null ? classBoundType.getClass().toString() : "" ) );
								}
							}
							else
							{
								throw new IllegalStateException(
										String.valueOf( resolvedBoundTypeparam.boundType ) + ( resolvedBoundTypeparam.boundType != null ? resolvedBoundTypeparam.boundType.getClass().toString() : "" ) );
							}
						}
					}
					else
					{
						if ( boundType instanceof AsmClassType )
						{
							final AsmClassType classBoundType = (AsmClassType) boundType;

							if ( classBoundType.typeArgs == null || classBoundType.typeArgs.isEmpty() )
							{
								matchingRelativType =
										new AsmClassType(
												createTypevarRelativeFqn(
														classBoundType ,
														part ) , 
												typeArgs );
							}
							else
							{
								//throw new IllegalStateException( String.valueOf( classBoundType ) + ( classBoundType != null ? classBoundType.getClass().toString() : "" ) );
								matchingRelativType =
										new AsmInnerClassType(
												//outerOrEnclosingType
												classBoundType ,
												//relativName
												//DottedName.valueOfDotSeparatedStr(
												//new JavaIdentifier(
														part.typeName ,
												typeArgs );
							}
						}
						else if ( boundType instanceof AsmInnerClassType )
						{
							matchingRelativType =
									new AsmInnerClassType(
											//outerOrEnclosingType
											(AsmInnerClassType) boundType ,
											//relativName
											//DottedName.valueOfDotSeparatedStr(
											//new JavaIdentifier(
													part.typeName ,
											typeArgs );
						}
						else
						{
							throw new IllegalStateException(
									String.valueOf( boundType ) + ( boundType != null ? boundType.getClass().toString() : "" ) );
						}
					}

					result =
							//new AsmMemberClassType(
							//		//outerOrEnclosingType
							//		result ,
							//		//simpleName
							//		part.typeName ,
							//		typeArgs );
							// Vermerken Typ-Variable und relativer Typ-Name mit Typ-Argumenten in ein temporres Objekt zum Ausflsen der Typ-Information anghand des desc-Parameters (Typ-Informationen ohne Generics) in der Methode parseFieldDescriptionAndSignature
							//new TmpAsmMemberClassOfTypevar(
							//		//outerOrEnclosingTypevar
							//		(AsmTypevar) result ,
							//		//relativName
							//		DottedName.valueOfDotSeparatedStr( part.typeName ) ,
							//		typeArgs );
							matchingRelativType;
				}
				else
				{
					if ( result instanceof AsmClassType )
					{
						final AsmClassType resultClassType = (AsmClassType) result;

						if ( resultClassType.typeArgs == null || resultClassType.typeArgs.isEmpty() )
						{
							result =
									//new AsmMemberClassType(
									//		//outerOrEnclosingType
									//		result ,
									//		//simpleName
									//		//relativName
									//		DottedName.valueOfDotSeparatedStr( part.typeName ) ,
									//		typeArgs );
									new AsmClassType(
											createTypevarRelativeFqn(
													resultClassType ,
													part ) , 
											typeArgs );
						}
						else
						{
							throw new IllegalStateException(
									String.valueOf( resultClassType ) + ( resultClassType != null ? resultClassType.getClass().toString() : "" ) );
						}
					}
					else if ( result instanceof AsmInnerClassType )
					{
						result =
								new AsmInnerClassType(
										//outerOrEnclosingType
										(AsmInnerClassType) result ,
										//relativName
										//DottedName.valueOfDotSeparatedStr(
										//new JavaIdentifier(
												part.typeName ,
										typeArgs );
					}
					else
					{
						throw new IllegalStateException( String.valueOf( result ) + ( result != null ? result.getClass().toString() : "" ) );
					}
				}
			}
		}

		return result;
		//while ( ! restTypeStr.isEmpty() )
		//{
		//	if ( ! restTypeStr.contains( "." ) )
		//	{
		//		if ( result == null )
		//		{
		//			// es gab keinen outer/enclosing type
		//			result = new AsmTypevar( restTypeStr );
		//		}
		//		else
		//		{
		//			result =
		//					new AsmMemberClassType(
		//							//outerOrEnclosingType
		//							result ,
		//							restTypeStr ,
		//							typeArgs );		
		//		}
		//	}
		//	else
		//	{
		//		// der Name der Typ-Variable enthlt einen Punkt, scheinbar ist der Typ ein inner/nested type of type variable
		//		// Testfall: TT.InnerIf<Ljava/lang/Integer;>;
		//		// Testfall: TT.InnerIf.InnerInnerIf<Ljava/lang/Integer;>;
		//		final String typevarNameOfTypevarInnerOrNestedType =
		//				restTypeStr.substring(
		//						//beginIndex
		//						0 ,
		//						//endIndex
		//						restTypeStr.indexOf( '.' ) );
		//
		//		final AsmTypevar asmTypevarWithInnerOrNestedType =
		//				new AsmTypevar(
		//						typevarNameOfTypevarInnerOrNestedType );
		//
		//		final String typevarInnerOrNestedTypStr =
		//				restTypeStr.substring(
		//						//beginIndex
		//						typevarNameOfTypevarInnerOrNestedType.length() + 1 );
		//
		//		final String typevarInnerOrNestedTypeName;
		//		final List<? extends AsmType> typeArgs;
		//		if ( typevarInnerOrNestedTypStr.contains( "<" ) )
		//		{
		//			typevarInnerOrNestedTypeName =
		//					typevarInnerOrNestedTypStr.substring(
		//							//beginIndex
		//							0 ,
		//							//endIndex
		//							typevarInnerOrNestedTypStr.indexOf( '<' ) );
		//
		//			final String typeArgsStr =
		//					typevarInnerOrNestedTypStr.substring(
		//							//beginIndex
		//							typevarInnerOrNestedTypStr.indexOf( '<' ) );
		//
		//			typeArgs =
		//					parseTypeArgs(
		//							typeArgsStr );
		//		}
		//		else
		//		{
		//			typevarInnerOrNestedTypeName = typevarInnerOrNestedTypStr;
		//			typeArgs = null;
		//		}
		//	}
		//}
		//return result;
	}

	private static AsmTopLevelOrInnerClassType resolveIntersectionBoundOfTypevarRelativType(
			final TypeNameAndTypeArgsStrTuple part ,
			final AsmIntersectionBound intersectionBound ,
			final ImmutableArray<? extends AsmType> typeArgs ,
			final ClasspathFiles classpathFiles )
	{
		AsmTopLevelOrInnerClassType matchingRelativTypeOfIntersectionBound = null;
		for ( final AsmType oneIntersectionBound : intersectionBound.boundTypes )
		{
			final AsmClassType oneIntersectionBoundClassType = (AsmClassType) oneIntersectionBound;

			final Fqn relativTypeClassFqn =
					createTypevarRelativeFqn(
							oneIntersectionBoundClassType ,
							part );

			if ( classpathFiles.containsClass(
					relativTypeClassFqn ) )
			{
				if ( oneIntersectionBoundClassType.typeArgs == null || oneIntersectionBoundClassType.typeArgs.isEmpty() )
				{
					matchingRelativTypeOfIntersectionBound =
							//(AsmClassType) oneIntersectionBound
							new AsmClassType(
									relativTypeClassFqn ,
									typeArgs );
				}
				else
				{
					matchingRelativTypeOfIntersectionBound =
							new AsmInnerClassType(
									//outerOrEnclosingType
									oneIntersectionBoundClassType ,
									//relativName
									//DottedName.valueOfDotSeparatedStr(
									//new JavaIdentifier(
											part.typeName ,
									typeArgs );
				}

				break;
			}
		}

		if ( matchingRelativTypeOfIntersectionBound == null )
		{
			throw new IllegalStateException(
					intersectionBound + " -> " + part );
		}

		return matchingRelativTypeOfIntersectionBound;
	}

	private static AsmAbstractTypeParam resolveBoundTypevar(
			final Map<String, ? extends AsmExtendsTypeParam> extendsTypeParamMap ,
			final AsmTypevar boundTypevar )
	{
		final AsmAbstractTypeParam typeParam = extendsTypeParamMap.get( boundTypevar.name );
		
		if ( typeParam instanceof AsmUnboundedTypeParam )
		{
			return typeParam;
		}

		final AsmExtendsTypeParam extendsTypeParam = (AsmExtendsTypeParam) typeParam;

		if ( extendsTypeParam.boundType instanceof AsmTypevar )
			// nochmal auf Typevar gemappter Typeparam: T2 extends T1, T1 extends T0
		{
			return resolveBoundTypevar(
					extendsTypeParamMap ,
					(AsmTypevar) extendsTypeParam.boundType );
		}

		return extendsTypeParam;
	}

	public static Fqn createTypevarRelativeFqn(
			final AsmTopLevelOrInnerClassType base ,
			TypeNameAndTypeArgsStrTuple part )
	{
		return new Fqn(
				base.getEntireFqn().packageName ,
				DottedName.valueOfDotSeparatedStr(
						base.getEntireFqn().dottedName.toJavaName() + 
						"." + 
						part.typeName ) );
	}

	// Testfall: TT.InnerIf<Ljava/lang/Integer;>;
	// Testfall: TT.InnerIf.InnerInnerIf<Ljava/lang/Integer;>;
	// fhrendes 'TT.' bereits abgeschnitten
	private static List<? extends TypeNameAndTypeArgsStrTuple> separateInnerOrNestedClassParts(
			final String strToSeparate )
	{
		if ( ! strToSeparate.contains( "." ) )
		{
			return Arrays.asList(
					separateTypeNameAndTypeArgs(
							strToSeparate ) );
		}

		String restStrToSeparate = strToSeparate;

		final ArrayList<TypeNameAndTypeArgsStrTuple> result = new ArrayList<>();

		while ( ! restStrToSeparate.isEmpty() )
		{
			final int dotPos = restStrToSeparate.indexOf( '.' );
			final int leftAngleBracePos = restStrToSeparate.indexOf( '<' );

			if ( dotPos < 0 )
				// es gibt kein '.'
			{
				result.add(
						separateTypeNameAndTypeArgs(
								restStrToSeparate ) );

				restStrToSeparate = "";
			}
			else if (
					// es gibt ein '.'
					// es gibt kein '<'
					leftAngleBracePos < 0 ||
					// oder das '<' liegt hinter dem Punkt
					dotPos < leftAngleBracePos )
			{
				final String partStr =
						restStrToSeparate.substring(
								//beginIndex
								0 ,
								//endIndex
								dotPos );

				result.add(
						separateTypeNameAndTypeArgs(
								partStr ) );

				restStrToSeparate =
						restStrToSeparate.substring(
								//beginIndex
								dotPos + 1 );
			}
			else
				// es gibt ein '.'
				// es gibt ein '<'
				// das '<' liegt vor dem Punkt
			{
				final int afterGtPosRecursiv =
						AsmDescAndSignatureParserUtil.getAfterGtPosRecursiv(
								restStrToSeparate ,
								leftAngleBracePos + 1 );

				result.add(
						separateTypeNameAndTypeArgs(
								restStrToSeparate.substring(
										//beginIndex
										0 ,
										//endIndex
										afterGtPosRecursiv ) ) );

				restStrToSeparate =
						restStrToSeparate.substring(
								//beginIndex: trennenden Punkt berspringen
								afterGtPosRecursiv + 1 );
			}
		}

		return result;
	}

	private static List<? extends TypeNameAndTypeArgsStrTuple> joinInnerOrNestedClassPartsWithoutTypeArgs(
			final List<? extends TypeNameAndTypeArgsStrTuple> typeNameAndTypeArgsStrTupleListToJoin )
	{
		final ArrayList<TypeNameAndTypeArgsStrTuple> result = new ArrayList<>();

		for ( final TypeNameAndTypeArgsStrTuple currTplToJoin : typeNameAndTypeArgsStrTupleListToJoin )
		{
			if ( 
					// nicht das erste Element mit dem zweiten Element joinen, weil das erste Element die Typ-Variable ist
					( result.size() > 1 ) 
					&&
					( ! hasLastTplTypeargs( result ) ) )
				// das letzte vermerkte Element hat keine Typ-Argumente:
				// das aktuelle Element und das letzte vermerkte Element kann verschmolzen werden
			{
				final TypeNameAndTypeArgsStrTuple resultLastTpl =
						result.get( result.size() - 1 );

				final String joinedTypeName =
						resultLastTpl.typeName +
						// Benutzung des Trenners fr innere/geschachtelte Klassen, weil nur diese an dieser Stelle auftauchen knnen
						// TODO: doch nicht wegen relativName in AsmMemberClassType, welcher ein DottedName ist und bei welchem das visitInnerClass-Ersetzen nicht funktioniert: "$" +
						"." +
						currTplToJoin.typeName;

				final TypeNameAndTypeArgsStrTuple joinedTpl =
						new TypeNameAndTypeArgsStrTuple(
								joinedTypeName ,
								currTplToJoin.typeArgsStr );

				// Setzen auf letzte Position
				result.set(
						//index letzte Position
						result.size() - 1 ,
						//element to set
						joinedTpl );
			}
			else
			{
				result.add( currTplToJoin );
			}
		}

		return result;
	}

	private static TypeNameAndTypeArgsStrTuple separateTypeNameAndTypeArgs(
			final String strToSep )
	{
		if ( ! strToSep.contains( "<" ) )
		{
			return new TypeNameAndTypeArgsStrTuple(
					strToSep ,
					//typeArgsStr
					null );
		}

		final int leftAngleBracePos = strToSep.indexOf( '<' );
		
		return new TypeNameAndTypeArgsStrTuple(
				//typeName
				strToSep.substring(
						//beginIndex
						0 ,
						//endIndex
						leftAngleBracePos ) ,
				//typeArgsStr
				strToSep.substring(
						//beginIndex
						leftAngleBracePos ) );
	}

	private static boolean hasLastTplTypeargs(
			final List<? extends TypeNameAndTypeArgsStrTuple> listToCheck )
	{
		if ( listToCheck.isEmpty() )
        {
			return false;
        }

		final TypeNameAndTypeArgsStrTuple resultLastTpl =
				listToCheck.get( listToCheck.size() - 1 );

		// ob das letzte vermerkte Element Typ-Argumente hat
		return ! StringUtil.isNullOrEmpty( resultLastTpl.typeArgsStr );
	}

}
