package de.heinerkuecker.asm_improved;

import java.util.Map.Entry;
import java.util.TreeMap;

import de.heinerkuecker.asm_improved.util.DottedName;
import de.heinerkuecker.asm_improved.util.Fqn;
import de.heinerkuecker.asm_improved.util.PackageName;

/**
 * Hilfsklasse fr die punkt-getrennten
 * Namen von inneren oder
 * geschachtelten Klassen.
 * <br/><br/>
 * Im Java-Quellcode werden innere oder
 * geschachtelte Klassen ber einen
 * mit Punkten abgetrennten qualifizierten
 * Namen adressiert.
 * <br/><br/>
 * Der Java-Compiler erzeugt fr innere oder
 * geschachtelte Klassen *.class-Dateien,
 * deren Namens-Teile ber Dollar-Zeichen
 * abgetrennt sind.
 * <br/><br/>
 * Dieses Verfahren nennt man Name Mangling.
 * <br/><br/>
 * Leider sind Dollar-Zeichen als Zeichen
 * in Java-Klassen-Namen zwar nicht empfohlen,
 * aber erlaubt, so dass der entstehende
 * Name der *.class-Datei uneindeutig ist.
 * <br/><br/>
 * In der erzeugten *.class-Datei sind
 * Informationen hinterlegt, aus denen
 * der punkt-getrennte Name rekonstruiert
 * werden kann.
 * <br/><br/>
 * In der Methode
 * {@link de.heinerkuecker.asm_improved.AsmClasspathInfo.AsmClasspathInfoClassVisitor#visitInnerClass},
 * welche die Methode
 * {@link org.objectweb.asm.ClassVisitor#visitInnerClass}
 * berschreibt,
 * werden diese Informationen von der
 * ASM-Lib dem Anwendungscode
 * zugnglich gemacht.
 * <br/><br/>
 * Fr die mehrfach geschachtelte Klasse
 * {@link de.heinerkuecker.asm_improved.test_data.MultiNestedClass.NestedClass0.NestedClass1.NestedClass2.NestedClass3}
 * werden folgende Informationen geliefert:
 * <pre>
 * name:      de/heinerkuecker/asm/MultiEnclosingClass$EnclosingClass0
 * outerName: de/heinerkuecker/asm/MultiEnclosingClass
 * innerName: EnclosingClass0
 * 
 * name:      de/heinerkuecker/asm/MultiEnclosingClass$EnclosingClass0$EnclosingClass1
 * outerName: de/heinerkuecker/asm/MultiEnclosingClass$EnclosingClass0
 * innerName: EnclosingClass1
 * 
 * name:      de/heinerkuecker/asm/MultiEnclosingClass$EnclosingClass0$EnclosingClass1$EnclosingClass2
 * outerName: de/heinerkuecker/asm/MultiEnclosingClass$EnclosingClass0$EnclosingClass1
 * innerName: EnclosingClass2
 * 
 * name:      de/heinerkuecker/asm/MultiEnclosingClass$EnclosingClass0$EnclosingClass1$EnclosingClass2$NestedClass
 * outerName: de/heinerkuecker/asm/MultiEnclosingClass$EnclosingClass0$EnclosingClass1$EnclosingClass2
 * innerName: NestedClass
 * </pre>
 * Es wird im Parameter name der Methode
 * {@link de.heinerkuecker.asm_improved.AsmClasspathInfo.AsmClasspathInfoClassVisitor#visitInnerClass}
 * jeweils der Name der *.class-Datei ohne Postfix geliefert.
 * <br/><br/>
 * Im Parameter outerName wird der Name der
 * outer/enclosing class und
 * im Parameter innerName der relative Name
 * der inner/nested class geliefert.
 * <br/><br/>
 * In der Methode
 * {@link de.heinerkuecker.asm_improved.AsmClasspathInfo.AsmClasspathInfoClassVisitor#visitInnerClass}
 * wird der encodete Name (Parameter name) sowie
 * die beiden Namensteile outerName und innerName durch einen Punkt
 * verbunden/getennt in einer {@link TreeMap} vermerkt.
 * <br/><br/>
 * In der Methode
 * {@link de.heinerkuecker.asm_improved.InnerOrNestedClassNameResolver#resolve}
 * wird in dem bergebenen Dollar-Zeichen-separierten Namen
 * der jeweils entsprechende Namensteil durch den
 * in der TreeMap vermerkten Punkt-separierten Namens-Teil
 * ersetzt,
 * <br/>
 * wobei die sortierte Map {@link TreeMap} rckwrts
 * durchlaufen wird, damit rechts stehende Namens-Teile
 * vor links stehenden Namens-Teilen ersetzt werden.
 * <br/>
 * Andersherum, vorwrts durchlaufen und links
 * stehende Namens-Teile vor rechts stehenden
 * Namens-Teile ersetzen, wrde dazu fhren,
 * dass weitere Dollar-Zeichen-separierte Namen
 * nicht mehr passen und ersetzt werden wrden.
 * 
 * @author Heiner K&uuml;cker
 */
final class InnerOrNestedClassNameResolver
{

	/**
	 * Konstruktor private,
	 * Instanziieren verboten,
	 * Util-Klasse.
	 */
	private InnerOrNestedClassNameResolver()
	{
		super();
	}

	/**
	 * Auflsen Name nach Javadoc oben {@link InnerOrNestedClassNameResolver}.
	 * 
	 * @param dollarSeparatedRawName Name wie von ASM geliefert
	 *        (Package durch '/' abgetrennt, inner/nested class name durch '$' abgetennt)
	 * @param innerClassFileNameToDotSeparatedNameMap
	 *        Map mit Informationen zum Auflsen
	 *        der Namen von inneren oder 
	 *        geschachtelten Klassen welche
	 *        in der Methode
	 *        {@link de.heinerkuecker.asm_improved.AsmClasspathInfo.AsmClasspathInfoClassVisitor#visitInnerClass}
	 *        aufgesammelt wurden
	 * @return korrigierter Name als {@link Fqn}
	 */
	public static Fqn resolve(
			final String dollarSeparatedRawName ,
			final TreeMap<String, ? extends String> innerClassFileNameToDotSeparatedNameMap )
	{
		String tmpRawName = dollarSeparatedRawName;

		for ( final Entry<? extends String, ? extends String> innerClassFileNameToDotSeparatedNameEntry : innerClassFileNameToDotSeparatedNameMap
				// rckwrts laufen
				.descendingMap().entrySet() )
		{
			if ( tmpRawName.startsWith( innerClassFileNameToDotSeparatedNameEntry.getKey() ) )
			{
				tmpRawName =
						innerClassFileNameToDotSeparatedNameEntry.getValue() +
						tmpRawName.substring(
								//beginIndex
								innerClassFileNameToDotSeparatedNameEntry.getValue().length() );
			}
		}

		final int packageEndIndex =
				lastIndexOfSlashOrZero(
						tmpRawName );

		final String packageRawStr =
				tmpRawName.substring(
						//beginIndex
						0 ,
						//endIndex
						packageEndIndex );

		final String packageDotSepStr =
				packageRawStr.replace( '/' , '.' );

		final String classDotSepName =
				tmpRawName.substring(
						//beginIndex
						packageEndIndex + 1 );

		final PackageName packageName =
				new PackageName(
						packageDotSepStr );

		final DottedName dottedName =
				DottedName.valueOfDotSeparatedStr(
						classDotSepName );

		return new Fqn(
				packageName ,
				dottedName );
	}

	/**
	 * @param str abzufragender String
	 * @return letzte Position des '/'-Zeichens im bergebenen String oder 0, wenn '/'-Zeichen nicht vorhanden
	 */
    private static int lastIndexOfSlashOrZero(
    		final String str )
    {
    	final int lastIndexOfSlash =
    			str.lastIndexOf( '/' );

    	if ( lastIndexOfSlash < 0 )
        {
	        return 0;
        }

    	return lastIndexOfSlash;
    }

}
