package de.heinerkuecker.asm_improved.annotation_use;

import java.util.Map.Entry;

import de.heinerkuecker.asm_improved.AsmClassInfo;
import de.heinerkuecker.asm_improved.AsmClasspathInfo;
import de.heinerkuecker.asm_improved.AsmMethodInfo;
import de.heinerkuecker.asm_improved.type.AsmClassType;
import de.heinerkuecker.asm_improved.util.DottedName;
import de.heinerkuecker.asm_improved.util.Fqn;
import de.heinerkuecker.asm_improved.util.ImmutableArray;
import de.heinerkuecker.asm_improved.util.ImmutableMap;
import de.heinerkuecker.asm_improved.util.PackageName;
import de.heinerkuecker.asm_improved.util.SeparatedStrBuff;

/**
 * Benutzung einer Annotation
 * im Java-Byte-Code.
 * 
 * @author Heiner K&uuml;cker
 */
public final class AsmAnnotationUse
{
	/**
	 * Using of annotation of class {@link FunctionalInterface}.
	 */
	public static final AsmAnnotationUse JAVA_LANG_FUNCTIONAL_INTERFACE =
			new AsmAnnotationUse(
					new AsmClassType(
							new Fqn(
									PackageName.JAVA_LANG_PACKAGE_NAME ,
									DottedName.valueOfNoDotSeparatedStr( "FunctionalInterface" ) ) ) ,
					null );

	/**
	 * Klasse der Annotation.
	 * 
	 * TODO rename to type
	 */
	public final AsmClassType annotationType;

	/**
	 * Parameter-Werte der Annotation.
	 */
	public final ImmutableMap<String, ? extends Object> values;

	// TODO knnte hier eingebaut werden, weil ASM diese Information immer liefert: public final boolean visibleAtRuntime;

	/**
	 * Konstruktor.
	 * 
	 * @param annotationType Klasse der Annotation
	 * @param values an die Annotation bergebene Werte
	 */
	public AsmAnnotationUse(
			final AsmClassType annotationType ,
			final ImmutableMap<String, ? extends Object> values )
	{
		this.annotationType = annotationType;
		this.values = values;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode()
	{
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((annotationType == null) ? 0 : annotationType.hashCode());
		result = prime * result + ((this.values == null) ? 0 : this.values.hashCode());
		return result;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(
			final Object obj)
	{
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		AsmAnnotationUse other = (AsmAnnotationUse) obj;
		if (annotationType == null)
		{
			if (other.annotationType != null)
				return false;
		} else if (!annotationType.equals(other.annotationType))
			return false;
		if (this.values == null)
		{
			if (other.values != null)
				return false;
		} else if (!this.values.equals(other.values))
			return false;
		return true;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString()
	{
		//return "AnnotationUse [annotationType=" + annotationType + ", values=" + this.values + "]";
		final String valuesStr;
		if ( values == null || values.isEmpty() )
		{
			valuesStr = "";
		}
		else
		{
			final SeparatedStrBuff buff = new SeparatedStrBuff( " , " );

			for ( final Entry<String, ? extends Object> valueEntry : values.entrySet() )
			{
				buff.add(
						valueEntry.getKey() +
						" = " +
						valueEntry.getValue() );
			}

			valuesStr = " " + buff.toString() + " ";
		}

		return "@" + annotationType + "(" +valuesStr + ")";
	}

	/**
	 * Lesen des gewnschten Annotations-Wertes.
	 * Falls ein default verwendet wurde,
	 * wird dieser aus dem Byte-Code der
	 * Annotation beschafft.
	 * 
	 * @param asmClasspathInfo zum Holen der {@link AsmClassInfo} dieser Annotation
	 * @param name Name des Annotations-Wertes
	 * @return Annotations-Wert unter Beachtung Default
	 */
	public Object getValueUnderConsiderationDefault(
			final AsmClasspathInfo asmClasspathInfo ,
			final String name )
	{
		if ( this.values != null && ( ! this.values.isEmpty() ) )
		{
			final Object annotationValue = this.values.get( name );

			if ( annotationValue != null )
				// diese funktioniert deshalb korrekt, weil null niemals als Value erlaubt ist
			{
				return annotationValue;
			}
		}

		return getDefault(
				asmClasspathInfo ,
				name );
	}

	/**
	 * Lesen des Default Annotations-Wertes.
	 * Dieser wird aus dem Byte-Code der
	 * Annotation beschafft.
	 * 
	 * @param asmClasspathInfo zum Holen der {@link AsmClassInfo} dieser Annotation
	 * @param name Name des Annotations-Wertes
	 * @return Default-Annotations-Wert
	 */
	public Object getDefault(
			final AsmClasspathInfo asmClasspathInfo ,
			final String name )
	{
		final AsmClassInfo annotationClassInfo =
				asmClasspathInfo.getClassInfoMandantory(
						this.annotationType.fqn );

		final ImmutableArray<? extends AsmMethodInfo> methodInfos = annotationClassInfo.methods.get( name );

		final AsmMethodInfo methodInfo = methodInfos.get( 0 );

		return methodInfo.annotationDefaultValue;
	}

}
