package de.heinerkuecker.asm_improved.type;

import java.util.Map;
import java.util.Stack;

import de.heinerkuecker.asm_improved.AsmClasspathInfo;
import de.heinerkuecker.asm_improved.annotation_use.AsmAnnotationUse;
import de.heinerkuecker.asm_improved.util.ImmutableArray;
import de.heinerkuecker.asm_improved.util.ImmutableMap;

/**
 * Array-Typ.
 * <p>
 * Fr mehrdimensionale Arrays werden mehrere Instanzen dieser Klasse ineinander geschachtelt:
 * <pre>
 * zweidimensionales String-Array
 * 
 * String[] []
 * 
 * wird codiert durch
 * 
 * new {@link AsmArray}( new {@link AsmArray}( {@link AsmClassType#JAVA_LANG_STRING} ) )
 * 
 * </pre>
 * Eventuell stelle ich diese Struktur fr die leichtere Abbildung der Typ-Annotationen je Array-Dimension so um,
 * dass am Array die Diemension als int vermerkt und je Dimension die Typ-Annotationen vermerkt werden.
 * 
 * @author Heiner K&uuml;cker
 */
public final class AsmArray
extends AsmType
{
	/**
	 * Typ der Array-Elemente.
	 */
	public final AsmType elementType;

	/**
	 * Konstruktor.
	 * 
	 * @param typeAnnotations 
	 * @param elementType see {@link #elementType}
	 */
	public AsmArray(
			final ImmutableArray<AsmAnnotationUse> typeAnnotations  ,
			final AsmType elementType )
	{
		super( typeAnnotations );
		this.elementType = elementType;
	}

	/**
	 * Konstruktor.
	 * 
	 * @param typeAnnotations 
	 * @param elementType see {@link #elementType}
	 */
	public AsmArray(
			final AsmType elementType )
	{
		super( null );
		this.elementType = elementType;
	}

	/**
	 * @return Anzahl Dimensionen dieses Arrays (Anzahl eckiger Klammerpaare)
	 */
	public int getDimensionCount()
	{
		if ( this.elementType instanceof AsmArray )
		{
			return ( (AsmArray) this.elementType ).getDimensionCount() + 1;
		}
		return 1;
	}
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean isPrimitive()
	{
		return false;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean isAssignableFrom(
			final AsmType typeToTest ,
			final AsmClasspathInfo asmClasspathInfo )
	{
		throw new RuntimeException( "nicht fertig" );
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public String toJavaString()
	{
		/*
		 * Ziemlich komplizierter Code zur korrekten Darstellung
		 * der Typ-Annotationen je Array-Dimension, siehe
		 * TestTypeAnnotationOnFieldArray1 und TypeAnnotationOnFieldArray1Test
		 */
		final StringBuilder elementArrayStrBuff = new StringBuilder();

		AsmType currElementType = this;

		//System.out.println( new Exception().getStackTrace()[ 0 ] + " " + currElementType.getClass() );

		while ( currElementType instanceof AsmArray )
		{
			elementArrayStrBuff.append(
					" " + currElementType.typeAnnotationsToStringWithDelimiterSpace() );

			elementArrayStrBuff.append(
					"[]" );

			System.out.println( new Exception().getStackTrace()[ 0 ] + " " + elementArrayStrBuff );

			currElementType = ( (AsmArray) currElementType ).elementType;

			System.out.println( new Exception().getStackTrace()[ 0 ] + " " + currElementType.getClass() );
		}

		return
				currElementType.typeAnnotationsToStringWithDelimiterSpace() +
				currElementType.toJavaStringWithoutTypeAnnotations() +
				elementArrayStrBuff.toString();
	}

	/**
	 * {@inheritDoc}
	 * 
	 * TODO kann weg
	 */
	@Override
	public String toJavaStringWithoutTypeAnnotations()
	{
		return elementType + "[]";
	}

	/**
	 * Umdrehen in toString noch mal umgedreht, weil die Verdrehung nur auf top-level gilt
	 */
	private String toArrayElementJavaString()
	{
		final String elementTypeStr;
		if ( ! ( elementType.typeAnnotations == null || elementType.typeAnnotations.isEmpty() ) )
		{
			elementTypeStr =
					elementType.typeAnnotationsToStringWithDelimiterSpace() +
					" " +
					elementType.toJavaStringWithoutTypeAnnotations();
		}
		else
		{
			elementTypeStr = String.valueOf( elementType );
		}
		return typeAnnotationsToStringWithDelimiterSpace() + "[]" + elementTypeStr;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public String toString()
	{
		return toJavaString();
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode()
	{
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((elementType == null) ? 0 : elementType.hashCode());
		return result;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(
			final Object obj )
	{
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		AsmArray other = (AsmArray) obj;
		if (elementType == null)
		{
			if (other.elementType != null)
				return false;
		} else if (!elementType.equals(other.elementType))
			return false;
		return true;
	}

	@Override
	public AsmArray resolveBound(
			final ImmutableMap<String, ? extends AsmType> typeargsNameToBoundMap )
	{
		return new AsmArray(
				elementType.resolveBound(
						typeargsNameToBoundMap ,
						new Stack<String>() ) );
	}

	@Override
	public AsmType resolveBound(
			final ImmutableMap<String, ? extends AsmType> typeargsNameToBoundMap ,
			final Stack<String> typevarNameStack )
	{
		return new AsmArray(
				elementType.resolveBound(
						typeargsNameToBoundMap ,
						typevarNameStack ) );
	}

	/**
	 * Ziemliche seltsame Spezifikation:
	 * <a href="https://docs.oracle.com/javase/specs/jls/se8/html/jls-9.html#jls-9.7.4">Java Language Specification - 9.7.4. Where Annotations May Appear </a>
	 * in "@C int @A [] @B [] f; @A applies to the array type
	 * int[][], @B applies to its component type int[], and @C
	 * applies to the element type int."
	 * 
	 * @param fixedTypeAnnotationsMap
	 * @param typePath
	 * @return
	 */
	@Override
	public AsmArray cloneAddTypeAnnotations(
			final Map<AsmTypePath, ImmutableArray<AsmAnnotationUse>> typeAnnotationsMap ,
			final AsmTypePath typePath )
	{
		//System.out.println( new Exception().getStackTrace()[ 0 ] + " " + this + " " + System.identityHashCode( this ) + " " + typePath );
		if ( this.typeAnnotations != null )
			// zweimal Typ-Annotationen hinzufgen sollte nicht vorkommen
		{
			throw new IllegalStateException();
		}

		final AsmTypePath typePathWithAddedElement =
				typePath.cloneAdd(
						new AsmTypePath.ARRAY_ELEMENT(
								null ) );

		if ( this.elementType instanceof AsmArray )
		{
			return new AsmArray(
					typeAnnotationsMap.get(
							typePath ) ,
					( (AsmArray) this.elementType ).cloneAddTypeAnnotations(
							typeAnnotationsMap ,
							typePathWithAddedElement ) );
		}
		else
		{
			return new AsmArray(
					typeAnnotationsMap.get(
							typePath ) ,
					this.elementType.cloneAddTypeAnnotations(
							typeAnnotationsMap ,
							typePathWithAddedElement ) );
		}
		//throw new RuntimeException( "nicht fertig" );
	}

}
