package de.heinerkuecker.asm_improved.type;

import java.util.HashMap;
import java.util.Map;
import java.util.Stack;

import de.heinerkuecker.asm_improved.AsmClassInfo;
import de.heinerkuecker.asm_improved.AsmClasspathInfo;
import de.heinerkuecker.asm_improved.annotation_use.AsmAnnotationUse;
import de.heinerkuecker.asm_improved.type.generics.param.AsmAbstractTypeParam;
import de.heinerkuecker.asm_improved.type.util.AsmTypeUtil;
import de.heinerkuecker.asm_improved.util.Fqn;
import de.heinerkuecker.asm_improved.util.ImmutableArray;
import de.heinerkuecker.asm_improved.util.ImmutableMap;
import de.heinerkuecker.asm_improved.util.SeparatedStrBuff;

/**
 * Top level class with type arguments.
 * 
 * @author Heiner K&uuml;cker
 */
public final class AsmClassType
extends AsmTopLevelOrInnerClassType
{
	/**
	 * Type of {@link java.lang.Object}.
	 */
	public static final AsmClassType JAVA_LANG_OBJECT =
			new AsmClassType(
					Fqn.JAVA_LANG_OBJECT_FQN ,
					(ImmutableArray<? extends AsmType>) null );

	/**
	 * Type of {@link java.lang.Integer}.
	 */
	public static final AsmClassType JAVA_LANG_BOOLEAN =
			new AsmClassType(
					Fqn.JAVA_LANG_BOOLEAN_FQN ,
					(ImmutableArray<? extends AsmType>) null );

	/**
	 * Type of {@link java.lang.Byte}.
	 */
	public static final AsmClassType JAVA_LANG_BYTE =
			new AsmClassType(
					Fqn.JAVA_LANG_BYTE_FQN ,
					(ImmutableArray<? extends AsmType>) null );

	/**
	 * Type of {@link java.lang.Integer}.
	 */
	public static final AsmClassType JAVA_LANG_SHORT =
			new AsmClassType(
					Fqn.JAVA_LANG_SHORT_FQN ,
					(ImmutableArray<? extends AsmType>) null );

	/**
	 * Type of {@link java.lang.Integer}.
	 */
	public static final AsmClassType JAVA_LANG_INTEGER =
			new AsmClassType(
					Fqn.JAVA_LANG_INTEGER_FQN ,
					(ImmutableArray<? extends AsmType>) null );

	/**
	 * Type of {@link java.lang.Integer}.
	 */
	public static final AsmClassType JAVA_LANG_LONG =
			new AsmClassType(
					Fqn.JAVA_LANG_LONG_FQN ,
					(ImmutableArray<? extends AsmType>) null );

	/**
	 * Type of {@link java.lang.Integer}.
	 */
	public static final AsmClassType JAVA_LANG_FLOAT =
			new AsmClassType(
					Fqn.JAVA_LANG_FLOAT_FQN ,
					(ImmutableArray<? extends AsmType>) null );

	/**
	 * Type of {@link java.lang.Integer}.
	 */
	public static final AsmClassType JAVA_LANG_DOUBLE =
			new AsmClassType(
					Fqn.JAVA_LANG_DOUBLE_FQN ,
					(ImmutableArray<? extends AsmType>) null );

	/**
	 * Type of {@link java.lang.String}.
	 */
	public static final AsmClassType JAVA_LANG_STRING =
			new AsmClassType(
					Fqn.JAVA_LANG_STRING_FQN ,
					(ImmutableArray<? extends AsmType>) null );

	/**
	 * Vollqualifizierter Name mit
	 * eindeutiger Abtrennung Package.
	 */
	public final Fqn fqn;

	/**
	 * Typ-Argumente.
	 */
	public final ImmutableArray<? extends AsmType> typeArgs;
	//public final ImmutableMap<String, ? extends AsmType> typeArgs;

	/**
	 * Konstruktor.
	 */
	public AsmClassType(
			final ImmutableArray<? extends AsmAnnotationUse> typeAnnotations ,
			final Fqn fqn ,
			final ImmutableArray<? extends AsmType> typeArgs
			//final ImmutableMap<String, ? extends AsmType> typeArgs
			)
	{
		super( typeAnnotations );
		this.fqn = fqn;
		this.typeArgs = typeArgs;

		//if ( ! JavaIdentifier.is( simpleName ) )
		//{
		//	throw new IllegalArgumentException( simpleName );
		//}
	}

	/**
	 * Konstruktor.
	 */
	public AsmClassType(
			final Fqn fqn ,
			final ImmutableArray<? extends AsmType> typeArgs
			//final ImmutableMap<String, ? extends AsmType> typeArgs
			)
	{
		this(
				null ,
				fqn ,
				typeArgs );
	}

	/**
	 * Konstruktor.
	 */
	public AsmClassType(
			final Fqn fqn ,
			final AsmType... typeArgs )
	{
		this(
				fqn , 
				( typeArgs == null || typeArgs.length < 1 
				? null 
						: ImmutableArray.create( typeArgs ) ) );
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int hashCode()
	{
		final int prime = 31;
		int result = 1;
		result = prime * result + (/*(fqn == null) ? 0 :*/ fqn.hashCode());
		result = prime * result
				+ ((typeArgs == null) ? 0 : typeArgs.hashCode());
		return result;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean equals(
			final Object obj)
	{
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final AsmClassType other = (AsmClassType) obj;
		/*if (fqn == null)
	    {
		    if (other.fqn != null)
			    return false;
	    } else*/ if (!fqn.equals(other.fqn))
	    	return false;
	    if (typeArgs == null)
	    {
	    	if (other.typeArgs != null)
	    		return false;
	    } else if (!typeArgs.equals(other.typeArgs))
	    	return false;
	    return true;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public String toJavaString()
	{
		return typeAnnotationsToStringWithDelimiterSpace() + this.fqn.toFqnStr() + typeArgsToStr();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public String toJavaStringWithoutTypeAnnotations()
	{
		return this.fqn.toFqnStr() + typeArgsToStr();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public String toString()
	{
		return toJavaString();
	}

	/**
	 * @return
	 */
	// TODO copy paste code
	private String typeArgsToStr()
	{
		if ( this.typeArgs == null || this.typeArgs.isEmpty() )
		{
			return "";
		}

		final SeparatedStrBuff commaSpaceSepStrBuff =
				new SeparatedStrBuff( ", " );

		for ( final AsmType typeArg : typeArgs )
		{
			commaSpaceSepStrBuff.add(
					typeArg.toString() );
		}

		return "<" + commaSpaceSepStrBuff.toString() + ">";
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean isPrimitive()
	{
		return false;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Fqn getEntireFqn()
	{
		return this.fqn;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean isAssignableFrom(
			final AsmType typeToTest ,
			final AsmClasspathInfo asmClasspathInfo )
	{
		//throw new RuntimeException( "nicht fertig" );
		if ( ! ( typeToTest instanceof AsmClassType ) )
		{
			return false;
		}

		final AsmClassType classToTest = (AsmClassType) typeToTest;

		//ClasspathClasses.ClassInfo classInfoToTest = asmClasspathInfo.getClassInfoMandantory( classToTest.fqn.toFqnStr() );
		final AsmClassInfo thisClassInfo =
				asmClasspathInfo.getClassInfoMandantory(
						this.fqn );

		if ( thisClassInfo.isAssignableFrom(
				classToTest.fqn ,
				asmClasspathInfo ) )
		{
			return true;
		}

		return false;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public ImmutableMap<String, AsmType> getEntireTypeArgs(
			final AsmClasspathInfo asmClasspathInfo )
	{
		if ( this.typeArgs == null )
		{
			return null;
		}

		final AsmClassInfo thisClassInfo =
				asmClasspathInfo.getClassInfoMandantory(
						this.fqn );

		final ImmutableArray<? extends AsmAbstractTypeParam> typeParams = thisClassInfo.typeParams;

		final Map<String, AsmType> result = new HashMap<>();

		for ( int index = 0 ; index < typeParams.size() ; index++ )
		{
			final AsmAbstractTypeParam typeParam = typeParams.get( index );
			result.put(
					typeParam.name ,
					this.typeArgs.get( index ) );
		}

		return ImmutableMap.create( result );
	}

	@Override
	public AsmClassType resolveBound(
			final ImmutableMap<String, ? extends AsmType> typeargsNameToBoundMap )
	{
		return new AsmClassType(
				fqn ,
				AsmTypeUtil.resolveTypeArgs(
						typeArgs ,
						typeargsNameToBoundMap ,
						new Stack<String>() ) );
	}

	@Override
	public AsmClassType resolveBound(
			final ImmutableMap<String, ? extends AsmType> typeargsNameToBoundMap ,
			final Stack<String> typevarNameStack )
	{
		return new AsmClassType(
				fqn ,
				AsmTypeUtil.resolveTypeArgs(
						typeArgs ,
						typeargsNameToBoundMap ,
						typevarNameStack ) );
	}

	@Override
	public AsmClassType cloneAddTypeAnnotations(
			final Map<AsmTypePath, ImmutableArray<AsmAnnotationUse>> typeAnnotationsMap ,
			final AsmTypePath typePath )
	{
		if ( this.typeAnnotations != null )
			// zweimal Typ-Annotationen hinzufgen sollte nicht vorkommen
		{
			throw new IllegalStateException();
		}

		//System.out.println( new Exception().getStackTrace()[ 0 ] + " " + this + " type path to correct " + typePath );
		//System.out.println( new Exception().getStackTrace()[ 0 ] + " typeAnnotationsMap type paths " + typeAnnotationsMap.keySet() );
		//new Exception().printStackTrace( System.out );

		return new AsmClassType(
				typeAnnotationsMap.get(
						typePath ) ,
				fqn ,
				AsmTypePath.typeArgsCloneAddTypeAnnotations(
						this.typeArgs ,
						typeAnnotationsMap ,
						typePath ) );
		//throw new RuntimeException( "nicht fertig" );
	}

}
