package de.heinerkuecker.asm_improved.type;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.objectweb.asm.TypePath;

import de.heinerkuecker.asm_improved.annotation_use.AsmAnnotationUse;
import de.heinerkuecker.asm_improved.util.ImmutableArray;

/**
 * Verbesserte Variante von
 * {@link TypePath} mit
 * Enums und {@link #equals} und
 * {@link #hashCode}-Methode.
 * 
 * @author Heiner K&uuml;cker
 */
public abstract class AsmTypePath
{
	/**
	 * A type path step that steps into the element type of an array type. See
	 * {@link TypePath#getStep getStep}.
	 */
	public static final class ARRAY_ELEMENT
	extends AsmTypePath
	{
		public static final ARRAY_ELEMENT FIRST_INSTANCE = new ARRAY_ELEMENT( null );

		/**
		 * Konstruktor.
		 */
		public ARRAY_ELEMENT(
				final AsmTypePath next )
		{
			super( next );
		}

		public static ARRAY_ELEMENT createForDimension(
				final int createDimension )
		{
			if ( createDimension < 1 )
			{
				throw new IllegalArgumentException( String.valueOf( createDimension ) );
			}

			ARRAY_ELEMENT tmp = FIRST_INSTANCE;

			for ( int currCreateDimension = 1 ; currCreateDimension < createDimension ; currCreateDimension++ )
			{
				tmp = (ARRAY_ELEMENT) tmp.cloneAdd( FIRST_INSTANCE );
			}

			return tmp;
		}

		/**
		 * {@inheritDoc}
		 */
		@Override
		public String toString()
		{
			return "[" + ( next != null ? String.valueOf( next ) : "" );
		}

		@Override
		protected ARRAY_ELEMENT cloneAddNotRecursiv(
				final AsmTypePath typePathToAdd )
		{
			//if ( this.next != null )
			//{
			//	throw new IllegalStateException();
			//}

			return new ARRAY_ELEMENT(
					typePathToAdd );
		}
	}

	/**
	 * A type path step that steps into the nested type of a class type. See
	 * {@link TypePath#getStep getStep}.
	 */
	public static final class INNER_TYPE
	extends AsmTypePath
	{
		/**
		 * Konstruktor.
		 */
		public INNER_TYPE(
				final AsmTypePath next )
		{
			super( next );
		}

		/**
		 * {@inheritDoc}
		 */
		@Override
		public String toString()
		{
			return "." + ( next != null ? String.valueOf( next ) : "" );
		}

		@Override
		protected INNER_TYPE cloneAddNotRecursiv(
				final AsmTypePath typePathToAdd )
		{
			//if ( this.next != null )
			//{
			//	throw new IllegalStateException();
			//}

			return new INNER_TYPE(
					typePathToAdd );
		}
	}

	/**
	 * A type path step that steps into the bound of a wildcard type. See
	 * {@link TypePath#getStep getStep}.
	 */
	public static final class WILDCARD_BOUND
	extends AsmTypePath
	{
		/**
		 * Konstruktor.
		 */
		public WILDCARD_BOUND(
				final AsmTypePath next )
		{
			super( next );
		}

		/**
		 * {@inheritDoc}
		 */
		@Override
		public String toString()
		{
			return "*" + ( next != null ? String.valueOf( next ) : "" );
		}

		@Override
		protected WILDCARD_BOUND cloneAddNotRecursiv(
				final AsmTypePath typePathToAdd )
		{
			//if ( this.next != null )
			//{
			//	throw new IllegalStateException();
			//}

			return new WILDCARD_BOUND(
					typePathToAdd );
		}
	}

	/**
	 * A type path step that steps into a type argument of a generic type. See
	 * {@link TypePath#getStep getStep}.
	 */
	public static final class TYPE_ARGUMENT
	extends AsmTypePath
	{
		/**
		 * Typ-Argument-Index.
		 */
		final int typeArgumentIndex;

		/**
		 * Konstruktor.
		 */
		public TYPE_ARGUMENT(
				final int typeArgumentIndex ,
				final AsmTypePath next )
		{
			super( next );
			this.typeArgumentIndex = typeArgumentIndex;
		}

		/**
		 * {@inheritDoc}
		 */
		@Override
		public String toString()
		{
			return String.valueOf( typeArgumentIndex ) + ";" + ( next != null ? String.valueOf( next ) : "" );
		}

		/* (non-Javadoc)
		 * @see java.lang.Object#hashCode()
		 */
		@Override
		public int hashCode() {
			final int prime = 31;
			int result = super.hashCode();
			result = prime * result + typeArgumentIndex;
			return result;
		}

		/* (non-Javadoc)
		 * @see java.lang.Object#equals(java.lang.Object)
		 */
		@Override
		public boolean equals(
				final Object obj )
		{
			if (this == obj)
			{
				return true;
			}
			if (!super.equals(obj))
			{
				return false;
			}
			if (getClass() != obj.getClass()) {
				return false;
			}
			TYPE_ARGUMENT other = (TYPE_ARGUMENT) obj;
			if (typeArgumentIndex != other.typeArgumentIndex) {
				return false;
			}
			return true;
		}

		@Override
		protected TYPE_ARGUMENT cloneAddNotRecursiv(
				final AsmTypePath typePathToAdd )
		{
			//if ( this.next != null )
			//{
			//	throw new IllegalStateException();
			//}

			return new TYPE_ARGUMENT(
					typeArgumentIndex ,
					typePathToAdd );
		}
	}

	/**
	 *Weil das verkackte ASM <code>null</code> als
	 *gltigen Typ-Pfad verwendet, muss ich das hier nachbilden,
	 *Null-Objekt-Pattern.
	 */
	public static final class NULL
	extends AsmTypePath
	{
		public static final NULL INSTANCE = new NULL();
		
		/**
		 * Konstruktor.
		 */
		private NULL()
		{
			super( null );
		}

		/* (non-Javadoc)
		 * @see java.lang.Object#hashCode()
		 */
		@Override
		public int hashCode()
		{
			return this.getClass().getName().hashCode();
		}

		/* (non-Javadoc)
		 * @see java.lang.Object#equals(java.lang.Object)
		 */
		@Override
		public boolean equals(
				final Object obj )
		{
			if (this == obj)
			{
				return true;
			}
			if (getClass() != obj.getClass()) {
				return false;
			}
			return true;
		}

		/**
		 * {@inheritDoc}
		 */
		@Override
		public String toString()
		{
			return "NULL";
		}

		@Override
		protected AsmTypePath cloneAddNotRecursiv(
				final AsmTypePath typePathToAdd )
		{
			return typePathToAdd;
		}

		/* (non-Javadoc)
		 * @see de.heinerkuecker.asm_improved.type.AsmTypePath#cloneAdd(de.heinerkuecker.asm_improved.type.AsmTypePath)
		 */
		//@Override
		//public AsmTypePath cloneAdd(
		//		final AsmTypePath typePathToAdd )
		//{
		//	???return typePathToAdd;
		//}
	}

	/**
	 * Nachfolger, darf <code>null</coe> sein.
	 */
	public final AsmTypePath next;

	/**
	 * Konstruktor.
	 * 
	 * @param stepKind
	 * @param next
	 */
	protected AsmTypePath(
			final AsmTypePath next )
	{
		//this.stepKind = stepKind;
		this.next = next;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((next == null) ? 0 : next.hashCode());
		return result;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		AsmTypePath other = (AsmTypePath) obj;
		if (next == null) {
			if (other.next != null) {
				return false;
			}
		} else if (!next.equals(other.next)) {
			return false;
		}
		return true;
	}

	public static AsmTypePath fromNativeTypePath(
			final TypePath nativeTypePath )
	{
		if ( nativeTypePath == null )
		{
			// Null-Objekt, damit spteres Hinzufgen keine Null-Prfung erfordert
			return NULL.INSTANCE;
		}

		final AsmTypePath result =
				fromString(
						nativeTypePath.toString() );

		if ( ! result.toString().equals( nativeTypePath.toString() ) )
			// TODO only for develop
		{
			throw new IllegalStateException(
					nativeTypePath.toString() +
					" " +
					result.toString() );
		}

		return result;
	}

	private static AsmTypePath fromString(
			final String typePathStr )
	{
		if ( typePathStr.isEmpty() )
			// Rekursion terminiert
		{
			// hier wirklich null, weil kein weiteres Element zu einem bergeordneten Pfad-Element hinzugefgt werden soll
			return null;
		}

		if ( typePathStr.charAt( 0 ) == '[' )
		{
			return new ARRAY_ELEMENT(
					fromString(
							typePathStr.substring(
									1 ) ) );
		}
		else if ( typePathStr.charAt( 0 ) == '.' )
		{
			return new INNER_TYPE(
					fromString(
							typePathStr.substring(
									1 ) ) );
		}
		else if ( typePathStr.charAt( 0 ) == '*' )
		{
			return new WILDCARD_BOUND(
					fromString(
							typePathStr.substring(
									1 ) ) );
		}
		else if ( Character.isDigit( typePathStr.charAt( 0 ) ) )
		{
			final StringBuilder buff = new StringBuilder();
			int digitIndex = 0;
			while ( digitIndex < typePathStr.length() &&
					Character.isDigit( typePathStr.charAt( digitIndex ) ) )
			{
				buff.append( typePathStr.charAt( digitIndex ) );
				digitIndex++;
			}

			if ( digitIndex < typePathStr.length() &&
					typePathStr.charAt( digitIndex ) == ';' )
			{
				digitIndex++;
			}

			final int typeArgIndex =
					Integer.parseInt(
							buff.toString() );

			return new TYPE_ARGUMENT(
					typeArgIndex ,
					fromString(
							typePathStr.substring(
									digitIndex ) ) );
		}
		else
		{
			throw new IllegalArgumentException(
					typePathStr );
		}
	}

	/**
	 * Addieere rekursiv wegen Unvernderlichkeit.
	 * 
	 * @param typePathToAdd
	 * @return
	 */
	public AsmTypePath cloneAdd(
			final AsmTypePath typePathToAdd )
	{
		//System.out.println( new Exception().getStackTrace()[ 0 ] + " " + this + " " + System.identityHashCode( this ) );
		if ( this.next == null )
		{
			return cloneAddNotRecursiv(
					typePathToAdd );
		}

		return cloneAddNotRecursiv(
				this.next.cloneAdd(
						typePathToAdd ) );
	}

	abstract protected AsmTypePath cloneAddNotRecursiv(
			final AsmTypePath typePathToAdd );

	public static ImmutableArray<? extends AsmType> typeArgsCloneAddTypeAnnotations(
			final ImmutableArray<? extends AsmType> typeArgsToAddTypeAnnotations ,
			final Map<AsmTypePath, ImmutableArray<AsmAnnotationUse>> typeAnnotationsMap ,
			final AsmTypePath typePath )
	{
		if ( typeArgsToAddTypeAnnotations == null || typeArgsToAddTypeAnnotations.isEmpty() )
		{
			return null;
		}

		final List<AsmType> result = new ArrayList<>();

		for ( int index = 0 ; index < typeArgsToAddTypeAnnotations.size() ; index++ )
		{
			result.add(
					typeArgsToAddTypeAnnotations.get( index ).cloneAddTypeAnnotations(
							typeAnnotationsMap ,
							typePath.cloneAdd(
									new TYPE_ARGUMENT(
											index ,
											null ) ) ) );
		}

		return ImmutableArray.create( result );
	}

	public AsmTypePath cloneAddInnerClasses(
			final int innerClassCount )
	{
		if ( innerClassCount < 0 )
		{
			throw new IllegalArgumentException();
		}
		if ( innerClassCount == 0 )
		{
			return this;
		}
		return this.cloneAdd(
				new AsmTypePath.INNER_TYPE(
						null ) ).cloneAddInnerClasses(
								innerClassCount - 1 );
	}
}
