package de.heinerkuecker.asm_improved.util;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

/**
 * Bei inneren oder geschachtelten (nested)
 * Klassen werden die Namens-Bestandteile
 * im Quelltext mit Punkten abgetrennt,
 * aber die Klassen-Datei hat einen
 * mit Dollar-Zeichen abgetrennten
 * Namen und so taucht der Name
 * auch in ASM auf.
 * 
 * @author Heiner K&uuml;cker
 */
public final class DottedName
{
	/**
	 * Instanz fr {@link Fqn#JAVA_LANG_OBJECT_FQN}.
	 */
	public static final DottedName OBJECT_DOTTED_NAME =
			valueOfNoDotSeparatedStr(
					"Object" );
	/**
	 * String-Inhalt.
	 */
	//public final String value;

	/**
	 * String-Liste mit den einzelnen Namens-Teilen.
	 */
	public final List<? extends String> parts;

	/**
	 * Konstruktor ausschliesslich fr ersten Namens-Teil.
	 * 
	 * @param value see {@link #value}
	 */
	private DottedName(
			final String value )
	{
		// Pruefung auf JavaIdentifier wegen anonymen Klassen entfernt
		//if ( ! ( JavaIdentifier.is( value ) ||
		//		"package-info".equals( value ) ) )
		//{
		//	throw new IllegalArgumentException(
		//			value );
		//}

		//this.value = value;

		if ( value.contains( "." ) )
		{
			throw new IllegalArgumentException(
					value );
		}

		this.parts = Collections.singletonList( value );
	}

	private DottedName(
			final List<? extends String> parts )
	{
		if ( parts.size() < 1 )
		{
			throw new IllegalArgumentException(
					"parts is empty" );
		}

		for ( final String part : parts )
		{
			if ( part.contains( "." ) )
			{
				throw new IllegalArgumentException(
						part );
			}
		}

		this.parts = parts;
	}

	/**
	 * @return dot separated Name
	 * 
	 * TODO rename to toJavaSourceName (Name, wie im Quellcode verwendet)
	 */
	public String toJavaName() {
		return ( new SeparatedStrBuff( "." ) ).addAll( this.parts ).toString();
	}

	/**
	 * @return Name(String) des entsprechenden *.class-File
	 */
    public String toClassFileName()
    {
    	return new SeparatedStrBuff( "$" ).addAll( this.parts ).toString();
    }

    @Override
    public int hashCode()
    {
    	//final int prime = 31;
    	//int result = 1;
    	//result = prime * result + (/*(value == null) ? 0 :*/ value.hashCode());
    	//return result;

	    //return value.hashCode();

    	return this.parts.hashCode();
    }

    @Override
    public boolean equals(
    		final Object obj )
    {
	    if (this == obj)
		    return true;
	    if (obj == null)
		    return false;
	    if (getClass() != obj.getClass())
		    return false;
	    final DottedName other = (DottedName) obj;
	    ///*if (value == null)
	    //{
	    //    if (other.value != null)
	    //	    return false;
	    //} else*/ if (!value.equals(other.value))
	    //    return false;
	    //return true;
	    return this.parts.equals( other.parts );
    }

    @Override
    public String toString()
    {
	    //return value;
    	return toJavaName();
    }

    public String simpleName()
    {
    	//if ( this.value.contains( "." ) )
    	//{
    	//	return this.value.substring(
    	//			//beginIndex
    	//			this.value.lastIndexOf( '.' ) + 1 );
    	//}
    	//return this.value;
    	return this.parts.get( this.parts.size() - 1 );
    }

    public DottedName concatMemberClassName(
    		final String memberClassName )
    {
    	// TODO bei lokalen und anonymen Klasse schlaegt folgende Pruefung fehl
    	//if ( ! JavaIdentifier.is( memberClassName ) )
    	//{
    	//	throw new IllegalArgumentException(
    	//			memberClassName );
    	//}

    	//return new DottedName( this.value + "." + memberClassName );

    	final ArrayList<String> tmpParts = new ArrayList<>();

    	tmpParts.addAll( this.parts );
    	tmpParts.add( memberClassName );

    	tmpParts.trimToSize();

    	return new DottedName( Collections.unmodifiableList( tmpParts ) );
    }

    public DottedName concatMemberClassName(
    		final DottedName memberClassName )
    {
    	final ArrayList<String> tmpParts = new ArrayList<>();

    	tmpParts.addAll( this.parts );
    	tmpParts.addAll( memberClassName.parts );

    	tmpParts.trimToSize();

    	return new DottedName( Collections.unmodifiableList( tmpParts ) );
    }
    
    /**
     * Factory-Methode.
     * 
     * @param noDotSeparatedStr Name (String) ohne Punkte
     * @return Instance
     */
    public static DottedName valueOfNoDotSeparatedStr(
    		final String noDotSeparatedStr )
    {
    		return new DottedName( noDotSeparatedStr );
    }
    
    /**
     * Factory-Methode.
     * 
     * @param dotSeparatedStr
     * @return Instance
     */
    public static DottedName valueOfDotSeparatedStr(
    		final String dotSeparatedStr )
    {
    	if ( ! dotSeparatedStr.contains( "." ) )
    	{
    		return new DottedName( dotSeparatedStr );
    	}
    	return new DottedName( Fqn.createNoDotFullQualifiedClassNameParts( dotSeparatedStr ) );
    }

    /**
     * Factory-Methode.
     * 
     * @param noDotSeparatedStrs Namen (Strings) ohne trennende Punkte (die trennenden Punkte werden in dieser Methode implizit eingefuegt)
     * @return Instance
     */
    public static DottedName valueOfNoDotSeparatedStrs(
    		final String... noDotSeparatedStrs )
    {
    	return valueOfNoDotSeparatedStrs( Arrays.asList( noDotSeparatedStrs ) );
    }
    
    /**
     * Factory-Methode.
     * 
     * @param noDotSeparatedStrs Namen (Strings) ohne trennende Punkte (die trennenden Punkte werden in dieser Methode implizit eingefuegt)
     * @return Instance
     */
    public static DottedName valueOfNoDotSeparatedStrs(
    		final Iterable<? extends String> noDotSeparatedStrs )
    {
    	final Iterator<? extends String> iterator = noDotSeparatedStrs.iterator();

    	DottedName tmpDottedName = new DottedName( iterator.next() );

    	while ( iterator.hasNext() )
    	{
    		tmpDottedName =
    				tmpDottedName.concatMemberClassName(
    						iterator.next() );
    	}

    	return tmpDottedName;
    }

    /**
     * @return count of name parts for inner/nested classes
     */
    public int partCount()
    {
    	return this.parts.size();
    }
    
    /**
	 * @return top level class or not
	 */
    public boolean isTopLevel()
    {
	    return this.parts.size() < 2;
    }

    public DottedName getOuterClassName()
    {
    	return new DottedName(
    			this.parts.subList(
    					//fromIndex
    					0 ,
    					//toIndex
    					this.parts.size() - 1 ) );
    }

    /**
     * Teil-Name von links beginnend.
     * 
     * @param leftPartCount left name part count
     * @return instance for left name part count
     */
    public DottedName instanceFromLeftPartCount(
    		final int leftPartCount )
    {
    	if ( leftPartCount < 1 )
    	{
    		throw new IllegalArgumentException();
    	}

    	return valueOfNoDotSeparatedStrs(
    			this.parts.subList(
    					//fromIndex
    					0 ,
    					//toIndex
    					leftPartCount ) );
    }

    /**
     * Ob der bergebene linke Namens-Teil
     * ein linker Namens-Teil ist, der aus eventuell
     * mehreren Parts bestehen kann,
     * und der bergebene rechte Namens-Teil
     * der entsprechend nachfolgende Part
     * dieses {@link DottedName] ist.
     * Beide bergebene Namens-Teile
     * mssen aber nicht den kompletten
     * Namen ergeben sondern drfen ein
     * Teil-Stck sein.
     * 
     * @param leftJavaClassFileName
     * @param rightPartJavaClassFileName
     * @return
     */
    public boolean arePartsOfJavaClassFileName(
    		final String leftJavaClassFileName ,
    		final String rightPartJavaClassFileName )
    {
    	for ( int currPartCount = 1 ; currPartCount < this.partCount() ; currPartCount++ )
    	{
    		if ( instanceFromLeftPartCount( currPartCount ).toClassFileName().equals( leftJavaClassFileName ) &&
    				this.parts.get( currPartCount ).equals( rightPartJavaClassFileName ) )
    		{
    	    	return true;
    		}
    	}
    	return false;
    }


    /**
     * Teil-Name von links beginnend.
     * 
     * @param leftPartCount left name part count
     * @return instance for left name part count
     */
    public DottedName instanceFromLeftPartsJavaClassName(
    		final String leftPartsJavaClassFileName )
    {
    	for ( int currPartCount = 1 ; currPartCount < this.partCount() ; currPartCount++ )
    	{
    		if ( instanceFromLeftPartCount( currPartCount ).toClassFileName().equals( leftPartsJavaClassFileName ) )
    		{
    	    	return instanceFromLeftPartCount( currPartCount );
    		}
    	}
    	throw new IllegalStateException(
    			this +
    			" " +
    			leftPartsJavaClassFileName );
    }
    
}
