package de.heinerkuecker.asm_improved.util;

import java.util.ArrayList;
import java.util.List;
import java.util.jar.JarFile;

/**
 * Vollqualifizierter Klassen-Name
 * mit eindeutiger Trennung
 * Package und Klassen-Name.
 * <br/><br/>
 * Diese Klasse ist vor allem
 * wichtig, um innere/geschachtelte
 * Klassen eindeutig zu
 * bezeichnen.
 * 
 * @author Heiner K&uuml;cker
 */
public final class Fqn
{
	/**
	 * {@link Fqn} fr {@link java.lang.Object}.
	 */
	public static final Fqn JAVA_LANG_OBJECT_FQN =
			new Fqn( java.lang.Object.class );

    /**
     * {@link Fqn} fr {@link java.lang.Object}.
     */
    public static final Fqn JAVA_LANG_CLASS_FQN =
            new Fqn( java.lang.Class.class );

	/**
	 * {@link Fqn} fr {@link java.lang.String}.
	 */
	public static final Fqn JAVA_LANG_STRING_FQN =
			new Fqn( java.lang.String.class );

	/**
	 * {@link Fqn} fr {@link java.lang.Double}.
	 */
	public static final Fqn JAVA_LANG_DOUBLE_FQN =
			new Fqn( java.lang.Double.class );

	/**
	 * {@link Fqn} fr {@link java.lang.Float}.
	 */
	public static final Fqn JAVA_LANG_FLOAT_FQN =
			new Fqn( java.lang.Float.class );

	/**
	 * {@link Fqn} fr {@link java.lang.Long}.
	 */
	public static final Fqn JAVA_LANG_LONG_FQN =
			new Fqn( java.lang.Long.class );

	/**
	 * {@link Fqn} fr {@link java.lang.Integer}.
	 */
	public static final Fqn JAVA_LANG_INTEGER_FQN =
			new Fqn( java.lang.Integer.class );

	/**
	 * {@link Fqn} fr {@link java.lang.Short}.
	 */
	public static final Fqn JAVA_LANG_SHORT_FQN =
			new Fqn( java.lang.Short.class );

	/**
	 * {@link Fqn} fr {@link java.lang.Byte}.
	 */
	public static final Fqn JAVA_LANG_BYTE_FQN =
			new Fqn( java.lang.Byte.class );

	/**
	 * {@link Fqn} fr {@link java.lang.Boolean}.
	 */
	public static final Fqn JAVA_LANG_BOOLEAN_FQN =
			new Fqn( java.lang.Boolean.class );

	/**
	 * {@link Fqn} fr {@link java.lang.Character}.
	 */
	public static final Fqn JAVA_LANG_CHARACTER_FQN =
			new Fqn( java.lang.Character.class );

	/**
	 * Package in Punkt-Notation,
	 * Default-Package als Leerstring codiert.
	 */
	public final PackageName packageName;

	/**
	 * Bei inneren oder geschachtelten (nested)
	 * Klassen werden die Namens-Bestandteile
	 * im Quelltext mit Punkten abgetrennt,
	 * aber die Klassen-Datei hat einen
	 * mit Dollar-Zeichen abgetrennten
	 * Namen und so taucht der Name
	 * auch in ASM auf.
	 */
	public final DottedName dottedName;

	/**
	 * Konstruktor.
	 * 
	 * @param packageName
	 * @param dottedName
	 */
	public Fqn(
			final PackageName packageName ,
			final DottedName dottedName )
	{
		if ( packageName == null )
		{
		    throw new IllegalArgumentException(
					"packageName is null" );
		}

		if ( dottedName == null )
        {
	        throw new IllegalArgumentException(
					"dottedName is null" );
        }

		this.packageName = packageName;
		this.dottedName = dottedName;
	}

	/**
	 * Konstruktor nur fr Tests.
	 * 
	 * @param clazz Klasse, fr welche ein {@link Fqn}-Objekt erzeugt werden soll
	 */
	public Fqn(
			final Class<?> clazz )
	{
		final PackageName packageName =
				new PackageName(
						clazz.getPackage() );

		final DottedName dottedName;
		if ( packageName.isEmpty() )
		{
			dottedName =
					DottedName.valueOfDotSeparatedStr(
							clazz.getCanonicalName() );
		}
		else
		{
			dottedName =
					DottedName.valueOfDotSeparatedStr(
							clazz.getCanonicalName().substring(
									clazz.getPackage().getName().length() +
									//trennenden Punkt berspringen
									1 ) );
		}
		
		this.packageName = packageName;
		this.dottedName = dottedName;
	}
			
	/**
	 * {@inheritDoc}
	 */
    @Override
    public String toString()
    {
    	return toFqnStr();
    }

	/**
	 * @return punktseparierter vollqualfizierter Klassen-Name fr Verwendung in SourceDump
	 */
    public String toFqnStr()
    {
	    if ( this.packageName == null || this.packageName.value == null || this.packageName.value.isEmpty() )
    	{
    		return this.dottedName.toString();
    	}
	    return this.packageName + "." + this.dottedName;
    }

    /**
	 * @return Name wie von {@link JarFile#getEntry}
	 *         oder im Dateisystem(Classpath) erwartet
     */
    public String toClassFileFqnStr()
    {
    	//if ( this.packageName == null || this.packageName.value == null || this.packageName.value.isEmpty() )
    	//{
    	//	return this.dottedName./*value*/toString().replace( '.' , '$' );
    	//}
    	//return this.packageName + "." + this.dottedName./*value*/toString().replace( '.' , '$' );
	    if ( this.packageName == null || this.packageName.isEmpty() )
    	{
    		return this.dottedName.toClassFileName();
    	}
	    return this.packageName.slashed() + "/" + this.dottedName.toClassFileName();
    }

    /**
     * Erzeugen eines neuen {@link Fqn} aus diesem
     * {@link Fqn} und dem bergebenen Namen einer
     * inneren oder geschachtelten Klasse.
     * 
     * @param memberClassName einfacher Name der inneren oder geschachtelten Klasse
     * @return {@link Fqn} der inneren oder geschachtelten Klasse
     */
    public Fqn concatMemberClassName(
    		final String memberClassName )
    {
    	final DottedName concatedDottedName =
    			this.dottedName.concatMemberClassName(
    					memberClassName );

    	return
    			new Fqn(
    					this.packageName , 
    					concatedDottedName );
    }

    /**
     * Erzeugen eines neuen {@link Fqn} aus diesem
     * {@link Fqn} und dem bergebenen Namen einer
     * inneren oder geschachtelten Klasse.
     * 
     * @param memberClassName einfacher Name der inneren oder geschachtelten Klasse
     * @return {@link Fqn} der inneren oder geschachtelten Klasse
     */
    public Fqn concatMemberClassName(
    		final DottedName memberClassName )
    {
    	final DottedName concatedDottedName =
    			this.dottedName.concatMemberClassName(
    					memberClassName );

    	return
    			new Fqn(
    					this.packageName , 
    					concatedDottedName );
    }

	/**
	 * {@inheritDoc}
	 */
    @Override
    public int hashCode()
    {
	    final int prime = 31;
	    int result = 1;
	    result = prime * result
	            + ((packageName == null) ? 0 : packageName.hashCode());
	    result = prime * result
	            + (/*(dottedName == null) ? 0 :*/ dottedName.hashCode());
	    return result;
    }

	/**
	 * {@inheritDoc}
	 */
    @Override
    public boolean equals(
    		final Object obj )
    {
	    if (this == obj)
		    return true;
	    if (obj == null)
		    return false;
	    if (getClass() != obj.getClass())
		    return false;
	    final Fqn other = (Fqn) obj;
	    /*if (packageName == null)
	    {
		    if (other.packageName != null)
			    return false;
	    } else*/ if (!packageName.equals(other.packageName))
		    return false;
	    /*if (dottedName == null)
	    {
		    if (other.dottedName != null)
			    return false;
	    } else*/ if (!dottedName.equals(other.dottedName))
		    return false;
	    return true;
    }

    /**
     * Zerlegen des bergebenen punkt-getrennten Strings.
     * 
     * @param dotSeparatedStr punkt-getrennter String
     * @return Liste mit den punkt-getrennten Teilen des bergebenen Strings
     */
    public static List<? extends String> createNoDotFullQualifiedClassNameParts(
    		final String dotSeparatedStr )
    {
    	final ArrayList<String> result = new ArrayList<>();

    	int currPos = 0;

    	while ( true )
    	{
    		final int nextDotPos =
    				dotSeparatedStr.indexOf(
    						//ch
    						'.' ,
    						//fromIndex
    						currPos );

    		if ( nextDotPos < 0 )
    		{
    			result.add(
    					dotSeparatedStr.substring(
    							//beginIndex
    							currPos ) );

    			return result;
    		}

    		result.add(
    				dotSeparatedStr.substring(
    						//beginIndex
    						currPos ,
    						//endIndex
    						nextDotPos ) );

    		currPos = nextDotPos + 1;
    	}
    }

    /**
     * @param listToJoin zu verbindende Liste
     * @return String bestehend aus den Elementen der bergebenen Liste mit Punkten verbunden
     */
    public static String joinWithDotSeparatorChar(
    		final List<? extends String> listToJoin )
    {
    	final SeparatedStrBuff dotSeparatedStrBuff =
    			new SeparatedStrBuff( "." );

    	dotSeparatedStrBuff.addAll( listToJoin );

    	return dotSeparatedStrBuff.toString();
    }

	/**
	 * @return top level class or not
	 */
 // TODO repariert: @Deprecated
    public boolean isTopLevel()
    {
    	// TODO repariert: bei anonymen und lokalen Klassen liefert diese methode eine falsche aussage
	    return this.dottedName.isTopLevel();
    }

	/**
	 * @return top level fqn
	 */
	public Fqn getToplevelFqn()
	{
		if ( this.isTopLevel() )
		{
			return this;
		}
		
		return new Fqn(
				this.packageName ,
				DottedName.valueOfNoDotSeparatedStr(
						this.dottedName.parts.get( 0 ) ) );
	}

}
