package de.heinerkuecker.asm_improved.util;

import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;

/**
 * Unvernderliches Array mit Generics.
 * 
 * @param <T> Element type
 * @author Heiner K&uuml;cker
 * 
 * TODO unter Java8 kann dies ein Interface mit default-Methoden sein
 */
public final class ImmutableArray<T>
implements SizeableIterable<T>
{
	/**
	 * Inneres gekapseltes Array.
	 */
	private final T[] innerArr;

	/**
	 * Konstruktor.
	 * 
	 * @param arrayToSet Array mit initialen Werten/Objekten
	 */
	@SafeVarargs
	private ImmutableArray(
			final T... arrayToSet )
	{
		this.innerArr = arrayToSet.clone();
	}

	/**
	 * Konstruktor.
	 * 
	 * @param listToSet {@link List} mit initialen Werten/Objekten
	 */
	@SuppressWarnings("unchecked")
	public ImmutableArray(
			final /*List*/Iterable<? extends T> listToSet )
	{
		//this.innerArr = (T[]) listToSet.toArray();
		if ( listToSet instanceof Collection )
		{
			this.innerArr = (T[]) ( (Collection<T>) listToSet ).toArray();
		}
		else
		{
			this.innerArr = (T[]) ( new ExtendedArrayList<T>( listToSet ) ).toArray();
		}
	}

	/**
	 * Zurckgeben eines gewnschten Elements.
	 * 
	 * @param index Index des gewnschten Elements
	 * @return gewnschtes Element
	 */
	public T get(
			final int index )
	{
		return this.innerArr[ index ];
	}

	/**
	 * @return element count
	 */
	public int size()
	{
		return this.length();
	}

	/**
	 * @return element count
	 */
	public int length()
	{
		return this.innerArr.length;
	}

	/**
	 * @return empty or not
	 */
	public boolean isEmpty()
	{
		return this.innerArr.length == 0;
	}

	/**
	 * Contains.
	 * 
	 * @param elemToCheck element to check
	 * @return contains this array the specified element or not
	 * @see Collection#contains(Object)
	 */
	public boolean contains(
			final Object elemToCheck )
	{
		if ( elemToCheck == null )
		{
			for ( final T currElem : innerArr )
			{
				if ( currElem == null )
				{
					return true;
				}
			}
			return false;
		}

		for( final T currElem : innerArr )
		{
			if ( elemToCheck.equals( currElem ) )
			{
				return true;
			}
		}
		return false;
	}

	/**
	 * Contains typesafe.
	 * 
	 * @param immuArrToCheck immutable array to check
	 * @param elemToCheck element to check
	 * @return contains specified array the specified element or not
	 * @see Collection#contains(Object)
	 */
	public static <T> boolean typesafeContains(
			final ImmutableArray<? extends T> immuArrToCheck ,
			final T elemToCheck )
	{
		return immuArrToCheck.contains( elemToCheck );
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Iterator<T> iterator()
	{
		return
				new Iterator<T>()
				{
					private int currIndex = 0;

					@Override
					public boolean hasNext()
					{
						return currIndex < innerArr.length;
					}

					@Override
					public T next()
					{
						if ( ! hasNext() )
						{
							throw new NoSuchElementException();
						}

						return innerArr[ currIndex++ ];
					}

					@Override
					public void remove()
					{
						throw new UnsupportedOperationException();
					}
				};
	}

	/**
	 * Create copy and append.
	 * 
	 * @param elementToAdd element to append
	 * @return new {@link ImmutableArray} with specified element subsequently
	 */
	public ImmutableArray<T> copyAdd(
			final T elementToAdd )
	{
		@SuppressWarnings("unchecked")
		final T[] tmpArr = (T[]) new Object[ this.size() + 1 ];

		System.arraycopy( innerArr , 0 , tmpArr , 0 , innerArr.length );

		tmpArr[ tmpArr.length - 1 ] = elementToAdd;

		return new ImmutableArray<>( tmpArr );
	}

	/**
	 * Create copy and append.
	 * 
	 * @param arrToAppend {@link ImmutableArray} to append
	 * @return new {@link ImmutableArray} with all elements of this and all elements of specified {@link ImmutableArray} subsequently
	 */
	public ImmutableArray<T> copyAddAll(
			final ImmutableArray<? extends T> arrToAppend )
	{
		@SuppressWarnings("unchecked")
		final T[] tmpArr = (T[]) new Object[ this.size() + arrToAppend.size() ];

		System.arraycopy( innerArr , 0 , tmpArr , 0 , innerArr.length );

		System.arraycopy( arrToAppend.innerArr , 0 , tmpArr , innerArr.length , arrToAppend.innerArr.length );

		return new ImmutableArray<>( tmpArr );
	}

	/**
	 * Create copy and insert.
	 * 
	 * @param elementToInsert element to insert
	 * @param insertIndex index of array to insert
	 * @return new {@link ImmutableArray} with specified element inserted at specified index
	 */
	public ImmutableArray<T> copyInsert(
			final T elementToInsert ,
			final int insertIndex )
	{
		@SuppressWarnings("unchecked")
		final T[] tmpArr = (T[]) new Object[ this.size() + 1 ];

		System.arraycopy( innerArr , 0 , tmpArr , insertIndex , innerArr.length );

		System.arraycopy( innerArr , insertIndex + 1 , tmpArr , insertIndex , innerArr.length - insertIndex );

		tmpArr[ insertIndex ] = elementToInsert;

		return new ImmutableArray<>( tmpArr );
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public int hashCode()
	{
		return Arrays.hashCode( this.innerArr );
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean equals(
			final Object obj )
	{
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final ImmutableArray<?> other = (ImmutableArray<?>) obj;
		return Arrays.equals( innerArr , other.innerArr );
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public String toString()
	{
		//return "ImmutableArray [innerArr=" + Arrays.toString(innerArr) + "]";
		return Arrays.toString( this.innerArr );
	}

	/**
	 * Leer-Instanz.
	 */
	private static final ImmutableArray<?> EMPTY = new ImmutableArray<>();

	/**
	 * Factory-Methode null safe.
	 * 
	 * @param <T> Element type
	 * @param arrayToSet Array mit initialen Werten/Objekten
	 * @return Instanz oder <code>null</code>
	 */
	@SafeVarargs
	@SuppressWarnings( "unchecked" )
	public static <T> ImmutableArray<T> create(
			final T... arrayToSet )
	{
		if ( arrayToSet == null )
		{
			return null;
		}
		if ( arrayToSet.length == 0 )
		{
			return (ImmutableArray<T>) EMPTY;
		}
		return new ImmutableArray<T>( arrayToSet );
	}

	/**
	 * Factory-Methode null safe.
	 * 
	 * @param <T> Element type
	 * @param listToSet {@link List} mit initialen Werten/Objekten
	 * @return Instanz oder <code>null</code>
	 */
	@SuppressWarnings( "unchecked" )
	public static <T> ImmutableArray<T> create(
			final /*List*/Iterable<? extends T> listToSet )
	{
		if ( listToSet == null )
		{
			return null;
		}
		//if ( listToSet.isEmpty()  )
		if ( ! listToSet.iterator().hasNext() )
		{
			return (ImmutableArray<T>) EMPTY;
		}
		return new ImmutableArray<T>( listToSet );
	}

	/**
	 * Factory-Methode null safe.
	 * 
	 * @param <T> Element type
	 * @param immuArrToSet {@link ImmutableArray} mit initialen Werten/Objekten
	 * @return Instanz oder <code>null</code>
	 */
	@SuppressWarnings( "unchecked" )
	public static <T> ImmutableArray<T> create(
			final ImmutableArray<? extends T> immuArrToSet )
	{
		if ( immuArrToSet == null )
		{
			return null;
		}
		if ( immuArrToSet.isEmpty()  )
		{
			return (ImmutableArray<T>) EMPTY;
		}
		return new ImmutableArray<T>( immuArrToSet.innerArr );
	}

	/**
	 * Factory-Methode Singleton-Array mit genau einem einzigen Element.
	 * 
	 * @param <T> Element type
	 * @param singleElementToSet initialer Wert/Objekt
	 * @return Instanz oder <code>null</code>
	 */
	@SuppressWarnings( "unchecked" )
	public static <T> ImmutableArray<T> createSingleton(
			final T singleElementToSet )
	{
		return new ImmutableArray<T>( (T[]) new Object[]{ singleElementToSet } );
	}

	/**
	 * Factory-Methode leeres Array.
	 * 
	 * @return leeres Array
	 * @param <T> Element type
	 */
	@SuppressWarnings( "unchecked" )
	public static <T> ImmutableArray<T> empty()
	{
		return (ImmutableArray<T>) EMPTY;
	}
	
}
