package de.heinerkuecker.asm_improved.util;

import java.util.Collection;
import java.util.Collections;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

/**
 * Unvernderliche {@link Map}.
 * 
 * @param <K> the type of keys maintained by this map
 * @param <V> the type of mapped values
 * @author Heiner K&uuml;cker
 */
public abstract class ImmutableMap<K, V>
{
	/**
	 * @return see {@link Map#size}
	 * @see Map#size
	 */
	abstract public int size();

	/**
	 * @return see {@link Map#isEmpty}
	 * @see Map#isEmpty
	 */
	abstract public boolean isEmpty();	

	/**
	 * @return see {@link Map#containsKey}
	 * @see Map#containsKey
	 */
	abstract public boolean containsKey( K key );

	/**
	 * @return see {@link Map#get}
	 * @see Map#get
	 */
	abstract public V get( K key );

	/**
	 * @return see {@link Map#entrySet}
	 * @see Map#entrySet
	 * 
	 * TODO hier sollte ein unveraenderliches Set zurueckgegeben werden
	 */
	abstract public Set<Map.Entry<K, V>> entrySet();
	
	/**
	 * @return see {@link Map#values}
	 * @see Map#values
	 * 
	 * TODO hier sollte eine unveraenderliche Collection zurueckgegeben werden
	 */
	abstract public Collection<V> values();

	/**
	 * Implementierungsklasse des Interface {@link ImmutableMap}.
	 *
	 * @param <K> the type of keys maintained by this map
	 * @param <V> the type of mapped values
	 */
	static class ImmutableMapImpl<K, V>
	extends ImmutableMap<K, V>
	{
		/**
		 * vor Vernderung geschtzte Map
		 */
		private final Map<K, V> innerMap;

		/**
		 * Konstruktor.
		 * 
		 * @param mapToWrap
		 */
        private ImmutableMapImpl(
        		final Map<K, V> mapToWrap )
        {
        	if ( mapToWrap == null )
        	{
        		throw new IllegalArgumentException(
        				"mapToWrap is null" );
        	}
        	// TODO um ganz sicher zu gehen, msste die bergebene Map geklont werden
	        this.innerMap = mapToWrap;
        }

		/**
		 * {@inheritDoc}
		 */
        @Override
        public int size()
        {
	        return this.innerMap.size();
        }

		/**
		 * {@inheritDoc}
		 */
        @Override
        public boolean isEmpty()
        {
	        return this.innerMap.isEmpty();
        }

		/**
		 * {@inheritDoc}
		 */
        @Override
        public boolean containsKey(
        		final K key )
        {
	        return this.innerMap.containsKey( key );
        }

		/**
		 * {@inheritDoc}
		 */
        @Override
        public V get(
        		final K key )
        {
	        return this.innerMap.get( key );
        }

		/**
		 * {@inheritDoc}
		 */
        @Override
        public String toString()
        {
	        //return "ImmutableMapImpl [innerMap=" + innerMap + "]";
        	return innerMap.toString();
        }

		/**
		 * {@inheritDoc}
		 */
        @Override
        public Set<Entry<K, V>> entrySet()
        {
	        return Collections.unmodifiableSet( this.innerMap.entrySet() );
        }

		/**
		 * {@inheritDoc}
		 */
		@Override
		public Collection<V> values()
		{
			return Collections.unmodifiableCollection( this.innerMap.values() );
		}

		/* (non-Javadoc)
		 * @see java.lang.Object#hashCode()
		 */
        @Override
        public int hashCode()
        {
	        final int prime = 31;
	        int result = 1;
	        result = prime * result
	                + ((innerMap == null) ? 0 : innerMap.hashCode());
	        return result;
        }

		/* (non-Javadoc)
		 * @see java.lang.Object#equals(java.lang.Object)
		 */
        @Override
        public boolean equals(
        		final Object obj)
        {
	        if (this == obj)
		        return true;
	        if (obj == null)
		        return false;
	        if (getClass() != obj.getClass())
		        return false;
	        final ImmutableMapImpl<?, ?> other = (ImmutableMapImpl<?, ?>) obj;
	        if (innerMap == null)
	        {
		        if (other.innerMap != null)
			        return false;
	        } else if (!innerMap.equals(other.innerMap))
		        return false;
	        return true;
        }

	}

	/**
	 * Leere Implemntierungsklasse des Interface {@link ImmutableMap}.
	 *
	 * @param <K> the type of keys maintained by this map
	 * @param <V> the type of mapped values
	 */
	static class ImmutableMapEmptyImpl<K, V>
	extends ImmutableMap<K, V>
	{
		/**
		 * Singleton-Instanz.
		 */
		static final ImmutableMapEmptyImpl<?, ?> INSTANCE = new ImmutableMapEmptyImpl<>();

		/**
		 * Konstruktor.
		 */
        private ImmutableMapEmptyImpl()
        {
        }

		/**
		 * {@inheritDoc}
		 */
        @Override
        public int size()
        {
	        return 0;
        }

		/**
		 * {@inheritDoc}
		 */
        @Override
        public boolean isEmpty()
        {
	        return true;
        }

		/**
		 * {@inheritDoc}
		 */
        @Override
        public boolean containsKey(
        		final K key )
        {
	        return false;
        }

		/**
		 * {@inheritDoc}
		 */
        @Override
        public V get(
        		final K key )
        {
	        return null;
        }

		/**
		 * {@inheritDoc}
		 */
        @Override
        public String toString()
        {
	        return "empty";
        }

		/**
		 * {@inheritDoc}
		 */
        @Override
        public Set<Entry<K, V>> entrySet()
        {
	        return Collections.<Entry<K, V>>emptySet();
        }
        

		/**
		 * {@inheritDoc}
		 */
		@Override
		public Collection<V> values()
		{
			return Collections.<V>emptySet();
		}
	}

	/**
	 * Factory-Methode.
	 * 
	 * @param <K>
	 * @param <V>
	 * @param mapToWrap
	 * @return Instanz oder <code>null</code>
	 */
	@SuppressWarnings( "unchecked" )
    public static <K, V> ImmutableMap<K, V> create(
			final Map<K, V> mapToWrap )
	{
		if ( mapToWrap == null )
		{
			return null;
		}
		if ( mapToWrap.isEmpty() )
		{
			return (ImmutableMap<K, V>) ImmutableMapEmptyImpl.INSTANCE;
		}
		return new ImmutableMapImpl<>( mapToWrap );
	}

	/**
	 * @return empty Map
	 * @param <K> the type of keys maintained by returned map
	 * @param <V> the type of mapped values
	 */
    @SuppressWarnings("unchecked")
    public static <K, V> ImmutableMap<K, V> empty()
    {
	    return (ImmutableMap<K, V>) de.heinerkuecker.asm_improved.util.ImmutableMap.ImmutableMapEmptyImpl.INSTANCE;
    }
}
