package de.heinerkuecker.asm_improved.util;

import java.util.HashMap;
import java.util.HashSet;


/**
 * Klasse fr einen bei Bedarf
 * (faul) zu initialisierenden
 * Wert.
 *
 * TODO {@link LazyInitializer} und {@link VisitorFutureExecutor} tun das selbe, LazyInitializer ist weiter entwickelt
 *
 * @param <T> Type-Parameter, Typ des Wertes
 * @author Heiner K&uuml;cker
 */
public abstract class LazyInitializer<T>
{
    /**
     * Bei Bedarf (faul) zu
     * initialisierender Wert.
     */
    private T t;

    /**
     * Merker, ob die Initialisierung
     * bereits erfolgte.
     * Das Prfen der Initialisierung
     * durch Vergleich des vermerkten
     * Wertes mit <code>null</code> knnte
     * inkorrekt sein, weil auch
     * <code>null</code> ein erlaubter
     * Wert ist.
     */
    private boolean isInitialized = false;

    /**
     * Von Client-Seite auzurufende Methode
     * zum Beschaffen des verzgerten Wertes.
     *
     * @return faul (lazy) initialisierter Wert
     */
    public final T get()
    {
        if ( false == this.isInitialized )
        {
            this.t = init();
            this.isInitialized = true;
        }
        return this.t;
    }

    /**
     * @return ob die Initialisierung erfolgte
     */
    public final boolean isInitialized()
    {
        return this.isInitialized;
    }

    /**
     * Zu implementierende Methode zum
     * Initialisieren des Wertes.
     *
     * @return initialisierter Wert
     */
    abstract protected T init();

    /**
     * Setzen des Wertes unter Umgehung der
     * Backend-Logik.
     *
     * @param value value to set
     */
    public final void set(
            final T value )
    {
        this.isInitialized = true;
        this.t = value;
    }

    /**
     * Verwendung als Key in {@link HashMap}/{@link HashSet} wegen Vernderlichkeit nicht mglich
     */
    @Override
    public int hashCode()
    {
        throw new UnsupportedOperationException();
    }

    /**
     * Verwendung als Key in {@link HashMap}/{@link HashSet} wegen Vernderlichkeit nicht mglich
     */
    @Override
    public boolean equals(
            final Object obj )
    {
        throw new UnsupportedOperationException();
    }

}
