package de.heinerkuecker.asm_improved;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import org.junit.Assert;
import org.junit.Test;
import org.junit.runners.Suite.SuiteClasses;

import de.heinerkuecker.asm_improved.annotation_use.AsmAnnotationUse;
import de.heinerkuecker.asm_improved.type.AsmClassType;
import de.heinerkuecker.asm_improved.util.ContainsTypesafeSet;
import de.heinerkuecker.asm_improved.util.Fqn;
import de.heinerkuecker.asm_improved.util.ImmutableArray;

/**
 * JUnit4-Testcase fr die Klasse {@link AsmClassInfo}.<br/>
 * Pr&uuml;fen, ob jeder untergeordnete Test in mindestens einer Test-Suite notiert ist
 * und alle Tests von {@link AllAsmTests} aufgerufen werden.
 * 
 * @author Heiner K&uuml;cker
 */
public class AllTestsInTestsuiteTest
{
	/**
	 * Die {@link Fqn} der Wurzel-JUnit-Test-Suite.
	 */
	private static final Fqn ROOT_TEST_SUITE_FQN = new Fqn( AllAsmTests.class );
	
	/**
	 * Test method for {@link AsmClassInfo}.
	 */
	@Test
	public final void testAllJUnitTestcaseReachableFromRootTestsuite()
	{
		/*
		 * Suchen nach Tests in Test-Suites von der Wurzel-Test-Suite aus
		 */
		final AsmClasspathInfo asmClasspathInfo =
				new AsmClasspathInfo(
						AsmClasspathInfo.Mode.EAGER ,
						AllAsmTests.RT_JAR_FILE ,
						AllAsmTests.BIN_DIR_FILE );

		final AsmClassInfo rootTestSuiteAsmClassInfo = asmClasspathInfo.getClassInfoMandantory( ROOT_TEST_SUITE_FQN );
		
		final Set<Fqn> fromRootTestSuiteCalledTestClassFqns =
				getFromRootTestSuiteCalledTestClassFqns(
						rootTestSuiteAsmClassInfo ,
						asmClasspathInfo );
		
		final ContainsTypesafeSet<Fqn> typesafeFromTestRootSuitCalledTestClassFqns =
				ContainsTypesafeSet.create( fromRootTestSuiteCalledTestClassFqns );

		/*
		 * Suchen nach Test-Klassen
		 */
		for ( final AsmClassInfo classInfo : asmClasspathInfo.getAllClassInfos() )
		{
			if ( isTestCase( classInfo ) )
			{
				if ( ! typesafeFromTestRootSuitCalledTestClassFqns.contains( classInfo.fqn ) )
				{
					Assert.fail(
							"no from root suite called test " +
									classInfo.fqn.toFqnStr() );
				}
			}

		}
	}

	private Set<Fqn> getFromRootTestSuiteCalledTestClassFqns(
			final AsmClassInfo rootTestSuiteAsmClassInfo ,
			final AsmClasspathInfo asmClasspathInfo )
	{
		final AsmAnnotationUse suiteClassesAnnotation = getTestSuiteAnnotation( rootTestSuiteAsmClassInfo );

		if ( suiteClassesAnnotation == null )
		{
			throw new IllegalStateException( "annotation of class " + SuiteClasses.class.getSimpleName() + " not found" );
		}

		final Set<Fqn> result = new HashSet<>();

		final Object[] suiteClassesAnnotationValueObjArr = (Object[]) suiteClassesAnnotation.values.get( "value" );

		for ( final Object suiteClassesAnnotationValueObj : suiteClassesAnnotationValueObjArr )
		{
			final AsmClassType suiteClassType = (AsmClassType) suiteClassesAnnotationValueObj;

			result.add( suiteClassType.fqn );
			
			final AsmClassInfo suiteClassInfo =
					asmClasspathInfo.getClassInfoMandantory(
							suiteClassType.fqn );
			
			result.addAll(
					getTestSuiteCalledTestClassFqns(
							suiteClassInfo ,
							asmClasspathInfo ) );
		}

		return result;
	}

	private Set<Fqn> getTestSuiteCalledTestClassFqns(
			final AsmClassInfo rootTestSuiteAsmClassInfo ,
			final AsmClasspathInfo asmClasspathInfo )
	{
		final AsmAnnotationUse suiteClassesAnnotation = getTestSuiteAnnotation( rootTestSuiteAsmClassInfo );

		if ( suiteClassesAnnotation == null )
		{
			return Collections.emptySet();
		}

		final Set<Fqn> result = new HashSet<>();
		
		final Object[] suiteClassesAnnotationValueObjArr = (Object[]) suiteClassesAnnotation.values.get( "value" );

		for ( final Object suiteClassesAnnotationValueObj : suiteClassesAnnotationValueObjArr )
		{
			final AsmClassType suiteClassType = (AsmClassType) suiteClassesAnnotationValueObj;

			result.add( suiteClassType.fqn );

			final AsmClassInfo suiteClassInfo =
					asmClasspathInfo.getClassInfoMandantory(
							suiteClassType.fqn );
			
			result.addAll(
					getTestSuiteCalledTestClassFqns(
							suiteClassInfo ,
							asmClasspathInfo ) );
		}

		return result;
	}

	/**
	 * {@link Fqn} der Annotation, welche an Methoden notiert wird,
	 * die von Junit4 als Tests ausgef&uuml;hrt werden.
	 */
	private static final Fqn TEST_METHOD_ANNOTATION_FQN = new Fqn( org.junit.Test.class );

	private static AsmAnnotationUse getTestSuiteAnnotation(
			final AsmClassInfo classInfo )
	{
		final ImmutableArray<? extends AsmAnnotationUse> rootTestSuiteAnnotations = classInfo.annotations;

		AsmAnnotationUse suiteClassesAnnotation = null;

		if ( rootTestSuiteAnnotations != null )
		{
			for ( final AsmAnnotationUse annotation : rootTestSuiteAnnotations )
			{
				if ( annotation.annotationType.fqn.equals( new Fqn( SuiteClasses.class ) ) )
				{
					suiteClassesAnnotation = annotation;
					break;
				}
			}
		}
		return suiteClassesAnnotation;
	}
	
	/**
	 * @param classInfo zu pr&uuml;fende Java-Klasse
	 * @return ob es sich bei der &uuml;rgebenen Java-Klasse einen Testcase handelt
	 */
	private static boolean isTestCase(
			final AsmClassInfo classInfo )
	{
		if ( classInfo.methods != null )
		{
			for ( final ImmutableArray<? extends AsmMethodInfo> methods : classInfo.methods.values() )
			{
				for ( final AsmMethodInfo method : methods )
				{
					if ( method.annotations != null )
					{
						for ( AsmAnnotationUse annotation : method.annotations )
						{
							if ( annotation.annotationType.fqn.equals( TEST_METHOD_ANNOTATION_FQN ) )
							{
								return true;
							}
						}
					}
				}

			}
		}

		return false;
	}
	
	/**
	 * Ermitteln aller untergeordneten Tests aus der
	 * &uuml;bergebenen {@link org.junit.runners.Suite.SuiteClasses}-Annotation.
	 * 
	 * @param testSuiteAnnotation Annotation {@link org.junit.runners.Suite.SuiteClasses} mit Testklassen(Testcases)
	 * @return Testklassen(Testcases) der &uuml;bergebenen Annotation als {@link Fqn}s
	 */
	private static Collection<? extends Fqn> getSubTestFqns(
			final AsmAnnotationUse testSuiteAnnotation )
	{
		//System.out.println( testSuiteClassInfo.fqn.toFqnStr() );

		if ( ! testSuiteAnnotation.annotationType.fqn.equals( new Fqn( SuiteClasses.class ) ) )
		{
			throw new IllegalArgumentException( String.valueOf( testSuiteAnnotation ) );
		}

		final Object[] testAsmClassTypeObjArr = (Object[]) testSuiteAnnotation.values.get( "value" );

		final Set<Fqn> subTestClassFqns = new HashSet<>();

		for ( final Object testAsmClassTypeObj : testAsmClassTypeObjArr )
		{
			final AsmClassType testAsmClassType = (AsmClassType) testAsmClassTypeObj;

			final Fqn testClassFqn = testAsmClassType.fqn;

			//System.out.println( "\t" + fqn.toFqnStr() );

			subTestClassFqns.add( testClassFqn );
		}

		return subTestClassFqns;
	}

}
